;       Last change    August 1993
(in-package epilib)

(tweakable '*key-threshold* 14 
  "Interest threshold beyond which all literals are keys"
  '(float * *))
(tweakable '*default-indicators* '(tp.unknown) 
  "List of indicators to use for new predicates"
  'symbol-list)
(tweakable '*useful-nonepisodic-topics* 
  '(tp.causal-relationship tp.kinship tp.happening tp.location)
  "List of topics of useful nonepisodic preds"
  'symbol-list)
(tweakable '*subnet-topics* '(tp.mental-attitude );tp.part
  "Topics which indicate that subnets are appropriate"
  'symbol-list)

(traceable 'class "Traces classification of formulas")
(traceable 'entry-class 
  "Traces classification of asserted and inferred formulas only")
(traceable 'keys "Traces Key selection")

(defvar *typed-variables* nil)
(defvar *variables* nil)
(defvar *current-arg* nil)
(defvar *lambda-arg* nil)
(defvar *embedded* nil)
(defvar *main-class* nil)

; 
; Determination of classifications
;

; each term and subpart may contribute to a particular classification, and
; may also generate classifications of its own
; component from predicate, component from arguments

(defun classify (wff &aux class s tp nclass)
  ; intializations
  ; if leftovers from previous assertion then cleanup (just in case)
  (when *typed-variables* (class-cleanup))
  (setq *main-class* nil)

  ; Select keys, save type information for variables
  (cond ((or (getp wff '+trigger) (getp wff '-trigger))
	 (dolist (a (getp wff 'types))
		 (setp (car a) 'types (cdr a))
		 (push (car a) *typed-variables*))
	 )
	(t 
	 (save-types wff)
	 ; first check to make sure at least one wh-variable is typed
	 (setp wff 'types (mapcar #'(lambda (a) 
				(setq tp (union (getp a 'types)
						(if (setq s (find-sort a))
						    (list s))))
			        (unless (or tp (getp a 'lambda))
					(setq tp '(entity))
					(setp a 'types tp)
					(push a *typed-variables*))
				(cons a tp))
				  *variables*))
	 (keys wff)
	 ))
  ; classification
  (if *meaning-postulate*
      (setq class (mp-classify wff))
      (setq class (cadr (class-wff wff t)))
      )

  ;(setq nclass (remove-if #'(lambda (a) (equal a '(episode ** object))) class))
  ;(if nclass (setq class nclass))
  ;(setq nclass (remove-if #'(lambda (a) (equal a '(episode ** object))) *main-class*))
  ;(if nclass (setq *main-class* nclass))

  (setq class (remove-duplicates class :test 'equal))
  (unless *meaning-postulate*
	  (setq *intermediate-class* class)
	  (if *main-class*
	      (setq class *main-class*)
	      (setq class (main-class *intermediate-class*)))
	  (when (type-lit-p wff)
	        (setq class (push (list (car (extract-type wff))) class))
	        (setq *intermediate-class* 
		   (push (list (car (extract-type wff))) *intermediate-class*))
		)
	  (setq class (remove-duplicates class :test 'equal))
	  )

  (setq nclass (remove-if #'(lambda (a) (equal a '(episode episode-specialist))) class))
  (setq class nclass)
  ;(cond (nclass (setq class nclass))
;	((or *meaning-postulate* *simplification-schema*) nil)
;	(t (setq class nclass)))

  ; cleanup
  (class-cleanup)

  (when (traced-p 'class)
	(print-line "Classified " (getp wff 'print) " under: ")
	(dolist (c class) (print-line "     " c))
	)
  class)

(defun main-class (int-classes &aux tclasses)
  (dolist (c int-classes)
	  (setq tclasses (append tclasses (class-int c t))))
  tclasses)

(defun topical-class (item main-classes &aux tclasses)
  (dolist (c main-classes)
	  (setq tclasses (append tclasses (class-int c nil))))
  (setq tclasses (remove-duplicates tclasses :test 'equal))
  (dolist (c tclasses) (store-class item c t))
  tclasses
  )

(defun class-int (class &optional (main t) &aux tclasses
			      concept topic role)
  (cond ((null class) nil)
	((listp (car class))
	 (let ((first (class-int (car class) main))
	       (second (class-int (cdr class) main)))
	      (when (null second) (setq tclasses first))
	      (dolist (f first)
		      (dolist (s second)
			      (setq tclasses (append tclasses
						     (list (make-class f s))))))
	      ))
	(main
	 (let ((specs (getp (cadr class) 'specialists)))
	      (cond (specs
		     (dolist (s specs)
			     (setq tclasses (append tclasses 
						 (list (list (car class) s))))))
		    (t (setq tclasses (append tclasses (list class))))) ))
	(t (setq concept (car class))
		       (if (cadr class)
	       (setq topic (get-indicators (cadr class))) 
	       (setq topic (list 'tp.spec)))
	   (setq role (caddr class))
	   (dolist (i topic)
		   (cond ((listp i) ;specific argument
			  (if (or (null role) (equal (cadr i) role))
			      (setq tclasses (append tclasses
					     (list (list concept (car i))))))
			  (if (and (numberp (cadr i))
				   (if (equal (cadr i) 1) (equal role 'subject)
				       (member role '(object iobj))))
			      (setq tclasses (append tclasses
					     (list (list concept (car i))))))
			  )
			 (t (setq tclasses (append tclasses
                                 (list (list concept i)))))
			 )) )
	)
  tclasses)

(defun class-cleanup ()
  (dolist (v *typed-variables*) 
	  (remprop v 'types)
	  (remprop v 'interest)
	  (remprop v 'topical-class))
  (setq *typed-variables* nil)
  (setq *variables* nil)
  (setq *current-arg* nil)
  (setq *lambda-arg* nil)
  (setq *embedded* nil)
  )

; all class routines return a list - the first part is a list of concepts,
; the second a list of topics, the third classifications

(defun class-wff (wff &optional (top nil) (subnet '/) &aux 
		      classes lit-classes cl cpart clpart mclass nclasses
		      hardway
		      (emb *embedded*))
  (setq hardway nil)
  (case (getp wff 'wff-type)

	; the first few have literals associated with them
	; just classify the literals and use that?
	((quantified logical modified)
	 ; save the type info on variables
	 ; get combined classifications (use keys only!)
	 (dolist (k (getp wff '+trigger))
		 (cond ((equal k wff) (setq hardway t))
		       (t
		 (setq clpart (class-wff k nil subnet))
		 (setq cl (cadr clpart))
		 (setq classes (append classes cl))
		 (setq cl (main-class cl))
		 (setq mclass (append mclass cl))
		 (push (list k '+ cl (wff-interest k)) lit-classes))  ) )
	 (dolist (k (getp wff '-trigger))
		 (setq clpart (class-wff k nil subnet))
		 (setq cl (cadr clpart))
		 (setq classes (append classes cl))
		 (setq cl (main-class cl))
		 (setq mclass (append mclass cl))
		 (push (list k '- cl (wff-interest k)) lit-classes))
	 (when (or hardway (not (or (getp wff '+trigger) (getp wff '-trigger))))
		 (unless hardway (setq *embedded* t))
		 (if (equal (getp wff 'wff-type) 'quantified)
		     (let* ((r (getp wff 'restrict))
			    (cr (if r (class-wff r nil subnet)))
			    (mr (class-wff (getp wff 'main) nil subnet)))
			   (setq classes (append (cadr cr) (cadr mr)))
			   (setq cpart (union (car cr) (car mr) :test 'equal))
			   )
		     (dolist (cl (mapcar #'(lambda (a) (class-wff a nil subnet)) 
					 (getp wff 'arguments)))
			     (setq classes (append classes (cadr cl)))
			     (setq cpart (union cpart (car cl) :test 'equal))
			     )) 
		 (setq *embedded* emb)
		 )
	 )

	; these are their own literals
	(episodic
	 ; class is that indicated by last argument and predicate,
	 ; as well as all independent classes of 1st argument
	 (let* ((args (getp wff 'arguments))
		(pred (get wff 'pred))
		(lasta (last1 args)) )
	       (if (equal pred '_) (setq *embedded* t))
	       (setq cl (class-op-and-args 'pred pred args subnet))
	       (setq classes (cadr cl))
	       (setq cpart (car cl))
	       (when (and (symbolp (setq lasta (last1 args))) (equal pred '_))
		     (if (epi-variable lasta) (pushnew lasta *typed-variables*))
		     (setq classes (remove-duplicates classes :test 'equal))
		     (setq classes (remove-if #'(lambda (a) 
			(and (equal (cadr a) '_) (equal (caddr a) 'subject)))
					      classes))
		     (cond ((epi-variable-p lasta)
			    (setp lasta 'topical-class classes) 
			    (setp lasta 'interest (update-interest-num (wff-interest wff)) ))
			   (t (change-property lasta 'topical-class classes) 
			      (change-property lasta 'interest 
					       (update-interest-num (wff-interest wff)) )))
		     (setq classes 
			   (remove-if-not #'(lambda (a) (equal (cadr a) '_))
					  classes))
		     )
	       (setq nclasses (remove-if #'(lambda (a) 
                      (and (or (member (car a) 
				       '(propos episode time wff entity))
			       (epi-variable (car a)))
			   (member (cadr a) '(@ * ** _)))) classes))
	       (when nclasses (setq classes nclasses))
	       (setq *embedded* emb)
	       (when top
		     (setq mclass (main-class classes))
		     (setq lit-classes 
			   (list (list wff '+ mclass (wff-interest wff)))))
	       )
	 )

	(causal
	 (let ((args (getp wff 'arguments))
	       (pred (get wff 'pred)))
	      (setq cl (class-op-and-args 'pred pred args subnet))
	      (setq classes (cadr cl))
	      (setq cpart (car cl))
	      (when top
		    (setq mclass (main-class classes))
		    (setq lit-classes 
			  (list (list wff '+ mclass (wff-interest wff)))))
	      )
	 )

	(prefix
	 ; what about modal predicates and subnets?
	 (unless (or top (if *embedded* (would-be-key wff) t))
		 (return-from class-wff nil))
	 (when (equal (getp wff 'pred) 'true)
	       (return-from class-wff (class-item (car (getp wff 'arguments)) subnet)))

	 (setq cl (class-op-and-args 'pred (getp wff 'pred)
				     (getp wff 'arguments) subnet))
	 (setq classes (cadr cl))
	 (setq cpart (car cl))
	 (when top
	       (setq mclass (main-class classes))
	       (setq lit-classes 
		     (list (list wff '+ mclass (wff-interest wff)))))
	 )

	(quasi-quoted-expression
	 (return-from class-wff (class-wff (getp wff 'quote-wff) nil subnet)))

	(quoted-expression (setq cpart (list wff)))

	(constant (setq cpart (list wff)))

	(variable (setq cpart (list 'wff)))

	(otherwise (setq cpart (list wff)))
	)
  (when top 
	(if lit-classes (setp wff 'lit-classes lit-classes))
	(if mclass (setq *main-class* mclass)) )
  (list cpart classes)
  )

; returns a list of items which can be used in the terms position of 
; classification

(defun class-term (term &optional (subnet '/) &aux cpart classes s cl)
  (cond ((symbolp term)
	 (if (epi-variable term) (setp term 'term-type 'variable))
	 (case (getp term 'term-type)
	       (variable
		; types/sorts
		(setq cpart (getp term 'types) )
		(if (setq s (getp term 'sort)) (pushnew s cpart))
		(setq classes (getp term 'topical-class))
		(when (and (equal (getp term 'sort) 'propos) (null classes))
		      (setq classes '((propos _ object)))) ) 
	       (constant 
		(setq cpart (union (list term) (equivalent-items term subnet)))
		(setq classes (getp term 'topical-class))
		(when (and (equal (getp term 'sort) 'propos) (null classes))
		      (setq classes (list (list term '_ 'object)))) )
	       ;; The above causes (wolf1 believe p1) to be classified
	       ;; under (p1 _ object), which seems bizarre.  But I
	       ;; haven't changed it because I'm sure this was put here
	       ;; for a reason, and I haven't figured out what it was. -ANK
	       (modified  
		(if (equal (getp term 'operator) 'to)
		    (setq cl (class-op-and-args 'pred (car (getp term 'arguments))
						(list *current-arg*) subnet))
		(setq cl (class-op-and-args 'operator (getp term 'operator)
						 (getp term 'arguments) subnet)) )
                (setq classes (cadr cl))
		(setq cpart (car cl)) )
	       (function
		; the function itself contributes to the classification
		; and may optionally add its own
		(setq cl (class-op-and-args 'function
						 (getp term 'operator)
						 (getp term 'arguments) subnet))
                (setq classes (cadr cl))
		(setq cpart (car cl)) )
	       (quoted-expression)
	       (quasi-quoted-expression)
	       (record)
	       (wff
		(setq cl (class-wff term nil subnet))
		(setq cpart (car cl))
		(setq classes (cadr cl)) )
	       (otherwise (setq cpart (list term)))
	       ))
	(t nil); if string or number, don't classify under it
	)
  (list cpart classes)
  )

(defun class-pred (pred &optional subnet &aux cpart classes cl)
  (cond ((symbolp pred)
	 (if (epi-variable pred) (setp pred 'pred-type 'variable))
	 (case (getp pred 'pred-type)
	       (variable (setq cpart (list 'pred)))
	       (constant (setq cpart (list pred)))
	       (lambda 
		(if (getp pred 'wff-type)
		    (let ((larg *lambda-arg*))
			 (unless *lambda-arg*
				 (setq *lambda-arg* (getp pred 'variable)))
			 (setq cl (class-wff pred nil subnet))
			 (setq classes (cadr cl)) 
			 (setq cpart (car cl)) 
			 (setq *lambda-arg* larg)
			 )
		    (return-from class-pred
				 (class-pred (getp pred 'operator) subnet))
		    )
		)
	       (modified 
		(setq cl (class-op-and-args 'operator (getp pred 'operator)
						 (getp pred 'arguments) subnet))
                (setq classes (cadr cl))
		(setq cpart (car cl)))
	       (otherwise (setq cpart (list pred)))
	       ))
	; numbers in any case other than implications are not legal and 
	; will not be classified; similarly for strings, etc
	(t nil))
  (list cpart classes)
  )

(defun class-op (op &optional subnet &aux cpart classes cl)
  (cond ((symbolp op)
	 (if (epi-variable op) (setp op 'op-type 'variable))
	 (case (getp op 'op-type)
	       (variable (setq cpart (list 'operator)))
	       (modified 
		(setq cl (class-op-and-args 'operator (getp op 'operator)
					       (getp op 'arguments) subnet))  
		(setq classes (cadr cl))
		(setq cpart (car cl)))
	       (otherwise 
		(if (and (not (getp op 'function))
			 (or (not *insertion*) (getp op 'indicators)))
		    (setq cpart (list op))) )
	       ))
	(t nil))
  (list cpart classes)
  )

; returns only a list of classifications
(defun class-op-and-args (optype op args &optional (subnet '/) &aux classes oclass 
				 indicators arg-classes subclass ac
				 cpart oldclass usedold n-subnet
				 (oarg *current-arg*))
  (when (and (equal optype 'pred) (subnet-pred op))
	(setq n-subnet (build-symbol (append (list subnet)
		   (remove (last1 args) args) (list '-) (list op)))) )
  (setq oclass (case optype
		     (pred (class-pred op subnet))
		     ((op function) (class-op op subnet))
		     (otherwise (class-pred op subnet))) )
  (setq indicators (car oclass))
  (setq classes (cadr oclass))
  (dolist (a args)
	  (when (and (equal optype 'pred) (symbolp a)
		     (member (getp a 'term-type) '(constant variable))
		     )
		(setq *current-arg* a) )
	  (setq ac nil)
	  (when (and (epi-variable a) (getp a 'lambda) oarg (equal a *lambda-arg*))
		(setq ac (class-item oarg subnet))
		(setq oarg nil)
		)
	  (unless ac (setq ac (if *meta* (class-term a subnet) 
				  (class-item a (if (and (equal a (last1 args)) n-subnet) n-subnet subnet)))))
	  (setq arg-classes (append arg-classes (list (car ac))))
	  (unless (and (type-pred-p op) (equal op (getp op 'sort)))
		  (setq classes (append classes (cadr ac))) ))
	
  (setq oldclass classes)
  (setq classes nil)

  (when (and (equal optype 'operator) (symbolp op) (getp op 'indicators))
        (setq arg-classes (append (list nil) arg-classes))
        (setq cpart (list op)) )
  (when (or (equal optype 'pred) (and (symbolp op) (getp op 'indicators)))
  ; use indicators with args
  (dolist (i indicators) ; each is an operator or predicate name
	  (when (member-if 'subnet-pred indicators)
		(setq subclass oldclass)
		(unless subclass (setq subclass '((propos _ object))))
		(setq usedold t) )
	  (setq classes (append classes 
	     (class-with-role subclass i arg-classes)))
	  ) )
  (unless (and usedold classes) (setq classes (append classes oldclass)))
                      
  (setq *current-arg* oarg)
  (unless (or cpart (and (equal optype 'pred) (not (type-pred-p op))))
	  (setq cpart (remove-duplicates (apply 'append arg-classes))))
  (list cpart classes)
  )

;;; With the addition of the belief specialist, I (ANK) have modified
;;; subnet-pred so that subnets are not maintained for a verb (such as
;;; 'believe) in which some specialist is interested, because the
;;; specialist interface can't handle it.  The problem is that whereas
;;; ordinarily
;;;
;;;            (lrrh believe (that (wolf1 wolf)))
;;;
;;; would be classified under
;;;
;;;            ((lrrh believe subject) (wolf1 type-specialist)) 
;;;            (wolf1 say object),
;;;
;;; when 'believe' is a specialist predicate they get classified under
;;;
;;;            (lrrh belief-specialist)
;;;            (wolf1 belief-specialist)
;;;            (wolf1 type-specialist)
;;;
;;; because subject, object, and iobject classifications are combined
;;; for subnet verbs.  If we continued to maintain subnets for
;;; 'believe' in this situation, the (x belief-specialist)
;;; classification would contain a combination of wffs (ones that
;;; would have been classified under (x believe object)) and subnets
;;; (ones that would have been classified under (x believe subject).
;;; This causes all kinds of confusion.

;;; This is a rather ad hoc solution, and there might be some reason
;;; to want to maintain belief subnets in addition to simulations.  If
;;; so, perhaps we could avoid the combination of subject and
;;; object classifications for subnet verbs.

(defun subnet-pred (item)
  (and (not (getp item 'specialists))
       (member-if #'(lambda (a)
		      (member (if (listp a) (car a) a)
			      *subnet-topics*))
		  (if (symbolp item) (getp item 'indicators)))))


	
(defun class-with-role (subclass op args)
  (apply 'append (mapcar #'(lambda (a b) (class-it subclass op a b))
			 args (make-role-list (list-length args)) ) )
  )

(defun class-it (subclass ind arg &optional role &aux classes nclass)
  (dolist (a arg)
	  (setq classes (append classes
               (list (if role (list a ind role) (list a ind)) )))
	  )
  (when (and subclass (member role '(subject iobj)))
        (dolist (s subclass)
	        (setq nclass (append nclass
                   (apply 'append (mapcar #'(lambda (a) (list (make-class a s)))
					  classes)))))
        (setq classes nclass))
  classes)

(defun make-role-list (num)
  (cond ((equal num 1) '(subject))
	((equal num 2) '(subject object))
	(t (append '(subject iobj) 
		   (make-list (- num 2) :initial-element 'object)))
	)
  )

(defun default-classes (topic)
  (case topic
	(tp.mental-attitude (list (cons 'propos 'tp.content)))
	(tp.part (list (cons 'entity nil)))
	(otherwise nil)
	)
  )

(defun topical-classes (topic args classes) 
  (if (member topic *subnet-topics*)
      (apply 'append
	     (mapcar #'(lambda (a) 
			       (when (or (not (equal topic 'tp.spec))
					 (type-pred-p a))
		(mapcar #'(lambda (b) (make-class (cons a topic) b))
			(if classes 
			    classes 
			    (default-classes topic)))) )
	      args))
      (apply 'append
      (mapcar #'(lambda (a) (when (or (not (equal topic 'tp.spec))
				      (type-pred-p a))
			(list (cons a topic)))) args)))
  )


(defun class-item (item &optional (subnet '/) &aux (usetype (if (symbolp item)
					  (cond ((entity-type item)) (t 'term))
					  nil)))
  (when usetype
	(case usetype 
	      (wff (class-wff item nil subnet))
	      (pred (class-pred item subnet))
	      (op (class-op item subnet))
	      (term (class-term item subnet))
	      (otherwise (class-term item subnet))
	      ))
  )

(defun get-indicators (item &optional (itype 'pred))
  (cond ((not (symbolp item)) nil)
	((getp item 'indicators))
	((apply 'append (mapcar #'(lambda (a) (getp a 'indicators)) 
				(getp item 'specialists))))
	((equal itype 'pred)
	 (change-property item 'indicators *default-indicators*)
	 (warn-error "Classification: " item " has no topic indicators - assuming "
		     *default-indicators*)
	 *default-indicators*)
	(t nil)
	)
  )

;
; Storage of wffs under classifications
;

(defun classify-and-store (wff &aux class)
  (when (and (equal (wff-type wff) 'quantified)
	     (neg-p wff))
	(warn-error "Normalization: " (print-info wff) 
		    " is a negated quantified expression "
		    *nl* "  No rules have applied to distribute the negation"
		    *nl* "  The formula may not be used properly"))
  (when *save-class*
	(setq class *save-class*)
	(setq *intermediate-class* *save-intermediate*))
  (unless class (setq class (classify wff)) )
  (dolist (c class) (store-class wff c))
  class
  )

(defun store-class (wff class &optional topic-net)
  (unless (member wff (get-class class topic-net))
    (store-class-in-net (if topic-net
			    (environment-topic-net *environment*)
			  (environment-main-net *environment*))
			wff class)))

(defun store-class-in-net (subnet item class &aux l class+ tbl)
  (when (listp (car class))
	(setq class+ (cdr class))
	(setq class (car class))
	)
  ; get current contents of that classification - will be either a list
  ; of propositions or another hash table 
  (setq l (get-class-from-net subnet class nil))
  (if (hash-table-p (car l)) (setq tbl (car l)))

  ; if it is nil, and there is more classifying to do, build a subnet hash
  ; table and store it
  (when (and (null tbl) class+)
	(setq tbl (make-hash-table :test #'equal))
	(change-hash class subnet tbl)
	)

  (if (listp (car class+)) 
      (setq class+ (car class+)))

  ; if there is a further level, continue classifying with new subnet,
  ; else just add this item to the current list of propositions
  (if (hash-table-p tbl)
      (store-class-in-net tbl item class+)
    (if (member item l)
	(assert nil) ;; should no longer happen; see store-class. -ANK
      (change-hash class subnet (insert-proposition item l))))
;       (unless (member item l)
; 	      (change-hash class subnet (insert-proposition item l))))
    )

(defun insert-proposition (wff-name plist); &aux wff-c p-c) comment because are not used
  ; insert wff-name in the appropriate spot in plist and return the
  ; new list

  (cond ((null plist)
	 (list wff-name))
	((proposition-order-lessp wff-name (car plist))
	 (cons wff-name plist))
	(t (cons (car plist)
		 (insert-proposition wff-name (cdr plist))))))

;; The ordering predicate for proposition lists.  Order is first by
;; increasing complexity, then by decreasing probability when there's
;; a tie in complexity.
(defun proposition-order-lessp (wff1 wff2)
  (let ((c1 (getp wff1 'complexity))
	(c2 (getp wff2 'complexity)))
    (cond ((< c1 c2)
	   t)
	  ((and (= c1 c2)
		(> (get-prob wff1)
		   (get-prob wff2)))
	   t)
	  (t nil))))
	   

;
; Manipulation of classifications
;

; climb up and down classifications - take into consideration functions and
; other complex entities - may climb up them as well
; e.g. ((start-of e1) . time-relationship) -> ((start-of episode) . time-rel..)
; also perhaps even (time . time-relationship) if function has known sort


(defun grab-types (arg &aux tps)
  (setq tps (remove-duplicates
  (apply #'append
         (mapcar #'(lambda (tp) (extract-type tp))
                 (get-class (list arg 'type-specialist))))))
  (unless tps (setq tps '(entity)))
  tps
  )

(defun extract-type (wff)
  (when (type-lit-p wff)
	(if (equal (getp wff 'wff-type) 'prefix)
	    (list (getp wff 'pred))
	    (list (getp (car (getp wff 'arguments)) 'pred))))
  )



; sort-classes applies a heuristic to the initial ordering of
; classifications
;   The order is as follows - highest to lowest
;         constant pred role
;         type-pred  pred
;         anything else

(defun sort-classes (cl &optional access &aux clist tlist olist slist tp)
  (dolist (class cl)
          (cond ((type-pred-p (setq tp 
			  (if (listp (car class)) (caar class) (car class))))
                 (if (member tp '(episode set number integer real time))
                     (pushnew class slist :test 'equal)
                 (if (cdr class)
                         (pushnew class tlist :test 'equal)
                     (pushnew class olist :test 'equal))) )
                (t (if (cdr class)
                       (if (and (symbolp tp) (getp tp 'INDEF))
                           (setq clist (append clist (list class)))
                           (pushnew class clist :test 'equal))
                       (pushnew class olist :test 'equal))))
          )
  (setq tlist (sort tlist #'(lambda (classa classb)
           (let* ((tpreda (if (listp (car classa)) (caar classa) (car classa)))
		  (tpredb (if (listp (car classb)) (caar classb) (car classb)))
		  )
		 (when (and (symbolp tpreda)
			    (symbolp tpredb))
		       (order-types tpreda tpredb))
		 ))))
  (if access
      nil
      (setq clist (sort clist #'(lambda (classa classb)
           (let* ((tpreda (if (listp (car classa)) (caar classa) (car classa)))
		  (tpredb (if (listp (car classb)) (caar classb) (car classb)))
		  )
		 (when (and (symbolp tpreda)
			    (symbolp tpredb))
		       (order-instances tpreda tpredb))
		 ))))
      )
  (append clist tlist slist olist)
  )

(defun order-instances (a b &aux int1 int2)
  (cond ((equal a b) t)
	((and (setq int1 (getp a 'interest)) (setq int2 (getp b 'interest))
	      (not (= int1 int2)))
	 (> int1 int2))
	(t t)))

;; returns t if types are in proper order
(defun order-types (a b &aux h int1 int2)
  (cond ((equal a b) nil)
	((and (setq int1 (getp a 'interest))
	      (setq int2 (getp b 'interest))
	      (not (= int1 int2)))
	 (> int1 int2))
	((and (setq int1 (getp a 'pred-interest)) 
	      (setq int2 (getp b 'pred-interest))
	      (not (= int1 int2)))
	 (> int1 int2))
	(t
	 (case (hier-rel a b)
	   ((subsumed equivalent) t)
	   (subsumes nil)
	   (disjoint
	    (when (setq h (car (intersection (getp a 'HIER) (getp b 'HIER))))
	      (if (getp h 'hier-change)
		  (number-hier h))
	      (and (= (get-max h a) (get-id h a))
		   (>= (get-max h b) (get-id h b))))
	    )
	   (otherwise t))
	 )
	)
  )

(defun find-equiv-classes (class &aux eclasses classes e2classes)
  (setq eclasses (equiv-classes (car class)))
  (cond ((null class) nil)
	((null (cdr class)) (mapcar 'list eclasses))
	(t (setq e2classes (find-equiv-classes (cdr class)))
	   (setq classes (mapcar #'(lambda (a) (append (list (car class)) a))
				 e2classes))
	   (dolist (e eclasses)
		   (setq classes (append classes (list (append (list e) (cdr class)))))
		   (dolist (e2 e2classes)
			   (setq classes (append classes (list (append (list e) e2))))
		   ))
	   (remove-duplicates classes))
	)
  )

; Make sure this gets called with subnet-info!!
(defun equivalent-items (concept &optional (subnet-info '/))
  (get-equivalence concept subnet-info)
  )

(defun equiv-classes (class)
  (cond ((null class) nil)
	((listp (car class)) ;modal embedding
	 ; check for equivalent first parts, then repeat for other parts
	 (find-equiv-classes class))
	(t (let ((concept (car class))
		 classes)
		(dolist (e (remove concept (equivalent-items concept)))
			(setq classes (append classes
					  (list (append (list e) (cdr class)))))
			)
		classes))
	)
  )

(defun super-class (class &optional mp already)
  (if (listp (car class)) 
      (remove-duplicates (get-subnet-class nil class already) :test 'equal)
      (remove-duplicates (get-super-class class mp already) :test 'equal))
  )

(defun part-pred-p (pred &aux pp)
  (cond ((setq pp (pred-parts pred))
         (part-pred-p (cadr pp)))
        (t (and (type-pred-p pred)
                (member-if #'(lambda (b)
                                     (or (getp b 'part-type-hier)
                                         (and (getp b 'part-hier)
                                              (not (equal (getp b 'root)
                                                          pred)))))
                           (getp pred 'hier))) )
        ))

(defun gimme-super (item &aux pp)
  (if (and (setq pp (getp item 'pred-parts))
	   (not (member (last1 pp) *remainder-indicators*))
	   (type-pred-p (car pp)) (type-pred-p (cadr pp)))
      (let* ((p1 (car pp))
	     (p2 (cadr pp))
	     (p1sup (gimme-super p1))
	     (p2sup (gimme-super p2))
	     res newp)
	    (if (and (null p1sup) (equal p2 'part))
		(setq res (append res (list 'part))))
	   (dolist (p p1sup)
		   (unless (member p '(entity part))
		   (if (and (part-pred-p p) (equal p2 'part)
			    (equal (cadr (pred-parts p)) 'part))
		       (setq newp (build-symbol 
				   (list (car (pred-parts p)) '- p2) 'epilog))
		       (setq newp (build-symbol (list p '- p2) 'epilog)))
		   (unless (getp newp 'pred-parts)
			   (setp newp 'specialists (getp p2 'specialists))
			   (setp newp 'indicators (getp p2 'indicators))
			   (setp newp 'pred-parts (list p p2)))
		   (setq res (append res (list newp)))) )
	   (dolist (p p2sup)
		   (unless (member p '(entity))
		   (setq newp (build-symbol (list p1 '- p) 'epilog))
		   (unless (getp newp 'pred-parts)
			   (setp newp 'specialists (getp p 'specialists))
			   (setp newp 'indicators (getp p 'indicators))
			   (setp newp 'pred-parts (list p1 p)))
		   (setq res (append res (list newp)))) )
	   res)
      (let ((res (get-all-super item)))
	   (if (and (null res) (equal item 'part))
	       (list 'entity)
	       res)) )
  )
(defun gimme-sub (item &aux pp)
  (if (and (setq pp (getp item 'pred-parts))
	   (not (member (last1 pp) *remainder-indicators*)))
      (let* ((p1 (car pp))
	     (p2 (cadr pp))
	     (p1sub (gimme-sub p1))
	     (p2sub (if (equal p2 'part) (list 'part) (gimme-sub p2)))
	     res newp)
	   (dolist (p p1sub)
		   (setq newp (build-symbol (list p '- p2) 'epilog))
		   (unless (getp newp 'pred-parts)
			   (setp newp 'specialists (getp p2 'specialists))
			   (setp newp 'indicators (getp p2 'indicators))
			   (setp newp 'pred-parts (list p p2)))
		   (setq res (append res (list newp))) )
	   (dolist (p p2sub)
		   (setq newp (build-symbol (list p1 '- p) 'epilog))
		   (unless (getp newp 'pred-parts)
			   (setp newp 'specialists (getp p 'specialists))
			   (setp newp 'indicators (getp p 'indicators))
			   (setp newp 'pred-parts (list p1 p)))
		   (setq res (append res (list newp))) )
	   res)
      (get-all-sub item))
  )

;also check for compound predicates here!
(defun get-super-class (class &optional mp already &aux 
			      (concept (car class))
			      (topic (cadr class))
			      (role (caddr class))
			      phiers t2 tlist c res s)
			  
  (if (and mp (equal concept 'wff) (null topic)) 
      (return-from get-super-class nil))
  ; climb up concept part of classification
  (if (type-pred-p concept)
      (setq tlist (append (gimme-super concept) 
			  (if (and mp (equal concept 'quantifier)) (list 'wff))
			  (if (and mp (equal concept 'function)) (list 'term))
			  (get-super-parts concept)
	      (if (setq phiers (remove-if #'(lambda (a) (not (getp a 'part-hier))) (getp concept 'hier)))
		  (remove-if 'null (mapcar #'(lambda (b) 
				    (let ((newp (build-symbol (list b '-part) 'epilog)))
					 (setp newp 'specialists (getp b 'specialists))
					 (setp newp 'indicators (getp b 'indicators))
					 (setp newp 'pred-parts (list b 'part))
					 newp))
			  (remove-if 'null (mapcar #'(lambda (a) (get-super a concept)) phiers))
		  )))) )
      (setq tlist (append (grab-types concept)
			  ; should get sort of fn name if there
			  (if (and (symbolp concept)
				   (setq s (find-sort concept)))
			      (list s) nil))))
  (dolist (typ tlist)
	  (setq c (if role (list typ topic role) (list typ topic)))
	  (unless (member c already :test 'equal)
		  (if (and mp (equal c '(entity entity))) (setq c '(wff)))
		  (setq res (append res (list c)))
		  ) )
  (when mp
	(if (type-pred-p topic)
	    (setq t2 (gimme-super topic)) 
	    (if (and (member topic *available-specialists*)
		     (not (equal topic 'meta-specialist)))
		(setq t2 (list 'pred))
		(setq t2 (append (grab-types topic)
			  (if (and (symbolp topic)
				   (setq s (find-sort topic)))
			      (list s) nil)))))
	(dolist (typ t2)
		(setq c (list concept typ))
		(unless (member c already :test 'equal)
			(setq res (append res (list c)))
			) )
	)
                
  (remove-duplicates res :test 'equal)
  )
 
(defun get-subnet-class (prev class &optional already (super t) (ok t) &aux 
			      result p1 nc tp prf lp)
  (cond ((null (cdr class)) ; this is the last part
	 (setq lp (if super (get-super-class (car class))
                        (get-sub-class (car class))))
	 (if (and (null lp) super) (setq lp '((propos _ object))))
         (dolist (c lp) 
		 (setq nc (if prev (combine-class prev c) c) )
		 (unless (member nc already :test 'equal)
			 (setq result (cons nc result)))
                 )
         result)
        (t (when ok (if (check-subnet
               (setq tp (if prev (combine-class prev (car class)) (car class))))
               (setq result (append result (get-subnet-class tp (cdr class)
                                            (append result already) super)))))
	   (setq p1 (if super (get-super-class (car class))
                        (get-sub-class (car class))))
           (dolist (p p1)
                   (setq prf (if prev (combine-class prev p) p))
                   (if (check-subnet prf)
		       (let ()
                       (setq result (union result
                                      (list (combine-class prf (cdr class) t))
				      :test 'equal))
                       (setq result (append result
				(get-subnet-class prf (cdr class)
                                              (append result already) super))))
		       ; nothing there, try with super of that one
                       (setq result (append result
                          (get-subnet-class prev (combine-class p (cdr class) t)
					    (append result already) super nil)))
                       ))
           result)
        )
  )

(defun combine-class (c1 c2 &optional (c2list nil))
  (if (listp (car c1))
      (if c2list
          (append c1 c2)
          (append c1 (list c2)))
      (if c2list
          (append (list c1) c2)
          (list c1 c2)))
  )
 
(defun check-subnet (class)
  (if (get-class class)
      t nil))
 
(defun sub-class (class &optional already)
  (if (listp (car class)) 
      (remove-duplicates (get-subnet-class nil class already nil) :test 'equal)
      (remove-duplicates (get-sub-class class already) :test 'equal))
  )
 
(defun get-sub-class (class &optional already &aux 
			    (concept (car class))
			    (topic (cadr class))
			    (role (caddr class))
			    c res)
  ; climb down concept part of classification
  (when (or (type-pred-p concept) ;or set
	    (equal (find-sort concept) 'set)
	    (and (listp concept)
		 (or (member (car concept) *quoted-indicators*)
		     (equal (car concept) 'set)
		     (equal (car concept) 'set-of))))
	(dolist (typ (append (if (symbolp concept)
				 (append (gimme-sub concept) 
					 (get-sub-parts concept)) )
			     (instances concept)))
		(setq c (if role (list typ topic role) (list typ topic)))
		(unless (member c already :test 'equal)
			(setq res (append res (list c)))
			) ))
  (remove-duplicates res :test 'equal)
  )

; should consider part-type-hier too?
(defun get-sub-parts (item &aux subs pp)
  (when (or (member-if #'(lambda (a) (and (getp a 'part-hier) 
					  (not (equal item (getp a 'root)))))
				 (getp item 'hier))
	    (and (setq pp (getp item 'pred-parts))
		 (equal (cadr pp) 'part)))
	(if (null pp) (setq pp (list item 'part)))
	(dolist (h (remove-if #'(lambda (a) (or (equal a 'part) 
						(not (getp a 'part-hier)) ))
				  (getp (car pp) 'hier)))
		(setq subs (append subs (get-sub h (car pp))))
		))
  subs)

(defun get-super-parts (item &aux newp super pp sup)
  (when (or (member-if #'(lambda (a) (and (getp a 'part-hier) 
					  (not (equal item (getp a 'root)))))
				 (getp item 'hier))
	    (and (setq pp (getp item 'pred-parts))
		 (equal (cadr pp) 'part)) )
	(if (null pp) (setq pp (list item 'part)))
	(dolist (h (remove-if #'(lambda (a) (or (equal a 'part)
						(not (getp a 'part-hier)) ))
				  (getp (car pp) 'hier)))
		(setq sup (get-super h (car pp)))
		(when sup 
		      (setq newp (build-symbol (list sup '-part) 'epilog))
		      (setp newp 'specialists (getp (car pp) 'specialists))
		      (setp newp 'indicators (getp (car pp) 'indicators))
		      (setp newp 'pred-parts (list sup 'part))
		      (setq super (append super (list newp))) )
		))
  super)


(defun make-class (c1 c2)
  (if (listp (car c2))
      (if (equal c1 (car c2))
	  c2
	  (cons c1 c2))
      (cons c1 (list c2))
      ) 
  )


(defun instances (item &aux insts)
  (cond ((type-pred-p item)
	 (dolist (w (get-class (list item)))
		 (when (type-lit-p w)
		       (if (equal (getp w 'wff-type) 'episodic)
			   (push (car (getp (car (getp w 'arguments)) 
					    'arguments)) insts)
			   (push (car (getp w 'arguments)) insts))
		       ))
	 (union insts (apply 'append (mapcar #'(lambda (a) (equivalent-items a))
					     insts)))
	 )
	; also handle sets etc
	)
  )


;
; Keys
;

 ;**********************************************************
 ;*                                                        *
 ;*   F I N D I N G   K E Y   P R E D I C A T E S  A N D   *
 ;*                                                        *
 ;*       K E Y   L I T E R A L S   O F   R U L E S        *
 ;*                                                        *
 ;**********************************************************

(defun save-types (wff &aux 
		       (phi-list (append (getp wff '+lits) (getp wff '-lits))))
  (mapcar 'save-type-info phi-list)
  (mapcar 'save-prop-info phi-list)
  )

(defun save-type-info (item &aux m)
  (cond ((null item) nil)
	((epi-variable item) 
	 (pushnew item *variables*)
	 (if (getp item 'sort) (pushnew item *typed-variables*) ))
	((not (symbolp item)) nil)
	((and (equal (getp item 'entity-type) 'wff)
	      (equal (getp item 'wff-type) 'quasi-quoted-expression))
	 (save-type-info (getp item 'quote-wff)))
	((quasi-quoted-expression-p item)
	 (let ((real-item (cadr (getp item 'print))))
	      (if (listp real-item)
		  (mapcar 'save-type-info real-item)
		  (save-type-info real-item)
		  )))
	(t
	 (if (type-lit-p item) (unless (neg-p item) (save-type item)))
	 (if (and (member (getp item 'pred) '(member-of member-of-0))
		  (not (neg-p item)))
	     (save-set-info item))
	 (cond ((setq m (getp item 'main) )
		  (save-type-info m)
		  (save-type-info (getp item 'restrict)))
		 (t (save-type-info (getp item 'pred))
		    (save-type-info (getp item 'operator))
		    (mapcar 'save-type-info (getp item 'arguments)))
		 )
	   ))
  )

(defun save-prop-info (item &aux m)
  (cond ((null item) nil)
	((propos-lit item) 
	 (class-wff item))
	((not (symbolp item)) nil)
	(t (cond ((setq m (getp item 'main) )
		  (save-prop-info m)
		  (save-prop-info (getp item 'restrict)))
		 (t (save-prop-info (getp item 'pred))
		    (mapcar 'save-prop-info (getp item 'arguments)))
		 )
	   ))
)


(defun check-arg-sorts (arg)
  (if (symbolp arg)
      (if (and (epi-variable arg) (getp arg 'sort))
	  (pushnew arg *typed-variables*)
	  (if (getp arg 'main)
	      (let () (check-arg-sorts (getp arg 'restrict))
		   (check-arg-sorts (getp arg 'main)))
	      (mapcar 'check-arg-sorts (getp arg 'arguments)) )
	  )
      )
  )

(defun propos-lit (item)
  (and (symbolp item)
       (equal (getp item 'wff-type) 'episodic) 
       (equal (getp item 'pred) '_))
  )

(defun save-type (phi)
  (case (getp phi 'wff-type)
	(prefix 
	 (let ((arg (car (getp phi 'arguments)))
	       (pred (getp phi 'pred)))
	      (when (and (not (epi-variable arg)) (symbolp arg) (symbolp pred)
			 (not (getp arg 'sort)))
		    (setp arg 'sort (getp pred 'sort)))
	      (when (epi-variable arg)
		    (setp arg 'types (cons pred (getp arg 'types)))
		    (push arg *typed-variables*))
	      ))
	(episodic
	 (save-type (car (getp phi 'arguments))))
	)
  )

(defun save-set-info (item &aux
			   (args (getp item 'arguments))
			   (a1 (car args))
			   (a2 (cadr args)))
  (if (epi-variable a1)
      (if (symbolp a2) 
	  (case (getp a2 'term-type) 
		(record 
		 (pushnew a1 *typed-variables*)
		 (setp a1 'types (append (getp a1 'types) 
			 (remove-if 'epi-variable (record-contents a2)))))
		(function
		 (when (equal (getp a2 'operator) 'set-of)
		       (pushnew a1 *typed-variables*)
		       (setp a1 'types (append (getp a1 'types)
			 (remove-if 'epi-variable (getp a2 'arguments))))) )
		((variable quoted-expresssion quasi-quoted-expression wff) nil)
		(otherwise 
		 (pushnew a1 *typed-variables*)
		 (setp a1 'types (append (getp a1 'types) (list a2))))
		)))
  )


(defun keys (phi &aux
		 (+lits (getp phi '+lits))
		 (-lits (getp phi '-lits))
		 nlits+ nlits- -ntype +ntype)
  (when (or +lits -lits)
	(cond (*meaning-postulate* 
	       (setp phi '+keys +lits)
	       (setp phi '-keys -lits)
	       (setp phi '+trigger +lits)
	       (setp phi '-trigger -lits))
	      (t
	       (when (traced-p 'keys) 
		     (print-line "  Selecting Keys for " phi ": " 
				 (getp phi 'print)))

        (flag-em -lits +lits)
	(setq nlits- (append (key-literals -lits) -ntype))
	(when (and (traced-p 'keys) (getp phi 'conditional))
	      (print-line "  Negative Keys :" (mapcar 'print-info nlits-) ))
	(setp phi '-keys nlits-)
	(setp phi '-other (set-difference -lits nlits-)) 
	(setp phi '-trigger (union (trigger-keys nlits-) -ntype))
	(when (traced-p 'keys) 
	      (print-line "  Negative Trigger Keys: " 
			  (mapcar 'print-info (getp phi '-trigger)) ))

	(setq nlits+ (append (key-literals +lits) +ntype))
	(when (and (traced-p 'keys) (getp phi 'conditional))
	      (print-line "  Positive Keys: " (mapcar 'print-info nlits+) ))
	(setp phi '+keys nlits+)
	(setp phi '+other (set-difference +lits nlits+)) 
	(setp phi '+trigger (union (trigger-keys nlits+) +ntype))
	(when (traced-p 'keys) 
	      (print-line "  Positive Trigger Keys: " 
			  (mapcar 'print-info (getp phi '+trigger)) ))
	))

	(remprop phi '-lits) 
	(remprop phi '+lits) 
	)
  )

(defun trigger-keys (result &aux keys)
  (if (= (length result) 1) 
      (setq keys result)

  (if (and *qa-mode* (> *key-threshold* 0) 
	   (member-if #'(lambda (a) (member (getp a 'pred) 
				    '(member-of member-of-0 part-of))) result))
      (setq keys result)
      ; take keys with highest interest, or greater than key threshold
      (let (int tkeys bestint)
	   (dolist (k result)
		   (setq int (wff-interest k))
		   (when (traced-p 'keys)
			 (print-line "    Interest of key " (print-info k) 
				     " is " int ))
		   (if (>= int *key-threshold*) (push k tkeys))
		   (cond ((extended-type-lit-p k) (setq keys (append keys (list k))))
			 ((or (null bestint) (> int bestint))
			  (setq bestint int)
			  (setq keys (list k)))
			 ((= int bestint)
			  (setq keys (append keys (list k))))
			 ))
	   (setq keys (union keys tkeys :test 'equal)))
      ) )
  keys)

(defun flag-em (phi-list alt-list)
  (mapc 'flag-free-vars phi-list)
  (mapc 'flag-free-vars alt-list)
  (mapc 'flag-free1 phi-list)
  (mapc 'flag-free1 alt-list)
  (mapc 'flag-free2 phi-list)
  (mapc 'flag-free2 alt-list)
  )

(defun key-literals (phi-list &aux result)
  ;;
  ;; For the given list of "literals", select those to be used as
  ;; "key literals" for matching input facts in forward-chaining.
  ;; Do this by eliminating all type-predications (including,
  ;; perhaps, episodic ones such as ((X FROG)**E2)) whose free
  ;; variables occur in non-type predications, and then eliminating
  ;; all non-episodic predications whose free variables occur in
  ;; episodic predications. (Flags (FREE 0), (FREE 1) and (FREE 2)
  ;; are used on property lists of variables to distinguish variables
  ;; that occur in type predications only, those that occur in non-
  ;; type predications but not in episodic predications, and those
  ;; that occur in episodic non-type predications respectively.)
  ;;
  (setq result (cond ((delete-extra-type-lit-ps phi-list)) (t phi-list)))
  (setq result (cond ((delete-extra-nonepisodic-lits result))
	(t result))) 
  result)

(defun would-be-key (lit &aux res)
  (setq res (free 3 lit))
  res)

(defun flag-free-vars (phi)
  ;;
  ;; Flag those variables of PHI which are not quantified within
  ;; PHI as (FREE 0). (Variables are assumed already to have
  ;; property VARIABLE.) 
  ;;
  (mapcar #'(lambda (a) (flag-term a 0)) (getp phi 'arguments))
  )


 (DEFUN MAPALL (PRED FUNC ARG)
  ;;
  ;; Apply FUNC to all subtrees of binary tree ARG (in left-to-right,
  ;; top-down manner) which satisfy PRED.
  ;;
 (IF (FUNCALL PRED ARG) (FUNCALL FUNC ARG))
 (UNLESS (ATOM ARG)
     (MAPALL PRED FUNC (CAR ARG))
     (MAPALL PRED FUNC (CDR ARG)) ))



(defun flag-free1 (phi)
  ;;
  ;; Flag those variables of PHI as (FREE 1) which occur in non-type
  ;; predications within PHI.
  ;;
  (if (not (type-pred-p (getp phi 'pred)))
      (mapcar #'(lambda (a)  (flag-term a 1))
	      (getp phi 'arguments))
      )
  )

(defun flag-term (item free)
  (if (epi-variable item) 
      (setp item 'free free)
      (if (symbolp item)
	  (mapcar #'(lambda (a) (flag-term a free)) (getp item 'arguments))))
  )


(defun flag-free2 (phi)
  ;;
  ;; Flag those FREE variables of PHI as (FREE 2) which occur in
  ;; episodic non-type predications within PHI.
  ;;
  (when (episodic-lit phi) (flag-free3 phi))
  )

(defun episodic-lit (phi)
  ;;
  ;; PHI is episodic if it is not a type literal and contains **, *, or @.
  ;;
  (and (symbolp phi)
       (equal (getp phi 'wff-type) 'episodic)
       (not (equal (getp phi 'pred) '_)))
  )

(defun type-lit-p (phi &aux pred)
  ;;
  ;; PHI is a type literal if it is an atomic type predication
  ;; or is made up entirely of type predications (possibly inside
  ;; **, *, or @ constructs)
  ;;
  (if (symbolp phi)
  (case (getp phi 'wff-type)
	(prefix (type-pred-p (setq pred (getp phi 'pred))))
	(episodic 
	 (unless (equal pred '_) (type-lit-p (car (getp phi 'arguments)))))
	))
  )
(defun extended-type-lit-p (phi &aux pred)
  (if (symbolp phi)
  (case (getp phi 'wff-type)
	(prefix (or (type-pred-p (setq pred (getp phi 'pred)))
		    (if (> *key-threshold* 0) 
			(or (member pred '(member-of member-of-0 part-of))
			    ; to handle coll-of type
			(and (getp pred 'operator)
			     (type-pred-p (car (getp pred 'arguments))))
			))))
	(episodic 
	 (unless (equal pred '_) (extended-type-lit-p (car (getp phi 'arguments)))))
	))
  )

(defun flag-free3 (phi)
  ;;
  ;; The input PHI is assumed to be an episodic wff. Mark all FREE
  ;; variables within (CAR PHI) and (CADDR PHI) as (FREE 2).
  ;;
  (case (getp phi 'wff-type)
	(episodic 
	 (unless (equal (getp phi 'pred) '_)
	 (let ((args (getp phi 'arguments)))
	      (flag-term (last1 args) 2)
	      (flag-free3 (car args)))) )
	(modified (mapcar 'flag-free3 (getp phi 'arguments)))
	(prefix
	 (mapcar #'(lambda (a) (flag-term a 2)) (getp phi 'arguments)))
	)
  )


(defun delete-extra-type-lit-ps (phi-list)
  ;;
  ;; Delete type-literals from PHI-LIST all of whose FREE variables
  ;; are (FREE 1)
  ;;
(filter #'(lambda (x) (or (not (free 1 x)) (not (extended-type-lit-p x)))) phi-list) )



(defun free-arg (n arg)
  (if (and (symbolp arg) (getp arg 'free))
      (equal (getp arg 'free) n)
      (if (and (symbolp arg) (getp arg 'arguments))
	  (eval (append (list 'and) (mapcar #'(lambda (a) (free-arg n a))
					    (getp arg 'arguments))))
	  t
	  ;(if (epi-variable arg) t (if (> n 1) nil t))
	  ))
  )

(defun free (n phi)
  ;;
  ;; Are all FREE variables in PHI marked (FREE N)?
  ;;
  (case (getp phi 'wff-type)
	(prefix (eval (append (list 'and) 
		       (mapcar #'(lambda (a) (free-arg n a))
			       (getp phi 'arguments)))))
	(modified 
	 (eval (append (list 'and) 
		     (mapcar #'(lambda (a) (free n a)) (getp phi 'arguments)))))
	(episodic 
	 (let ((args (getp phi 'arguments)))
	      (and (if (and (symbolp (last1 args)) (getp (last1 args) 'free))
		       (equal (getp (last1 args) 'free) n) t)
		   (free n (car args)))))
	)
  )

(defun test-free-term (term free)
  (cond ((symbolp term)
	 (if (getp term 'free) 
	     (equal (getp term 'free) free)
	     (eval (append (list 'and) (mapcar #'(lambda (a) 
			(test-free-term a free)) (getp term 'arguments))))) )
	(t nil))
  )


(defun useful-nonepisodic-lit (phi)
  ;;
  ;; PHI is a useful nonepisodic if it is not about time
  ;; cause-of and occurred-in are ok
  ;;
  (case (getp phi 'wff-type)
	(prefix 
	 (find-if #'(lambda (i) 
			    (member (if (listp i) (car i) i)
				    *useful-nonepisodic-topics*))
		  (getp (getp phi 'pred) 'INDICATORS))) 
	)
  )


(defun delete-extra-nonepisodic-lits (phi-list)
  ;;
  ;; Delete non-episodic wffs from PHI-LIST all of whose FREE
  ;; variables are (FREE 2)
  ;;
 (filter #'(lambda (x) (or (not (free 2 x)) 
			   (episodic-lit x)
			   (useful-nonepisodic-lit x)
			   ))
         phi-list))


; add-flags adds super and sub flags for classifications for goal directed
; reasoning

(defun add-flags (cl)
  (mapcar #'(lambda (class)
          (cond ((type-pred-p (if (listp (car class)) (caar class) (car class)))
                 ; already a type pred - add -sub and -super flags
                 (cons class (list '-super '-sub)) )
                (t (cons class (list '-super)) )))
          cl)
  )

; 
; Determination of mp classifications
;

(defun mp-classify (wff &optional topic &aux classes pred)
  (if (and (not *meaning-postulate*) (not (getp wff 'meaning-postulate)))
      (return-from mp-classify (remove-duplicates (mp-class-wff wff) 
						  :test 'equal)))
	      
  (setq pred (getp wff 'pred))
  ; get combined classifications (use keys only!)
  (dolist (k (getp wff '-keys))
	  (setq classes (append classes (mp-class-wff k topic))) )
  (dolist (k (getp wff '+keys))
	  (setq classes (append classes (mp-class-wff k topic))) )
  (remprop wff '-lits)
  (remprop wff '+lits)
  (unless (or (getp wff '-keys) (getp wff '+keys))
	  (if (equal (getp wff 'wff-type) 'quantified)
	      (let* ((r (getp wff 'restrict))
		     (cr (if r (mp-class-wff r topic)))
		     (mr (mp-class-wff (getp wff 'main) topic)))
		    (setq classes (append (append classes cr) mr))
		    )
	      (if (getp wff 'conditional)
		  (apply 'append (mapcar #'(lambda (a) 
			   (mp-class-wff a topic)) (getp wff 'arguments)))
		  (if topic
		      (setq classes (list (list pred topic)))
		      (dolist (cl (mapcar #'(lambda (a) 
						    (mp-class-wff a pred)) 
					 (getp wff 'arguments)))
			      (setq classes (append classes cl))
			      )) ) ) )
  (remove-duplicates classes :test 'equal))

(defun mp-class-wff (wff &optional topic &aux piece cl mc)
  (case (getp wff 'wff-type)

	; the first few have literals associated with them
	; just classify the literals and use that?
	(quantified
	 (cond ((equal (getp wff 'quantifier) 'L)
		(if (getp (setq mc (getp wff 'main)) 'pred)
		    (mp-class-pred (getp mc 'pred) topic)
		    (mp-class-wff mc topic))
		)
	       (t (setq piece (getp wff 'quantifier))
		  (if (epi-variable piece) 
		      (setq piece (cond ((getp piece 'sort)) (t 'quantifier))))
		  (if topic 
		      (final-mp-class piece topic)
		      (append (if (getp wff 'restrict)
				  (mp-class-wff (getp wff 'restrict) piece))
			      (mp-class-wff (getp wff 'main) piece)))
		  ) ) 
	 )
	(logical
	 (if topic 
	     (final-mp-class (getp wff 'pred) topic)
	     (let (cl)
		  (dolist (a (getp wff 'arguments))
			  (setq cl (append cl (mp-class-wff a (getp wff 'pred))))
			  )
		  cl))
	 )

	(modified
	 (mp-class-op (getp wff 'operator) topic (getp wff 'arguments))
	 )

	(episodic
	 (if topic
	     (append (final-mp-class (getp wff 'pred) topic) 
		     (unless *meaning-postulate* (final-mp-class 'wff topic))) 
	     (let* ((pred (getp wff 'pred))
		    (args (getp wff 'arguments))
		    (classes (mp-class-wff (car args) pred)))
		   (unless (and classes *meaning-postulate*)
			   (setq classes (append classes
						 (apply 'append
				 (mapcar #'(lambda (a) (mp-class-item a pred))
					 (cdr args)))) ))
		   classes)
	     )
	 )
	((prefix causal)
	 (cond ((equal (getp wff 'pred) 'true)
		(setq cl (mp-class-wff (car (getp wff 'arguments)) topic))
		(if cl cl (list (list 'meta-entity 'meta-entity)))
		)
	       (t (append
		   (mp-class-pred (getp wff 'pred) topic (getp wff 'arguments))
		   (unless (or (null topic) *meaning-postulate*)
			   (final-mp-class 'wff topic) 
			   )) )
	       ))

	(quasi-quoted-expression
	 (mp-class-wff (getp wff 'quote-wff) topic))

	(quoted-expression
	 (if topic (final-mp-class 'wff topic)))

	(otherwise ; constant variable otherwise 
	 (final-mp-class 'wff topic))
	)
  )

; returns a list of items which can be used in the terms position of 
; classification

(defun mp-class-term (term topic &aux s)
  (cond ((symbolp term)
	 (case (getp term 'term-type)
	       ((variable constant)
		(if (member (setq s (find-sort term)) 
			    '(meta-entity pred operator quantifier wff term))
		    (final-mp-class s topic)
		    (if topic (final-mp-class 'term topic))))
	       ((modified function)
		(mp-class-op (getp term 'operator) topic 
			     (getp term 'arguments)))
	       (wff (mp-class-wff term topic))
	       (otherwise
		(when topic (final-mp-class 'term topic) )
		)
	       ))
	((setq s (find-sort term))
	 (if topic (final-mp-class s topic))) 
	(t nil); if string or number, don't classify under it
	)
  )

(defun mp-class-quant (quant topic &aux s)
  (cond ((symbolp quant)
	 (when topic
	       (if (setq s (find-sort quant))
		   (final-mp-class s quant) 
		   (final-mp-class 'term quant) 
		   ))
	 )
	(t nil)
	)
  )

(defun mp-class-pred (pred topic &optional args)
  (cond ((symbolp pred)
	 (case (getp pred 'pred-type)
	       (variable 
		(if topic
		    (final-mp-class 'pred topic) 
		    (apply 'append 
			 (mapcar #'(lambda (a) (mp-class-item a 'pred)) args))
		    ))
	       (lambda 
		(if (getp pred 'wff-type)
		    (let ((larg *lambda-arg*)
			  classes)
			 (unless *lambda-arg*
				 (setq *lambda-arg* (getp pred 'variable)))
			 (setq classes (mp-class-wff pred topic))
			 (setq *lambda-arg* larg)
			 classes
			 )
		    (mp-class-pred (getp pred 'operator) topic args)
		    )
		)
	       (modified 
		(mp-class-op (getp pred 'operator) topic (getp pred 'arguments))
		)
	       (otherwise ; constant  
		(if topic
		    (append (final-mp-class pred topic) 
			 (unless *meaning-postulate* 
				 (final-mp-class 'pred topic))) 
		    (append (apply 'append 
			 (mapcar #'(lambda (a) (mp-class-item a pred)) args))
			    (unless *meaning-postulate* (apply 'append
			   (mapcar #'(lambda (a) (mp-class-item a 'pred)) args))
				))
		    ))
	       ))
	; numbers in any case other than implications are not legal and 
	; will not be classified; similarly for strings, etc
	(t nil))
  )

(defun mp-class-op (op topic &optional args)
  (cond ((symbolp op)
	 (case (getp op 'op-type)
	       (variable 
		(if topic 
		    (final-mp-class 'operator topic)
		    (apply 'append 
			 (mapcar #'(lambda (a) (mp-class-item a 'operator)) args))
		    ))
	       (modified 
		(mp-class-op (getp op 'operator) topic (getp op 'arguments))
		)
	       (otherwise ; constant 
		(if topic
		    (append (final-mp-class op topic) 
			 (unless *meaning-postulate* 
				 (final-mp-class 'operator topic))) 
		    (append (apply 'append 
			 (mapcar #'(lambda (a) (mp-class-item a op)) args))
			    (unless *meaning-postulate* (apply 'append
			   (mapcar #'(lambda (a) (mp-class-item a 'operator)) args))
				))
		    ))
	       ))
	((numberp op)
	 (if topic
	     (final-mp-class 'operator topic)
	     nil))
	(t nil))
  )

(defun mp-class-item (item topic &aux (usetype (if (symbolp item)
					  (cond ((entity-type item)) (t 'term))
					  nil)))
  (when usetype
	(case usetype 
	      (wff (mp-class-wff item topic))
	      (pred (mp-class-pred item topic))
	      (op (mp-class-op item topic))
	      (term (mp-class-term item topic))
	      (quantifier (mp-class-quant item topic))
	      (otherwise (mp-class-term item topic))
	      ))
  )

(defun final-mp-class (item topic &aux
			    (specs (if (and (symbolp topic)
					    (not (member topic 
					 '(meta-entity pred operator quantifier term wff))))
				       (getp topic 'specialists)))
			    )
  (if topic 
      (setq topic
	    (cond ((member topic '(meta-entity pred operator 
					       quantifier term wff)) topic)
		  ((not (symbolp topic)) topic)
		  ((member 'meta-specialist specs) 
		   (append specs (list 'meta-entity)))
		  (specs specs)
		  (t topic) 
		  )))
  (cond (topic
	 (if (listp topic) (mapcar #'(lambda (a) (list item a)) topic)
	     (list (list item topic))))
	(t (list (list item))))
  )

