;       Last change    August 1993
(in-package epilib)

; Normalization routines
; Normalization now combines syntax checking, normalization, sort removal
; and replacing variables by their skolem constants all in one step instead
; of several passes. The input is a list which is the formula; the result will
; be a single atom in which each subpart has been normalized as well.

;  Normalized entities are all atoms, and have an entity-type property
;  associated with them of: wff, pred, op, or term

(tweakable '*name-symbol* '! 
  "Symbol which indicates that the next item is a name for the previous structure"
  'symbol)
(tweakable '*expand-names* t 
  "Indicates whether or not to expand named objects when printing" 'boolean)
(tweakable '*stop-if-error* t 
  "Indicates whether to pause if a syntax error is found" 'boolean)
(tweakable '*split-episodic* t 
  "Indicates whether to automatically split top level episodic formulas with and within"
  'boolean)

(traceable 'lambda "Shows what happens to lambda expressions")
(traceable 'normalize "Traces normalization ")

(defvar *simplification-schemas* nil)
(defvar *interim* nil "Indicates whether in midst of forward infernce")
(defvar *normed-prob* 1)
(defvar *normed-subs* nil)
(defvar *skol-subs* nil)
(defvar *wh-vars* nil)
(defvar *e-vars* nil)
(defvar *new-vars* t)
(defvar *got-sort* nil)
(defvar *quoted* nil)
(defvar *cur-var*)
(defvar *cur-varnum*)
(defvar *vars-to-fix* nil)
(defvar *vars-to-restore* nil)
(defvar *+lits* nil)
(defvar *-lits* nil)
(defvar *cur-var-matchable* nil)
(defvar *cur-var-nonmatchable* nil)
(defvar *cur-varnum-matchable* 0)
(defvar *cur-varnum-nonmatchable* 0)

; pre-norm takes the input from the user, which may have some different
; symbols than system generated input

(defun pre-norm (phi)
  (setq phi (subst 'not '\~ phi))
  (setq phi (subst 'implies '=> phi))
  (setq phi (subst 'implies '-> phi))
  (setq phi (subst 'less-than '\< phi))
  (setq phi (subst 'lt= '<= phi))
  (setq phi (subst 'greater-than '\> phi))
  (setq phi (subst 'gt= '>= phi))
  (setq phi (subst 'equal '\= phi))
  (setq phi (subst 'or '\| phi))
  (setq phi (subst 'and '\& phi))
  (setq phi (subst 'qquote 'qq phi))
  phi)

; normalize returns a wff name
(defun normalize (phi &optional (context '+) (insertion nil) (new-vars t) &aux 
		      (oi *insertion*)
		      (ophi phi) phi-name p)
  ; this will be done before getting here - but is here temp for testing
  (setq *insertion* (or *insertion* insertion))
  (setq *normed-prob* 1)
  (setq *normed-subs* nil)
  (setq phi (pre-norm phi))
  (setq phi (distr-neg phi))
  (setq phi (collapse-connectives phi))
  (unless (or (member (setq p (prop-pred phi)) '(implies ~> <=>)) (numberp p))
	  (setq phi (distr-ops phi)) )

  ; initialize
  (setq *quoted* nil)
  (setq *+lits* nil)
  (setq *-lits* nil)
  (initialize-standard-vars)
  (setq *subs* nil)
  (setq *vars-to-fix* nil)
  (setq *vars-to-restore* nil)
  (setq *wh-vars* nil)
  (setq *e-vars* nil)
  (setq *new-vars* new-vars)

  ; normalize
  (catch 'normalize-catcher (setq phi-name (norm-wff phi context)))
  (when (and phi-name *simplification-schema* 
	     (not (getp phi-name 'equivalence)))
	(print-line "ERROR: Normalization: Substitution meaning postulate " 
		    ophi " must be an equivalence - <=>")
	(setq phi-name nil))

  (when (and (traced-p 'normalize) (not (member ophi '(YES NO inconsistent))))
	(print-line "Normalized: " ophi " to")
	(if phi-name
	    (print-line "        " phi-name ": " 
			(cond ((getp phi-name 'print)) (t phi-name))) 
	    (print-it "  --> Error "))
	)
  (when (null phi-name)
	(print-line phi " did NOT normalize, so will NOT be entered or used")
	(when *stop-if-error*
	      (terpri)(princ "WFF FORMAT ERROR: PRESS RETURN TO CONTINUE")
	      (read-line))
	)

  ; may end up with something skolemizable
  (unless (member phi-name '(YES NO INCONSISTENT))
  (when (and phi-name (null (getp phi-name '+lits)))
	(setp phi-name '+LITS (check-lits phi-name *+lits*) )
	(setp phi-name '-LITS (check-lits phi-name *-lits*))
	)
  (when (and phi-name *wh-vars*) (setp phi-name 'wh-vars *wh-vars*))
  (when (and phi-name *e-vars*) (setp phi-name 'e-vars *e-vars*)) )

  ; cleanup
  (dolist (v *vars-to-fix*) (unless (getp v 'main-sort) (remprop v 'sort)))
  (setq *insertion* oi)
  (setq *new-vars* t)

  ; if null is returned it means there was a fatal error somewhere
  phi-name)

; last minute check on lits - some may have been removed because of 
; evaluations
(defun check-lits (phi-name lits)
  (cond ((null lits) nil)
	(t (remove-if 'null 
		      (mapcar #'(lambda (a) (check-lit phi-name a)) lits))))
  )

(defun check-lit (phi-name lit &aux res)
  (when (and phi-name (symbolp phi-name))
  (if (equal phi-name lit) (return-from check-lit lit))
  (case (getp phi-name 'wff-type)
	(quantified (cond ((check-lit (getp phi-name 'main) lit))
			  (t (check-lit (getp phi-name 'restrict) lit) )))
	((logical causal) 
	 (dolist (a (getp phi-name 'arguments))
		 (setq res (check-lit a lit))
		 (if res (return-from check-lit res))) )
	((episodic modified) (check-lit (car (getp phi-name 'arguments)) lit))
	; others should be handled by earlier equal
	)
	)
  )

; if a name is found, a list consisting of the name, and the revised version
; of item with the name symbol and name removed are returned; otherwise nil
; optionally the entitty type may be put on 
; if the name already exists as a different! type of entity - yell

(defun name-scan (item  &optional ttype &aux pos name newitem)
  (when (listp item)
	(setq pos (position *name-symbol* item))
	(when pos
	      (setq name (car (subseq item (+ pos 1) (+ pos 2))))
	      (setq newitem (subseq item 0 pos))
	      (setq newitem (append newitem (subseq item (+ pos 2))))
	      (when (symbolp name) (setq name (remove-sort name)))
	      (when (and (symbolp name) (epi-variable name))
		    (let ((new (new-symbol 'name nil)))
			 (setp name 'alternate-names (list new))
			 (setq name new)))

	      (cond ((not (symbolp name)) 
		     ; fatal error - stop immediately
		     (fatal-error 'normalize-catcher
			"Normalization: " name " is not a"
			      " legal name to use after " *name-symbol*))
		    ((getp name 'entity-type)
		     ; fatal error - stop immediately
		     (fatal-error 'normalize-catcher
			"Normalization: Attempted to rename"
			      " previously named entity " name " of type " 
			      (getp name 'entity-type)))
		    (t (if ttype (change-property name 'entity-type ttype))
		       (change-property name 'explicit t)
		       (list name newitem))
		    )
	      )
	)
  )

; for keys and temporary normalizations
(defun normalize-wff (phi &optional (context '+) &aux phi-name p
			  (oi *insertion*)
			  (nv *new-vars*))
  (setq *new-vars* nil)
  (setq *subs* nil)
  (setq *insertion* nil)
  (setq phi (distr-neg phi))
  (setq phi (collapse-connectives phi))
  (unless (or (member (setq p (prop-pred phi)) '(implies ~> <=>)) (numberp p))
	  (setq phi (distr-ops phi)) )
  (catch 'normalize-catcher (setq phi-name (norm-wff phi context)))
  (setq *insertion* oi)
  (setq *new-vars* nv)
  phi-name)

; wff consists of:
;   (wff-op wff)
;   {negation} quantifier variable {rest} main
;              term pred term*
;              wff logical-conn wff+
;              wff episodic-conn term
;              wff causal-conn wff
; By the time it gets here, the negation, if present, may have been distributed
; inside - this is true for quantified expressions, as well as those involving
; logical connectives.
; Normalized wff has the following properties:
;  wff-type: quantified, logical, modified, episodic, causal, prefix, constant
;            variable
;  negated: true if negated, nil otherwise
;  quantifier, variable, restrict, main for quantified expressions
;  operator, arguments for modified expressions
;  pred, arguments for logical, episodic, causal and prefix
;  conditional: t if conditional for quantified or logical expressions

(defun norm-wff (wff &optional (context '+) (ttype 'wff) embed &aux 
							       (alt-context (if (equal context '+) '- '+))
							       pred args namestuff newname rr conv
							       quantifier variable main restrict predicate arguments
							       operator wff-type negated print conditional equivalence
							       old--lits old-+lits pr quote-wff controlled-vars
							       (save-evars *e-vars*))
					; returns a wff-name atom - which has property 'wff on it with the 
					; normalized version of the formula in it
  (cond ((member wff '(YES NO inconsistent)) wff)
	((and (symbolp wff) (setq pr (getp wff 'print)))
	 (norm-wff pr context ttype))
	((symbolp wff) 
	 (setq newname (replace-arg wff))
	 (when (and (epi-variable-p newname) *meaning-postulate*
		    (not *quoted*) (not (equal ttype 'term)))
	   (fatal-error 'normalize-catcher
			"Normalization: Meta-level variable " wff 
			" should only occur " *nl* 
			"       in WFF position within quasi-quoted expressions")
	   )
	 (when (symbolp newname)
	   (unless (getp newname 'entity-type) (change-property newname 'entity-type 'wff))
	   (unless (getp newname 'wff-type)
	     (change-property newname 'wff-type (if (epi-variable-p newname) 
						    'variable 'constant)))
	   (when (getp newname 'print)
	     (if (equal context '+) 
		 (push newname *+lits*)
	       (push newname *-lits*))) 
	   )
	 newname)
	((not (listp wff))
	 (fatal-error 'normalize-catcher
		      "Normalization: Expecting formula here - got " wff))

					; look for "names" embedded - if there is a name, use it, else 
					; generate one
	(t				;scan for name symbol
	 (setq namestuff (name-scan wff ttype))
	 (when namestuff
	   (setq newname (car namestuff))
	   (setq wff (cadr namestuff)) )
	 
	 (cond				;  quantifier variable {rest} main (negation already distrib)
	  ((and (> (list-length wff) 2)
		(not (and (numberp (car wff)) 
			  (if (symbolp (cadr wff)) 
			      (getp (cadr wff) 'predicate) t)))
		(or (is-quantifier (car wff)) (equal (car wff) 'L))
		)
	   (let ((match-quant (if (equal context '-)
				  '(E E! WH the)
				'(A most many some few)))
		 (save-subs *subs*)
		 new-sub
		 (var (remove-sort (cadr wff)))
		 sort)

	     (setq quantifier (norm-quant (car wff) context))
	     (unless (is-variable var)
	       (fatal-error 'normalize-catcher
			    "Normalization: " var 
			    " cannot be used as a variable ")) 
	     (when (and (equal quantifier 'L)
			(cdddr wff))
	       (fatal-error 'normalize-catcher
			    "Normalization: Lambda "
			    "expression may"
			    " only have one wff after the variable " 
			    wff)) 
	     
	     (when (cddddr wff)
	       (fatal-error 'normalize-catcher
			    "Normalization: Quantified "
			    "expression may"
			    " only have two wffs after the variable " 
			    wff)) 
	     (setq sort (getp var 'sort))
	     (make-variable var)
	     (change-property var 'sort sort)
	     (if (and *meaning-postulate* (not *quoted*)
		      (not (equal quantifier 'L))
		      (or (null sort)
			  (not (member (hier-rel 'meta-entity sort)
				       '(equivalent subsumes))))
		      )
		 (fatal-error 'normalize-catcher
			      "Normalization: Variables in mp " wff
			      *nl* "      must be meta variables"
			      "- i.e. of sort pred, term, op, wff, etc" )
	       )
	     
	     (if (and (if *qa-mode* nil (not *meaning-postulate*))
		      (not *interim*)
		      (member (getp var 'sort) '(wff pred op term)))
		 (fatal-error 'normalize-catcher
			      "Normalization: Wff, pred, term, and op variables " 
			      "are only allowed " *nl*
			      "         in meaning postulates - "
			      var " of sort " (getp var 'sort) " is illegal here")
	       )
	     (when (equal quantifier 'L) 
	       (setq old-+lits *+lits*)
	       (setq old--lits *-lits*))
					; rename variable, add to *subs* (replacing other sub 
					; for that same variable if it has been used before)
	     (setq variable var)
	     (when (and *new-vars*
			(if *quoted* 
			    (if (assoc variable *subs*)
				(let () (setq variable (replace-arg variable)) nil)
			      t)
			  t))
	       (cond ((or (and (equal quantifier 'wh)
			       (member variable *wh-pattern*))
			  (and (equal quantifier 'e)
			       (member variable *e-pattern*)))
					; ensure nothing else gets normalized to it
		      (setq *subs* (merge-subs *subs* 
					       (setq new-sub (list var variable)))) )
		     (t (setq variable (new-var var
						(member quantifier match-quant)
						quantifier context embed))
			(setq *subs* (merge-subs *subs* 
						 (setq new-sub (list var variable)))) )
		     ))
	     
	     (if (equal quantifier 'L) 
		 (change-property variable 'lambda t))
	     (unless (equal var variable)
	       (setq *normed-subs* (push new-sub *normed-subs*)))
	     (setq wff-type 'quantified)
	     (cond ((cdddr wff)
		    (setq restrict 
		      (norm-wff (caddr wff) (if (member quantifier
							'(WH E! E the))
						context alt-context)))
		    (setq main (norm-wff (cadddr wff) context))
		    )
		   (t (setq main (norm-wff (caddr wff) context))))
	     (if (or (null restrict) (member quantifier '(E E! WH the)))
		 (setq rr (eval-quantified restrict 
					   (if (member restrict '(YES NO inconsistent)) restrict nil)
					   main (if (member main '(YES NO inconsistent)) main nil)))
	       (setq rr (eval-conditional restrict 
					  (if (member restrict '(YES NO inconsistent)) restrict nil)
					  main (if (member main '(YES NO inconsistent)) main nil)))
	       )
	     (when rr
	       (when (symbolp rr) 
		 (setq *normed-prob* 
		   (combine-probs *normed-prob*
				  (cond ((getp quantifier 'prob)) (t 1))))
		 (return-from norm-wff rr))
	       (setq restrict (car rr))
	       (setq main (cadr rr))
	       ) 
	     
	     (if restrict
		 (setq print (list (print-info quantifier) variable
				   (print-info restrict) (print-info main)))
	       (setq print (list (print-info quantifier) variable
				 (print-info main))) )
	     
	     (if (not (member quantifier '(WH E! E the))) ;)
		 (setq conditional t))
	     (if (getp main 'equivalence) (setq equivalence t))
	     
	     (when new-sub
	       (setq *subs* (undo-sub *subs* save-subs new-sub)))
	     (when (equal (car wff) 'L)
	       (setq *+lits* old-+lits)
	       (setq *-lits* old--lits))
	     )
	   )
	  
	  ((is-function (car wff))
					;(if (or *meaning-postulate* *simplification-schema*
					;	  (equal (car wff) 'epilog::%subst))
					;      (return-from norm-wff (norm-term wff context)) )
	   (if (equal (car wff) 'set-of)
	       (return-from norm-wff (norm-term wff context))
	     (fatal-error 'normalize-catcher
			  "Normalization: Illegal syntax - " wff 
			  " - functional term cannot be wff")))
	  
	  ((is-record wff)
	   (if (equal (car (cdr (cadr wff))) 'set)
	       (return-from norm-wff (norm-term wff context))
	     (fatal-error 'normalize-catcher
			  "Normalization: Illegal syntax - " wff 
			  " - record cannot be wff")))
	  
	  ((is-quasi-quoted-expression wff)
					; do replace-args but no more
	   (let ((old-+lits *+lits*)
		 (old--lits *-lits*)
		 res)
	     (setq *quoted* t)
	     (setq res (norm-wff (cadr wff) context))
	     (setq *quoted* nil)
	     (setq *+lits* old-+lits)
	     (setq *-lits* old--lits)
	     (setq wff-type 'quasi-quoted-expression)
	     (setq quote-wff res)
	     (setq print (list 'qquote (print-info res)))
	     )
	   )
	  
	  ((is-quoted-expression wff)
	   (setq print wff)
	   (setq wff-type 'quoted-expression)
	   )
	  
					;look for top level operators   (wff-op wff)
	  ((and (not (and (numberp (car wff)) (> (list-length wff) 2)
			  (if (symbolp (cadr wff))
			      (getp (cadr wff) 'predicate) t)))
		(or (is-operator (car wff) 'wff)
		    (and (is-operator (car wff) 'wff t)
			 (is-wff (cadr wff)))))
	   (let* ((op (norm-op (car wff) context 'wff))
		  (atype (if (symbolp op) (getp op 'args))))
	     (setq wff-type 'modified)
	     (setq operator op)
	     (setq arguments 
	       (mapcar #'(lambda (a)
			   (case atype
			     (pred (norm-pred a context))
			     (wff (norm-wff a context))
			     (term (norm-term a context))
			     (otherwise (norm-wff a context))))
		       (cdr wff)))
	     (if (getp (car arguments) 'conditional)
		 (setq conditional t))
	     (if (getp (car arguments) 'equivalence)
		 (setq equivalence t))
	     (setq print (append (list (print-info op))
				 (mapcar 'print-info arguments)))
	     )
	   )
	  
	  ((and (equal (car wff) 'not) (is-quantifier (cadr wff)))
	   (let* ((old-+lits *+lits*)
		  (old--lits *-lits*)
		  (w (norm-wff (cdr wff) context)))
	     (setq print (append (list 'not) (getp w 'print)))
	     (setq negated t)
	     (setq predicate (getp w 'pred))
	     (setq main (getp w 'main))
	     (setq restrict (getp w 'restrict))
	     (setq operator (getp w 'operator))
	     (setq arguments (getp w 'arguments))
	     (setq quantifier (getp w 'quantifier))
	     (setq conditional (getp w 'conditional))
	     (setq equivalence (getp w 'equivalence))
	     (setq variable (getp w 'variable))
	     (setq wff-type 'quantified)
	     (setq *+lits* old-+lits)
	     (setq *-lits* old--lits)
	     )
	   )
	  
	  ((and (equal (car wff) 'not) (symbolp (cadr wff)) 
		(null (cddr wff)) (equal (find-sort (cadr wff)) 'wff))
	   (if (getp (cadr wff) 'print)
	       (return-from norm-wff (norm-wff 
				      (append (list 'not) (getp (cadr wff) 'print)) context)))
	   (setq print (list 'not (norm-wff (cadr wff) context)))
	   (setq negated t)
	   (setq wff-type 
	     (if (epi-variable-p (cadr wff)) 'variable 'constant)))
	  
	  ((and (equal (car wff) 'not) (listp (cadr wff)) 
		(null (cddr wff)))
	   (return-from norm-wff
	     (norm-wff (append (list 'not) (cadr wff)))) )
	  
	  ((not (and (setq pred (prop-pred wff))
		     (setq args (prop-args wff))))
	   (fatal-error 'normalize-catcher
			"Normalization: Illegal syntax - " wff 
			" - could not find predicate and arguments"))
	  
	  ;; true business - if this is the top level, just use
	  ;; first argument. Otherwise keep the true, but prepare
	  ;; keys differently and flag for normalization of variables
	  
	  ((equal pred 'true)
	   (setq wff-type 'prefix)
	   (cond ((cdr args)
		  (fatal-error 'normalize-catcher
			       "Normalization: Illegal syntax - " wff 
			       " - true can only have one argument") )
		 ((and (not *meaning-postulate*)
		       (not (is-negated wff))
		       (or (is-quasi-quoted-expression (car args))
			   (and (symbolp (car args))
				(getp (car args) 'wff-type)
				(getp (car args) 'print)
				(not (epi-variable-p (car args))))
			   (and (listp (car args))
				(or (equal (caar args) 'that)
				    (is-wff (car args))))
			   ))
		  (return-from norm-wff 
		    (norm-wff 
		     (if (is-quasi-quoted-expression (car args))
			 (cadar args)
		       (if (symbolp (car args))
			   (car args)
			 (cadar args))) context))
		  )
		 ;; e.g. (term4 true).  Case added by ANK
		 ((and (symbolp (car args))
		       (eq (getp (car args) 'entity-type) 'term)
		       (eq (getp (car args) 'operator) 'that))
		  (return-from norm-wff
		    ;; I don't see why the recursive call is
		    ;; necessary, since the value of the term's
		    ;; argument property is a wff symbol, which
		    ;; was already normalized when it was
		    ;; built.  But that's how the other cases
		    ;; are handled, so I'll follow suit.
		    (norm-wff (car (getp (car args) 'arguments)))))
		 (t (let ()
		      (setq predicate (norm-pred pred context))
		      (setq negated (is-negated wff))
		      (setq arguments (list (norm-term (car args) context)))
		      (setq print (append (if negated (list 'not))
					  (list (print-info (car arguments))
						(print-info predicate))))
		      ))
		 ))
	  
					;  wff logical-conn wff+ (negation already distributed)
	  ((logical-conn pred)
	   (setq wff-type 'logical)
	   (setq predicate pred)
	   (when (and (listp predicate)
		      (numberp (car predicate))
		      (<= (car predicate) 1.0))
	     (setq pred (car predicate))
	     (setq controlled-vars (cdr predicate))
					;	(setq predicate pred)
	     )
	   (if (or (member pred '(implies ~> <=>)) (numberp pred))
	       (let ()
		 (unless (= (list-length args) 2)
		   (fatal-error 'normalize-catcher
				"Normalization: " pred 
				" requires exactly 2 wff arguments - "
				wff)) 
		 (setq arguments
		   (list (norm-wff (car args) alt-context)
			 (norm-wff (cadr args) context))) 
		 (setq rr (eval-conditional (car arguments)
					    (if (member (car arguments) '(YES NO inconsistent))
						(car arguments) nil)
					    (cadr arguments)
					    (if (member (cadr arguments) '(YES NO inconsistent))
						(cadr arguments) nil) (equal pred '<=>)
						(equal pred '~>)))
		 (when rr
		   (if (numberp pred) 
		       (setq *normed-prob* (* *normed-prob* pred)))
		   (if (symbolp rr)
		       (return-from norm-wff rr))
		   (setq arguments rr))
		 (when (null (car arguments))
		   (return-from norm-wff (cadr arguments)))
		 (setq conditional t)
		 (if (equal pred '<=>) (setq equivalence t))
		 )
	     (let ()
	       (setq arguments 
		 (mapcar #'(lambda (a) (norm-wff a context)) 
			 (remove-duplicates args :test 'equal) 
			 ))
	       (setq rr (eval-logical arguments pred))
	       (when (symbolp rr) (return-from norm-wff rr))
	       (setq arguments rr)
	       (if (equal (list-length arguments) 1)
		   (return-from norm-wff (car arguments)))
	       )
	     )
	   (when controlled-vars
	     (setq controlled-vars (mapcar #'(lambda (a) 
					       (norm-term a context)) controlled-vars))
	     (setq predicate (append (list (car predicate))
				     controlled-vars))
	     )
	   (when (member pred '(and or))
					; order the arguments in some consistent manner
	     (setq arguments (sort arguments 'sort-args)))
	   (setq print (append (list (print-info (car arguments)) predicate)
			       (mapcar 'print-info (cdr arguments))))
	   )
	  
					;  {negation} wff episodic-conn term
	  ((episodic-conn pred)
	   (unless (= (list-length args) 2)
	     (fatal-error 'normalize-catcher
			  "Normalization: " pred " requires "
			  "exactly 2 arguments - a wff and a term - got "
			  wff)) 
	   (setq old-+lits *+lits*)
	   (setq old--lits *-lits*)
	   (setq wff-type 'episodic)
	   (if (symbolp (last1 args))
	       (if (member pred '(** * @))
		   (make-episode (last1 args))
		 (make-propos (last1 args))))
	   (if (epi-variable-p (last1 args)) (push (last1 args) *vars-to-fix*))
	   (setq predicate pred)
	   (setq negated (is-negated wff))
	   (setq arguments (list (norm-wff (car args) context 'wff
					   (equal pred '_))
				 (norm-term (cadr args) context)))
	   (setq print (append (if negated (list 'not))
			       (list (print-info (car arguments))
				     pred)
			       (mapcar 'print-info (cdr arguments))))
	   (setq *+lits* old-+lits)
	   (setq *-lits* old--lits)
	   )
	  
					;  {negation} wff causal-conn wff
	  ((causal-conn pred)
	   (unless (= (list-length args) 2)
	     (fatal-error 'normalize-catcher
			  "Normalization: " pred " requires "
			  "exactly 2 wff arguments - got " wff)) 
	   (setq old-+lits *+lits*)
	   (setq old--lits *-lits*)
	   (setq wff-type 'causal)
	   (setq predicate pred)
	   (setq negated (is-negated wff))
	   (setq arguments 
	     (mapcar #'(lambda (a) (norm-wff a context)) args))
	   (setq print (append (if negated (list 'not))
			       (list (print-info (car arguments))
				     pred)
			       (mapcar 'print-info (cdr arguments))))
	   (setq *+lits* old-+lits)
	   (setq *-lits* old--lits)
	   )
	  
	  ((is-lambda-expr pred)
					; do lambda conversion if possible (the term pred term* case,
					; but needs special treatment because conversion possible,
					; which involves arguments as well as predicate)
	   (cond ((setq conv 
		    (lambda-conversion (if (symbolp pred)
					   (getp pred 'print)
					 pred) args t))
		  (if (is-negated wff) 
		      (setq conv (append (list 'not) conv)))
		  (return-from norm-wff (norm-wff conv context)))
		 (t (setq wff-type 'prefix)
		    (setq predicate (norm-pred pred context))
		    (setq negated (is-negated wff))
		    (setq arguments 
		      (mapcar #'(lambda (a) 
				  (norm-term a context)) args))
		    (setq print (append (if negated (list 'not))
					(list (print-info (car arguments))
					      (print-info predicate))
					(mapcar 'print-info (cdr arguments))))
		    ))
	   )

	  (t
	   (setq wff-type 'prefix)
	   (when (and (listp pred) (not (is-lambda-expr pred))
		      (not (is-operator (car pred))))
					;needs to be flattened
	     (when (traced-p 'lambda)
	       (print-line "Normalization: Flattening " pred
			   " to " (car pred) " with arguments "
			   (cdr pred)))
	     (setq args (append (list (car args)) 
				(cdr pred) (cdr args)))
	     (setq pred (car pred)))

	   (setq predicate (norm-pred pred context))
	   (setq negated (is-negated wff))
	   (when (equal (pred-type predicate) 'lambda)
	     (let ((conv (lambda-conversion (getp pred 'print) args 
					    t)))
	       (when conv
		 (if negated 
		     (setq conv (append (list 'not) conv)))
		 (return-from norm-wff 
		   (norm-wff conv context)))
	       ))
	   (setq arguments 
	     (mapcar #'(lambda (a) (norm-term a context)) args))
	   (setq print (append (if negated (list 'not))
			       (list (print-info (car arguments))
				     (print-info predicate))
			       (mapcar 'print-info (cdr arguments))))
	   ))

					; save info just gathered
	 (setq newname (save-info print newname ttype))
	 (setp newname 'wff-type wff-type)
	 (setp newname 'print 
	       (if (and (getp newname 'explicit) (not *expand-names*)) 
		   newname print))
	 (when conditional (setp newname 'conditional conditional))
	 (when equivalence (setp newname 'equivalence equivalence))
	 (when operator (setp newname 'operator operator))
	 (when negated (setp newname 'negated negated))
	 (when quote-wff (setp newname 'quote-wff quote-wff))
	 (when predicate (setp newname 'pred predicate))
	 (when (or conditional (equal wff-type 'episodic))
	   (setq *e-vars* save-evars))
	 (when arguments (setp newname 'arguments arguments))
	 (when quantifier (setp newname 'quantifier quantifier))
	 (when controlled-vars (setp newname 'controlled-vars controlled-vars))
	 (when variable (setp newname 'variable variable))
	 (when main (setp newname 'main main))
	 (when restrict (setp newname 'restrict restrict))
					; save it in hash table
	 (when (or (member wff-type '(prefix causal episodic))
		   (and (member wff-type '(modified))
			(not (if (equal context '+) *+lits* *-lits*))) )
					; (or (and (equal quantifier 'E) (equal context '+)))
	   (if (equal context '+)
	       (push newname *+lits*)
	     (push newname *-lits*))
	   )
	 newname)
	)
  )

(defun presort (a b)
  (if (and (listp b) (is-quantifier (car b)))
      nil
      (and (listp a) (is-quantifier (car a)))))



(defun find-vars (wff)
  (when wff
  (case (wff-type wff)
	(quantified (append (list (getp wff 'variable))
			    (find-vars (getp wff 'restrict))
			    (find-vars (getp wff 'main))))
	(logical (apply 'append (mapcar 'find-vars (getp wff 'arguments))))
	(episodic (find-vars (car (getp wff 'arguments))))
	(t nil)))
  )

(defun find-args (wff &aux args)
  (when wff
  (case (wff-type wff)
	(quantified (append (find-args (getp wff 'restrict))
			    (find-args (getp wff 'main))))
	(logical (apply 'append (mapcar 'find-args (getp wff 'arguments))))
	(episodic 
	 (setq args (getp wff 'arguments))
	 (append (find-args (car args)) (list (cadr args))))
	(prefix (getp wff 'arguments))
	(t nil)))
  )


(defun sort-args (a b &aux
		    (pred1 (getp a 'pred))
		    (pred2 (getp b 'pred))
		    (q1 (getp a 'quantifier))
		    (q2 (getp b 'quantifier))
		    e1 e2 p1 p2 t1 t2 l1 l2)
  (unless (member pred1 '(implies ~> <=>)) (setq p1 pred1))
  (unless (member pred2 '(implies ~> <=>)) (setq p2 pred2))
  (when p1 (setq t1 (type-pred-p p1)))
  (when p2 (setq t2 (type-pred-p p2)))
  (when p1 (setq e1 (episodic-conn p1)))
  (when p2 (setq e2 (episodic-conn p2)))
  (when p1 (setq l1 (logical-conn p1)))
  (when p2 (setq l2 (logical-conn p2)))
  (cond ((and q2 (not q1)) (not (member-if 'epi-variable (find-args a))))
	((and q1 (not q2)) nil) 
	((and q1 q2) nil) 
	(q2 nil) 
        (l1 (if l2 (< (sxhash l1) (sxhash l2)) nil))
	(l2 t)
        (e1 (if e2 (< (sxhash p1) (sxhash p2)) nil))
	(e2 t)
        (t1 (if t2 (< (sxhash p1) (sxhash p2)) t))
	(t2 nil)
	(p1 (if p2 (< (sxhash p1) (sxhash p2)) t))
	(p2 nil)
	(t (< (sxhash pred1) (sxhash pred2)))
	))

; top level norm-term
(defun normalize-term (term &optional (context '+) &aux term-name)
  (setq *new-vars* nil)
  (catch 'normalize-catcher (setq term-name (norm-term term context)))
  term-name)

;term consists of:
;   constant
;   (nominalization-op pred)
;   (wff-nominalization-op wff)
;   (function term+)
; Normalized term has the following properties:
;  term-type: variable, constant, function, modified, wff, record, 
;             quoted-expression
;  operator, arguments for function and modified expressions
;  all wff properties for wff

(defun norm-term (term &optional (context '+) &aux namestuff newname skol s
						   (orig-term term)
						   term-type operator arguments print pr et
						   (save-wh *wh-vars*))
					; returns an atom corresponding to the term
					; must have property 'term with normalized term
  (cond ((and (symbolp term) (setq pr (getp term 'print)))
	 (if (and (setq et (getp term 'entity-type)) (not (equal et 'term)))
	     term
	   (norm-term pr context)))
	((symbolp term) 
	 (setq term (remove-sort term))
					; check to see if legal term
	 (when *new-vars*
	   (if (and (symbolp term) (setq skol (getp term 'alternate-names)))
	       (setq term (car skol)))
	   (if (and (symbolp term) 
		    (not (assoc term *subs*))
		    (or (setq skol (getp term 'skol-const))
			(setq skol (cadr (assoc term *vars-to-restore*)))))
	       (setq term skol)) )
	 (setq newname (replace-arg term))
	 (when (and (symbolp newname) 
		    (not (or *qa-mode* *meaning-postulate* *meta*))
		    (equal context '+)
		    *insertion*
		    (or (operator-p newname) (function-p newname)
			(getp newname 'indicators)))
	   (fatal-error 'normalize-catcher
			"Normalization: " newname " is not a legal term")) 
	 (when (symbolp newname)
	   (unless (getp newname 'entity-type) 
	     (change-property newname 'entity-type 'term)
	     (if *meta* (change-property newname 'meta t)) )
	   (unless (getp newname 'term-type)
	     (change-property newname 'term-type (if (epi-variable-p newname) 
						     'variable 'constant)))
	   )
	 newname)
	((atom term)
	 term)

	(t				;scan for name symbol
	 (setq namestuff (name-scan term 'term))
	 (when namestuff
	   (setq newname (car namestuff))
	   (setq term (cadr namestuff)) )
	 
	 (cond				;   (function term+)
	  ((is-function-term term)
	   (setq term-type 'function)
	   (setq operator (car term))
	   (setq arguments 
	     (mapcar #'(lambda (a) (norm-term a context)) (cdr term)))
	   (setq print (append (list (car term))
			       (mapcar 'print-info arguments)))
	   )
	  
					; combine operator tests
					;   (nominalization-op pred)
					;   (wff-nominalization-op wff)
	  ((is-operator (car term) 'term)
	   (let* ((op (norm-op (car term) 'term))
		  (atype (if (symbolp op) (getp op 'args)))
		  (o-lit *-lits*)
		  (o+lit *+lits*))
	     (setq term-type 'modified)
	     (setq operator op) 
	     (setq arguments 
	       (mapcar #'(lambda (a)
			   (case atype
			     (pred (norm-pred a context))
			     (wff (norm-wff a context 'wff nil)) ;t
			     (term (norm-term a context))
			     (otherwise (norm-term a context))))
		       (cdr term)))
	     (setq print (append (list (print-info op))
				 (mapcar 'print-info arguments)))
	     (setq *-lits* o-lit)
	     (setq *+lits* o+lit)
	     (if (equal op 'answer-to) (setq *wh-vars* save-wh))
	     ))

	  ((and (is-operator term)
		(member (setq s (find-sort term)) '(pred operator)))
					; find out what kind of operator and normalize for that
					; instead
	   (when (and (not (or *qa-mode* *meaning-postulate* *meta*))
		      (equal context '+) *insertion*)
	     (fatal-error 'normalize-catcher
			  "Normalization: " term " is not a legal term"))
	   (case s
	     (pred (return-from norm-term (norm-pred term context)))
	     (operator (return-from norm-term (norm-op term context)))
	     )
	   )

	  ((is-record term)
	   (setq print term)
	   (setq term-type 'record))

	  ((is-quasi-quoted-expression term)
	   (cond ((is-wff (cadr term))
		  (if namestuff (setf (symbol-plist newname) nil))
		  (return-from norm-term (norm-wff orig-term context)))
		 (t (setq print (replace-quoted-arg term))
		    (setq term-type 'quasi-quoted-expression)) ) )

	  ((is-quoted-expression term)
	   (setq print term)
	   (setq term-type 'quoted-expression))

	  ((equal (car term) 'L)
	   (if namestuff (setf (symbol-plist newname) nil))
	   (return-from norm-term (norm-pred orig-term context)) )

					; propositions - need a better test here - ?
	  (t				; must be proposition then
	   (let (old-+lits old--lits result)
	     (setq old-+lits *+lits*)
	     (setq old--lits *-lits*)
	     (if namestuff (setf (symbol-plist newname) nil))
	     (setq result (norm-wff orig-term context 'term))
	     (change-property result 'term-type 'wff)
	     (setq *+lits* old-+lits)
	     (setq *-lits* old--lits)
	     (return-from norm-term result)))
	  )
	 (setq newname (save-info print newname 'term))
	 (change-property newname 'term-type term-type)
	 (change-property newname 'print 
			  (if (and (getp newname 'explicit) (not *expand-names*)) 
			      newname print))
	 (when operator (change-property newname 'operator operator))
	 (when arguments (change-property newname 'arguments arguments))
	 newname)
	)
  )

(defun replace-quoted-arg (item)
  (cond ((listp item) (mapcar 'replace-quoted-arg item))
	(t (replace-arg item)))
  )

; top level norm-op
(defun normalize-op (op &optional (context '+) result &aux op-name)
  (setq *new-vars* nil)
  (catch 'normalize-catcher (setq op-name (norm-op op context result)))
  op-name)

; All operators combined into one test - operators consist of:
;    operator name
;    (operator-creating-operator arguments)
; Normalized operator has the following properties:
;  op-type: variable, constant, modified
;  operator, arguments for modified expressions

(defun norm-op (op &optional (context '+) result &aux 
		       op-type arguments operator namestuff newname optype
		       op-res args print nr pr)
  ; returns an atom corresponding to the operator
  ; must have properties 'op, 'result, 'args associated with result

  (cond ((and (symbolp op) (setq pr (getp op 'print)))
	 (norm-op pr context result))
	((atom op) 
	 (setq newname (replace-arg op))
	 (when (and (symbolp newname) 
		    (not (operator-p newname)) (not (epi-variable-p newname))
		    (or (quantifier-p newname) (function-p newname)))
	       (fatal-error 'normalize-catcher
                  "Normalization: " newname 
			" is not a legal operator")) 
	 (when (and (epi-variable-p newname) *meaning-postulate* (not *quoted*))
	       (fatal-error 'normalize-catcher
			   "Normalization: Meta-level variable " op 
			   " should only occur " *nl* 
			   "       in OPERATOR position within quasi-quoted expressions")
	       )
	 (when (symbolp newname)
	 (unless (getp newname 'entity-type) 
		 (change-property newname 'entity-type 'operator))
	 (unless (getp newname 'op-type)
		 (change-property newname 'op-type (if (epi-variable-p newname) 
					     'variable 'constant)))
	 (when (and result (setq nr (getp newname 'result)) 
		    (not (equal result nr)))
	       (fatal-error 'normalize-catcher
                "Normalization: " op " is expected to "
		      " return a result of " result *nl* 
		      "       but returns a " nr " instead"))
	 (unless (getp newname 'result)
		 (change-property newname 'result result))
	 (unless (getp newname 'args)
		 (if (setq optype (getp newname 'operator-type))
		     (change-property newname 'args (getp optype 'args)))
		 ) )
	 newname)

	(t ;scan for name symbol
	   (setq namestuff (name-scan op 'operator))
	   (when namestuff
		 (setq newname (car namestuff))
		 (setq op (cadr namestuff)) )
	   (let* ((opn (norm-op (car op) context))
		  (atype (if (symbolp opn) (getp opn 'args))))
		 (setq op-type 'modified)
		 (setq operator opn)
		 (setq arguments 
			       (mapcar #'(lambda (a)
				 (case atype
				       (pred (norm-pred a context))
				       (wff (norm-wff a context))
				       (term (norm-term a context))
				       (otherwise (norm-term a context))))
				    (cdr op)))
		 (setq print (append (list (print-info opn))
				     (mapcar 'print-info arguments)))
		 ;set result and arg types
		 (cond ((setq op-res (getp opn 'result))
			(setq result (getp op-res 'result))
			(setq args (getp op-res 'args)))
		       (t (setq result result)
			  ))
		 )
	   (setq newname (save-info print newname 'operator))
	   (change-property newname 'op-type op-type)
	   (change-property newname 'print 
		 (if (and (getp newname 'explicit) (not *expand-names*)) 
		     newname print))
	   (when operator (change-property newname 'operator operator))
	   (when arguments (change-property newname 'arguments arguments))
	   (when args (change-property newname 'args args))
	   (when result (change-property newname 'result result))
	   newname)
	)
  )

; top level norm-pred
(defun normalize-pred (pred &optional (context '+) &aux pred-name)
  (setq *new-vars* nil)
  (catch 'normalize-catcher (setq pred-name (norm-pred pred context)))
  pred-name)

; pred consists of:
;    predicate name
;    (modifier pred)
;    (pred-modifier pred+)
;    lambda-expr
; Normalized predicate has the following properties:
;  pred-type: variable, constant, modified, lambda
;  operator, arguments for modified and complex lambda expressions
;  all quantified wff properties for simple lambda predicates

(defun norm-pred (pred &optional (context '+) &aux namestuff newname
		       (gpred pred)
		       pred-type operator arguments print pr)
  ; returns an atom corresponding to the predicate
  ; must have properties 'pred associated with the result

  (cond ((and (symbolp pred) (setq pr (getp pred 'print)))
	 (norm-pred pr context))
	((atom pred) 
	 (setq pred (replace-arg pred))
	 (when (and (symbolp pred) 
		    (or (quantifier-p pred)
			(function-p pred) (operator-p pred)
			(and (equal (getp pred 'entity-type) 'term)
			     (not (getp pred 'meta))
			     (not (equal (find-sort pred) 'pred)))
			))
	       (fatal-error 'normalize-catcher
                "Normalization: " pred 
			" is not a legal predicate")) 
	 (when (and (symbolp pred) (epi-variable-p pred) 
		    *insertion* (if *qa-mode* nil (not *meaning-postulate*))
		    )
	       (fatal-error 'normalize-catcher
                "Normalization: Variable predicates are only "
		      "allowed in meaning postulates - " pred))
	 (when (and (epi-variable-p pred) *meaning-postulate* (not *quoted*))
	       (fatal-error 'normalize-catcher
			   "Normalization: Meta-level variable " gpred 
			   " should only occur " *nl* 
			   "       in PREDICATE position within quasi-quoted expressions")
	       )
	 (when (and (symbolp pred) (not (epi-variable-p pred))
		    (not (getp pred 'print))
		    (not (equal (get pred 'entity-type) 'pred))
		    (not (getp pred 'specialists))
		    (not (getp pred 'indicators))
		    (not (pred-info pred)))
		   (warn-error "Normalization: " pred " is a new predicate")
	       )
	 (setq newname pred)
	 (when (symbolp newname)
	 (unless (cond ((getp newname 'entity-type) 
			(if (and (equal (getp newname 'entity-type) 'term)
				 (getp newname 'meta))
			    nil
			    t)))
		 (change-property newname 'entity-type 'pred))
	 (unless (getp newname 'pred-type)
		 (change-property newname 'pred-type (if (epi-variable-p newname) 
					     'variable 'constant)))
	       )
	 newname)
	   
	((is-lambda-pred pred)
	 (setq newname (norm-wff pred context 'pred))
	 (change-property newname 'pred-type 'lambda)
	 (setq newname (cond ((lambda-reduce newname)) (t newname)))
	 (return-from norm-pred newname)
	 )

	(t ;scan for name symbol
	   (setq namestuff (name-scan pred 'pred))
	   (when namestuff
		 (setq newname (car namestuff))
		 (setq pred (cadr namestuff)) )

	   (cond ((is-lambda-expr pred)
		  (unless newname (setq newname (new-symbol 'pred)))
		  (change-property newname 'operator (norm-pred (car pred) context))
		  (change-property newname 'arguments 
		       (mapcar #'(lambda (a) (norm-term a context)) (cdr pred)))
		  (change-property newname 'print 
			(append (list (print-info (getp newname 'operator)))
				(mapcar 'print-info (getp newname 'arguments))))
		  (change-property newname 'pred-type 'lambda)
		  (return-from norm-pred newname)
		  )
		 ; combine operator tests into one - check first to make
		 ; sure its a legal operator!!
		 ((is-operator (car pred) 'pred)
		  (let* ((op (norm-op (car pred) context 'pred))
			   (atype (if (symbolp op) (getp op 'args) 'pred)))
			  (setq pred-type 'modified)
			  (setq operator op)
			  (setq arguments
					(mapcar #'(lambda (a)
                                           (case atype
						 (pred (norm-pred a context))
						 (wff (norm-wff a context))
						 (term (norm-term a context))
						 (otherwise (norm-pred a context))))
						(cdr pred)))
			  (setq print (append (list (print-info op))
					      (mapcar 'print-info arguments)))
			  ))
		 ; (pred terms)  - short form of lambda? unflattened version
		 (t 
		  ; how to represent it?
		  ; build into lambda for now
		  (return-from norm-pred (norm-pred 
			  (list 'L 'new-l-var (append (list 'new-l-var) pred)) 
			  context))
		  )
		 )
	   (setq newname (save-info print newname 'pred))
	   (change-property newname 'pred-type pred-type)
	   (change-property newname 'print 
		 (if (and (getp newname 'explicit) (not *expand-names*)) 
		     newname print))
	   (when operator (change-property newname 'operator operator))
	   (when arguments (change-property newname 'arguments arguments))
	   newname)
	)
  )

(defun norm-quant (quant &optional (context '+) &aux namestuff newname
		       quant-type operator arguments print pr)
  ; returns an atom corresponding to the quantifier

  (cond ((and (symbolp quant) (setq pr (getp quant 'print)))
	 (norm-quant pr context))
	((atom quant) 
	 (when (and (symbolp quant) 
		    (function-p quant) (operator-p quant)
		    (getp quant 'specialists) (getp quant 'indicators))
	       (fatal-error 'normalize-catcher
                "Normalization: " quant " is not a legal quantifier")) 
	 (when (and (symbolp quant) (epi-variable-p quant) 
		    *insertion*
		    (if *qa-mode* nil (not *meaning-postulate*))
		    )
	       (fatal-error 'normalize-catcher
                "Normalization: Variable quantifiers are only "
		      "allowed in meaning postulates - " quant))
	 (setq newname (replace-arg quant))
	 (when (and (epi-variable-p newname) *meaning-postulate* (not *quoted*))
	       (fatal-error 'normalize-catcher
			   "Normalization: Meta-level variable " quant 
			   " should only occur " *nl* 
			   "       in QUANTIFIER position within quasi-quoted expressions")
	       )
	 (when (symbolp newname)
	 (unless (getp newname 'entity-type) 
		 (change-property newname 'entity-type 'quantifier))
	 (unless (getp newname 'quant-type)
		 (change-property newname 'quant-type (if (epi-variable-p newname) 
					     'variable 'constant)))
	       )
	 newname)
	   
	(t ;scan for name symbol
	   (setq namestuff (name-scan quant 'quantifier))
	   (when namestuff
		 (setq newname (car namestuff))
		 (setq quant (cadr namestuff)) )

	   (let* ((op (norm-op (car quant) context 'quantifier))
		  (atype (if (symbolp op) (getp op 'args) 'quantifier)))
		 (setq quant-type 'modified)
		 (setq operator op)
		 (setq arguments (mapcar #'(lambda (a)
                                      (case atype
					    (pred (norm-pred a context))
					    (quantifier (norm-quant a context))
					    (wff (norm-wff a context))
					    (term (norm-term a context))
					    (otherwise (norm-quant a context))))
					 (cdr quant)))
		 (setq print (append (list (print-info op))
				     (mapcar 'print-info arguments)))
		 )
	   (setq newname (save-info print newname 'quantifier))
	   (change-property newname 'quant-type quant-type)
	   (change-property newname 'print 
		 (if (and (getp newname 'explicit) (not *expand-names*)) 
		     newname print))
	   (when operator (change-property newname 'operator operator))
	   (when arguments (change-property newname 'arguments arguments))
	   newname)
	)
  )

(defun replace-arg (arg &aux sub)
  (cond ((symbolp arg)
	 (if (setq sub (assoc arg *subs*)) (cadr sub) arg))
	(t arg))
  )

; should also be able to reduce within * ** @ or _ formula
(defun lambda-reduce (pred)
  (cond ((not (symbolp pred)) nil)
	((getp pred 'wff-type)
	 ; single level, simple predicate
	 (let* ((main (getp pred 'main))
		(p (getp main 'pred)))
	      (if (and (= (list-length (getp main 'arguments)) 1)
		       p (or (null (getp p 'pred-type))
			     (equal (getp p 'pred-type) 'constant)) )
		  p))
	 )
	((null (getp pred 'arguments))
	 (lambda-reduce (getp pred 'operator)))
	(t nil))
  )

(defun lambda-conversion (pred args &optional (first nil) &aux namestuff) 
  ; this is a recursive routine which can handle simple or complex lambda
  ; expressions - when they are used as the predicate it is sometimes possible
  ; to eliminate the lambda variables and use the arguments of the wff

  (setq namestuff (name-scan pred 'pred))
  (when namestuff 
	(setf (symbol-plist (car namestuff)) nil)
	(norm-pred pred)
	(setq pred (cadr namestuff)))
  ; simple lambda expression
  (cond ((and (is-lambda-pred pred) args)
	 (if (and (or (symbolp (car args)) (is-function-term (car args)))
		  (not (operator-p (car args))))
	     ; try lambda conversion
	     (let* ((lamargs (if first 
				 (append (cdr args) (list (car args))) 
				 args))
		    newprop useargs remargs)
		   (setq newprop (caddr (subst (car lamargs) (cadr pred) pred)))
		   (when (traced-p 'lambda)
		      (print-line " Normalization: Lambda converting " 
				pred " with " "subject " (car lamargs)) )
		   (setq remargs (cdr lamargs))
		   (when remargs
			 (setq useargs (append (list (last1 remargs))
					     (remove (last1 remargs) remargs))))
		   (if (equal (car newprop) 'L) 
		       (if useargs
			   (append (list (car useargs) newprop) (cdr useargs))
			   nil)
		       newprop)
		   )
	     )
	 )
	; complex lambda expression - first part acts as operator with
	; remainder being arguments which may be applied to that operator
	(t
	 (let* ((pargs (append (cdr pred) args)))
	       (lambda-conversion (car pred) pargs)
               ) 	
	 )
	)
  )

; The following routines are called after any quantifiers or sentential
; operators have been accounted for

;
; Creation routines 
;

(defun make-variable (var &optional top &aux (msort (getp var 'MAIN-SORT))
			  (lbda (getp var 'lambda))
			  (skol (getp var 'skol-const))
			  )
  (unless msort
	(remprop var 'MAIN-SORT)
	(remprop var 'SORT) )
  (change-property var 'VARIABLE t)
  (remprop var 'free)
  (remprop var 'result)
  (when lbda (change-property var 'LAMBDA lbda))
  (unless top (push (list var skol) *vars-to-restore*))
  (symbol-plist var)
  )

(defun make-episode (x) (change-property x 'SORT 'episode))

(defun make-propos (x) (change-property x 'SORT 'propos))
 
;; !!!! no checking is done to ensure tht the sort obtained is legal
;; !!!! or that it is attached to a variable or constant (illegal elsewhere)

(defun remove-sort (arg &aux 
			 (alist (unpack arg))
			 (p (position '_ alist)))
  (setq *got-sort* nil)
  (if p
      (let ((s (build-symbol (subseq alist (+ p 1)) 'epilog))
	    (aname (build-symbol (subseq alist 0 p) 'epilog)))
	   (setq *got-sort* t)
	   (setq s (cond ((getp s 'LONG-NAME)) (t s)))
	   ; check sort to see if legal?
	   (change-property aname 'SORT s)
	   (when (getp arg 'VARIABLE) (change-property aname 'VARIABLE t))
	   aname)
      arg)
  )

; is this really necessary?
(defun order-lits (lits)
	(sort lits '< :key #'(lambda (x) 
		    (if (and (equal (getp x 'WFF-TYPE) 'episodic)
			     (equal (getp x 'PRED) '_))
			0 1)))
	)

(defun merge-subs (main-subs new-sub &aux test-sub)
  (if (setq test-sub (assoc (car new-sub) main-subs))
      (append (remove test-sub main-subs) (list new-sub))
      (append main-subs (list new-sub)) )
  )

(defun undo-sub (main-subs old-subs new-sub &aux test-sub)
  (if (setq test-sub (assoc (car new-sub) old-subs))
      (append (remove (assoc (car new-sub) main-subs) main-subs) 
	      (list test-sub))
      main-subs)
  )

(defun initialize-standard-vars ()
  (setq *+lits* nil)
  (setq *-lits* nil)
  (setq *cur-var-matchable* 'X)
  (setq *cur-var-nonmatchable* 'U)
  (setq *cur-varnum-matchable* 0)
  (setq *cur-varnum-nonmatchable* 0)
  (setp 'x 'variable t)
  (setp 'y 'variable t)
  (setp 'z 'variable t)
  (setp 'u 'variable t)
  (setp 'v 'variable t)
  (setp 'w 'variable t)
  )

; build variable name with sort, lambda info, and matchable info

(defun make-varname (sort stand-var varnum quantifier &aux
			  vname)
  (setq vname
	(if (equal varnum 0)
	    stand-var
	    (build-symbol (list stand-var varnum) 'epilog)))
  (when sort
	(setq vname (build-symbol (list (cond ((getp sort 'SHORT-NAME)) (t sort)) 
				'- vname) 'epilog))
	)
  (when (equal quantifier 'L)
	(setq vname (build-symbol (list 'l- vname) 'epilog))
	(setp vname 'LAMBDA t))
  (when (equal quantifier 'WH)
	(setq vname (build-symbol (list 'wh- vname) 'epilog))
	(setq *wh-vars* (append *wh-vars* (list vname)))
	(setp vname 'WH t))
  (when (member quantifier '(E E! THE))
	(setq *e-vars* (append *e-vars* (list vname))))
  (when sort
	(setp vname 'SORT sort)
	(setp vname 'MAIN-SORT sort)
	)
  vname)


(defun new-var (phi matchable &optional quantifier context embed &aux
		    (lbda (equal quantifier 'L))
		    varname exist-unique)
  (setq exist-unique 
	(and (member quantifier '(E E! The)) (not matchable) 
	     (equal context '+) embed))
		    
  (setq varname (cond ((and matchable (not lbda))
		       (make-varname (getp phi 'SORT)
				     *cur-var-matchable*
				     *cur-varnum-matchable* quantifier))
  
		      (exist-unique
		       (let* ((sort (getp phi 'sort))
			      (n (new-symbol (if sort (build-symbol (list 
					 (cond ((getp sort 'short-name)) 
					       (t sort)) '- 'E-C) 'epilog) 
						 'E-C))) )
			    (change-property phi 'skol-const n)
			    n))
		      (t (make-varname (getp phi 'SORT)
				       *cur-var-nonmatchable*
				       *cur-varnum-nonmatchable* quantifier)) 
		      ))
  (push phi *vars-to-fix*)
  (make-variable varname)

  (if (and matchable (not lbda)) (setp varname 'MATCHABLE t))
	
  (unless exist-unique
  (if matchable
      (case *cur-var-matchable*
	    (X (setq *cur-var-matchable* 'Y))
	    (Y (setq *cur-var-matchable* 'Z))
	    (Z (setq *cur-var-matchable* 'X)
	       (setq *cur-varnum-matchable* (+ *cur-varnum-matchable* 1))))
      (case *cur-var-nonmatchable*
	    (U (setq *cur-var-nonmatchable* 'V))
	    (V (setq *cur-var-nonmatchable* 'W))
	    (W (setq *cur-var-nonmatchable* 'U)
	       (setq *cur-varnum-nonmatchable* 
		     (+ *cur-varnum-nonmatchable* 1))))
      ) )

  (if (assoc varname *subs*) (return-from new-var 
			  (new-var phi matchable quantifier context embed))) 
  (setp varname 'MAIN-VARIABLE t)
  (if (equal (getp varname 'sort) 'operator)
      (setp varname 'operator t))
  (if (equal (getp varname 'sort) 'quantifier)
      (setp varname 'quantifier t))
  varname)



 ;*******************************************
 ;*                                         *
 ;*   L O G I C A L   O P E R A T I O N S   *
 ;*                                         *
 ;*******************************************

 (DEFUN DISTR-NEG (PHI)
  ;;
  ;; Distribute all negations occurring within PHI to have
  ;; minimal scope. This involves "passing" NOT through AND, OR,
  ;; NOT, IMPLIES, NEC, POSS, and quantifiers (but not propositional
  ;; attitude operators.) The transformations are accomplished by
  ;; functions attached to the various operators under indicator
  ;; NEGATION.
  ;;
 (COND ((ATOM PHI) PHI)
       ((EQUAL (CAR PHI) 'NOT)
        (COND ((atom (cdr phi))
	       (case (cdr phi) (YES 'NO) (NO 'YES)))
	      ;quantifier negation
	      ((AND (SYMBOLP (CADR PHI)) (GETP (CADR PHI) 'NEGATION))
               (FUNCALL (GETP (CADR PHI) 'NEGATION) (CDR PHI)) )
	      ; predicate negation
              ((AND (SYMBOLP (CADDR PHI)) (GETP (CADDR PHI) 'NEGATION))
               (FUNCALL (GETP (CADDR PHI) 'NEGATION) (CDR PHI)) )
	      ; predicate is number 
	      ((numberp (caddr phi)) ; numeric conditional
	       ; (NOT phi p psi)  becomes  (phi (1-p) psi)
	       ;; TEMP CHANGE - until the probs are handled better in qa
	       (list (distr-neg (cadr phi)) 'and ;(- 1 (caddr phi))
		     (neg (car (cdddr phi))))
	       )
	      ((and (listp (caddr phi))
		    (numberp (car (caddr phi)))
		    (<= (car (caddr phi)) 1.0)) ; numeric conditional
	       ; (NOT phi p psi)  becomes  (phi (1-p) psi)
	       ;; TEMP CHANGE - until the probs are handled better in qa
	       (list (distr-neg (cadr phi)) 'and ;(append (list (- 1 (car (caddr phi)))) (cdr (caddr phi)))
		     (neg (car (cdddr phi))))
	       )
              (T (CONS 'NOT (DISTR-NEG (CDR PHI)))) ))
       (T (MAPCAR 'DISTR-NEG PHI))
 )) ; end of DISTR-NEG


 (DEFUN NEG (PHI) (DISTR-NEG (CONS 'NOT PHI)))

 (PROGN
  ;;
  ;; Put NEGATION functions on property lists of connectives
  ;; and quantifiers.
  ;;
 (SETP 'NOT 'NEGATION
   #'(LAMBDA (X) (DISTR-NEG (CDR X))) )

 (SETP 'NEC 'NEGATION
   #'(LAMBDA (X) (LIST 'POSS (NEG (CADR X)))) )

 (SETP 'POSS 'NEGATION
   #'(LAMBDA (X) (LIST 'NEC (NEG (CADR X)))) )

 (SETP 'A 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier 
		    (LIST 'E (CADR X)
                         (DISTR-NEG (CADDR X))
                         (NEG (CADDDR X)) )
                    (LIST 'E (CADR X) (NEG (CADDR X)))) ))

 (SETP 'E 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'A (CADR X)
                         (DISTR-NEG (CADDR X))
                         (NEG (CADDDR X)) )
                    (LIST 'A (CADR X) (NEG (CADDR X)))) ))

 (SETP 'WH 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'A (CADR X)
                         (DISTR-NEG (CADDR X))
                         (NEG (CADDDR X)) )
                    (LIST 'A (CADR X) (NEG (CADDR X)))) ))

 (SETP 'MOST 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'MOST (CADR X)
                         (DISTR-NEG (CADDR X))
                         (NEG (CADDDR X)) )
                    (LIST 'MOST (CADR X) (NEG (CADDR X)))) ))

 (SETP 'MANY 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'FEW (CADR X)
                         (DISTR-NEG (CADDR X))
                         (DISTR-NEG (CADDDR X)) )
                    (LIST 'FEW (CADR X) (DISTR-NEG (CADDR X)))) ))

 (SETP 'FEW 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'MANY (CADR X)
                         (DISTR-NEG (CADDR X))
                         (DISTR-NEG (CADDDR X)) )
                    (LIST 'MANY (CADR X) (DISTR-NEG (CADDR X)))) ))

  ;; Note - this used to be NO - but there is too much confusion
  ;; between the answer no and the quantifier
 (SETP 'NOne 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
                    (LIST 'E (CADR X)
                         (DISTR-NEG (CADDR X))
                         (DISTR-NEG (CADDDR X)) )
                    (LIST 'E (CADR X) (DISTR-NEG (CADDR X)))) ))

  ; the - non-referential case
 (SETP 'THE 'NEGATION
   #'(LAMBDA (X) (if (cdddr x)
		    (list (list 'the (cadr x) (distr-neg (caddr x))
				(neg (cadddr x))) 'or
			  (distr-neg (list 'not 'E! (cadr x) (caddr x))))
		    (distr-neg (list 'not 'E! (cadr x) (caddr x)))
		    )))

 (SETP 'E! 'NEGATION
   #'(LAMBDA (X) (IF (CDDDR X) ; restricted quantifier
     (list (distr-neg (LIST 'not 'E (CADR X) (CADDR X) (CADDDR X)))  'or
	   (distr-neg (list 'E 'q (list 'q (list 'L (cadr x) (caddr x)))
		 (list (list 'q (list 'L (cadr x) (cadddr x))) 'and 
		       (list 'E 'r (list (list 'r (list 'L (cadr x) (caddr x)))
				    'and '(not q equal r))
			     (list 'r (list 'L (cadr x) (cadddr x))))))))
     (list (distr-neg (list 'not 'E (cadr x) (caddr x))) 'or
	   (distr-neg (list 'E 'q (list 'q (list 'L (cadr x) (caddr x)))
		 (list 'E 'r '(not q equal r)
		       (list 'r (list 'L (cadr x) (caddr x)))))
	   ))
    ) ))

 (SETP 'AND 'NEGATION
   #'(LAMBDA (X) (CONS (NEG (CAR X))
                      (CONS 'OR (MAPCAR 'NEG (CDDR X))) )))

 (SETP 'OR 'NEGATION
   #'(LAMBDA (X) (CONS (NEG (CAR X))
                      (CONS 'AND (MAPCAR 'NEG (CDDR X))) )))

 (SETP 'IMPLIES 'NEGATION
   #'(LAMBDA (X) (LIST (DISTR-NEG (CAR X))
                      'AND (NEG (CADDR X)) )))

 (SETP '~> 'NEGATION
   #'(LAMBDA (X) (LIST (DISTR-NEG (CAR X))
                      '~> (NEG (CADDR X)) )))

 (SETP '<=> 'NEGATION
   #'(LAMBDA (X) (list (LIST (DISTR-NEG (CAR X)) 'AND (NEG (CADDR X))) 'OR
		      (list (neg (car x)) 'and (distr-neg (caddr x)))
		       )))

 ) ; end of NEGATION PROGN

 (MAPCAR #'(LAMBDA (X) (SETP X 'DISTRIBUTIVE T) X)
         '(NEC    IMPLIES  ;<=> 
		  ~> believe  know
           think  infer  tell     report   say
           want   wish   hope     fear ))



 (DEFUN DISTR-OPS (PHI)
  ;;
  ;; Distribute all occurrences of operators marked "DISTRIBUTIVE"
  ;; over conjunctions in PHI (occurring as "right-hand" operand).
  ;; Presumably this includes NEC, propositional attitudes, IMPLIES 
  ;; (?), perhaps OR (this will work only for binary OR), and
  ;; universal quantifier A.
  ;;
 (IF (OR (ATOM PHI) (MEMBER (CADR PHI) '(** _)) )
      PHI
     (TRY-DISTR-OP (MAPCAR 'DISTR-OPS PHI)) ))



 (DEFUN TRY-DISTR-OP (PHI)
  ;;
  ;; If the first or second element of PHI is an operator marked
  ;; "DISTRIBUTIVE", and the last element of PHI is a conjunction,
  ;; distribute the initial elements of PHI over the conjuncts in
  ;; the last element.
  ;;
 (IF (AND (OR (AND (SYMBOLP (CAR PHI))
                   (GETP (CAR PHI) 'DISTRIBUTIVE))
              (AND (SYMBOLP (CADR PHI))
                   (GETP (CADR PHI) 'DISTRIBUTIVE))
              (AND (SYMBOLP (CADDR PHI))
                   (GETP (CADDR PHI) 'DISTRIBUTIVE)) )
          (CONJUNCTIVE (LAST1 PHI)) )
     (DISTR-OP NIL PHI)
      PHI ))

 

 (DEFUN CONJUNCTIVE (PHI) (AND (LISTP PHI) 
                               (EQUAL (CADR PHI) 'AND) ))


 (DEFUN DISTR-OP (LIST PHI &AUX PHI-INV)
  ;;
  ;; If LIST = NIL (the initial value), then PHI is a formula
  ;; with an operator that distributes over conjunction, and a
  ;; last element which is a conjunctive formula. The initial
  ;; elements (i.e., the operator and any extra arguments) of
  ;; PHI are to be distributed over the conjuncts in the last
  ;; element of PHI. This is done by reversing PHI, separating
  ;; off the conjunct as LIST, and then adding the distributive
  ;; operator and extra arguments to one element of LIST at a
  ;; time. (Since the conjuncts of the resultant conjunction are
  ;; reversed, all the conjuncts are again reversed at the end,
  ;; using MAPCAR 'REVERS on the conjunction.)
  ;;
 (COND ((NULL LIST)
        (SETQ PHI-INV (REVERS PHI))
        (DISTR-OP (CAR PHI-INV) (CDR PHI-INV)) )
       (T
        ;; LIST is a conjunction, or remaining part thereof,
        ;; and PHI is the reversed prefix to be distributed
        ;;
        (IF (CDR LIST)                        ; > 1 remaining conjunct?
            (IF (EQUAL (CADR LIST) 'AND)      ; initial conjunction?
                (MAPCAR 'REVERS
                      (CONS (CONS (CAR LIST) PHI)
                            (CONS 'AND (DISTR-OP (CDDR LIST) PHI))))
                (CONS (CONS (CAR LIST) PHI)
                      (DISTR-OP (CDR LIST) PHI)))
            (LIST (CONS (CAR LIST) PHI)) ))   ; one remaining conjunct
 )) ; end DISTR-OP

     
             
 (DEFUN COLLAPSE-CONNECTIVES (PHI &AUX RESULT)
  ;; 
  ;; Collapse nested AND's or OR's into one level.
  ;; Eg: (PHI1 and ((PHI2 or (PHI3 or PHI4)) and (PHI5)))
  ;;        is collapsed into
  ;;	 (PHI1 and (PHI2 or PHI3 PHI4) PHI5).
  ;;
 (COND ((ATOM PHI) PHI)
       ((EQUAL (CAR PHI) 'NOT)
	(CONS 'NOT (COLLAPSE-CONNECTIVES (CDR PHI))) )
       ((QUANTIFIER-p (CAR PHI))
	(append (list (car phi) (cadr phi))
		(mapcar 'collapse-connectives (cddr phi))) )
       ((MEMBER (CADR PHI) '(** * @ _))
	(append (LIST (COLLAPSE-CONNECTIVES (CAR PHI)) (CADR PHI))
		(mapcar 'collapse-connectives (CDDR PHI))))
       ((MEMBER (CADR PHI) '(AND OR))
	(SETQ RESULT (COLLAPSE-CONNECTIVES (CAR PHI)))
	(UNLESS (and (listp result) (EQUAL (CADR RESULT) (CADR PHI)))
		(SETQ RESULT (LIST RESULT (CADR PHI))) )
	(MAPC #'(LAMBDA (X &AUX XRESULT)
		  (SETQ XRESULT (COLLAPSE-CONNECTIVES X))
		  (IF (and (listp xresult) (EQUAL (CADR XRESULT) (CADR PHI)))
		      (SETQ RESULT
			   (APPEND RESULT (CONS (CAR XRESULT)
						(CDDR XRESULT))))
		      (SETQ RESULT
			   (APPEND RESULT (LIST XRESULT))) ))
	      (CDDR PHI))
	 RESULT )
       (T PHI)
 )) 

(defun skolemizable (phi) 
  (if (atom phi)
      (and (symbolp phi) (equal (getp phi 'wff-type) 'quantified)
	   (member (getp phi 'quantifier) '(E E! THE)))
      (member (car phi) '(E E! THE)) )
  )


(defun skolemize (phi &optional (given nil) &aux
		      (unique (member (car phi) '(E! THE)))
		      (var (remove-sort (cadr phi))) 
		      (sort (getp var 'SORT))
		      res
		      (c (new-symbol 
			  (if sort (build-symbol (list (cond ((getp sort 'short-name)) 
						     (t sort)) '- 'C) 'epilog) 
			      'C))) )
  ;;
  ;; When this is called, the input formula is already known to be
  ;; existential. Skolemize the top-level existential variable
  ;; (additional existentials will be skolemized via the LOAD-FORMULA
  ;; cycle). Flag skolem constants as INDEF (since indefinites are
  ;; candidates for merging with other indefinites if one of them
  ;; is not of rank 0, while names are not).
  ;; 
 
  (unless (is-variable var)
	(print-error "Skolemization: " var " cannot be used as a variable ")
	(return-from skolemize nil))
  (when (cddddr phi)
	(print-error "Skolemization: Only two formulas allowed after " (car phi)
		     *nl* " - " phi " is invalid")
	(return-from skolemize nil))
  (unless (or (listp (caddr phi)) (symbolp (caddr phi)))
	(print-error "Skolemization: Invalid formula " (caddr phi) " in " phi)
	(return-from skolemize nil))
  (unless (or (listp (cadddr phi)) (symbolp (cadddr phi)))
	(print-error "Skolemization: Invalid formula " (cadddr phi) " in " phi)
	(return-from skolemize nil))
  (make-variable var given)
  (change-property c 'indef t)
  (change-property c 'var var)
  (when given (push (list var c) *skol-subs*))
  (when given 
	;; formula given by the story?, i.e., the variable
	;; being skolemized is not a standard variable?
	(change-property var 'SKOL-CONST c) 
	)
  (change-property c 'SORT sort)
  (setq res
  (if (cdddr phi)
      (subst c var (list (caddr phi) 'and (cadddr phi)))
      (subst c var (caddr phi))))
  (when unique
        (change-property c 'UNIQUE t)
        (setq res (list res 'and (list 'A 'r 
		      (if (cddr phi) 
			  (subst 'r var  (list (caddr phi) 'and (cadddr phi)))
			  (subst 'r var (caddr phi)))
		      (list 'r 'equal c)))))
  (dolist (v *vars-to-restore*) 
	  (change-property (car v) 'skol-const (cadr v)))
  (setq *vars-to-restore* nil)
  res) ; end SKOLEMIZE



(defun skolemize-recur (phi &optional (given nil) &aux c var sort)
  ;;
  ;; Recursive version of SKOLEMIZE.  Skolemize every existential
  ;; variable within PHI without applying SPLITs.
  ;; For each new constant introduced by skolemization, mark
  ;; whether DEF/INDEF, GIVEN/INFERRED, EPISODIC/NONEPISODIC,
  ;; and EXPLICIT/IMPLICIT for later use in inference merging.
  ;; (This function will eventually replace the above SKOLEMIZE.)
  ;;
 (when (skolemizable phi)
       (setq var (remove-sort (cadr phi))) 
       (setq sort (getp var 'SORT))
       (setq c (new-symbol 
		(if sort (build-symbol (list (cond ((getp sort 'short-name)) 
					   (t sort)) '- 'C) 'epilog)
		    'C)))
       (make-variable var)
       (if (equal (car phi) 'E)
	   (setp c 'INDEF t)
	   (setp c 'DEF t) )
       (cond (given
	      (setp var 'SKOL-CONST c) 
	      (setp c 'VAR var)
	      (setp c 'GIVEN t)
	      (if (getp var 'sort)
	 	  (setp c 'IMPLICIT t)
		  (setp c 'EXPLICIT t)) )
	     (t (setp c 'INFERRED t)
	        (setp c 'IMPLICIT t) ))
       (setp c 'SORT (getp var 'SORT))
       (if (cdddr phi)
	  (setq phi
	    (subst c var 
	      (list (skolemize-recur (caddr phi) given) 'AND
		    (skolemize-recur (cadddr phi) given) )))
	  (setq phi
	    (subst c var (skolemize-recur (caddr phi) given) )) )
       )
  phi ) ; end of SKOLEMIZE-RECUR


(defun skolemize-in-contxt (contxt-name phi &aux c var 
					(res (getp phi 'print))
					(args (getp phi 'arguments)))
  ;;
  ;; Given PHI that is originally embedded in a proposition,
  ;; skolemize all of its existential variables (without SPLIT).
  ;; Indicate that these skolem constants are "not actual" by putting
  ;; a context name (name of the wff that specifies the original
  ;; proposition) on their property list LOGICAL-FORM.
  ;; Note that we do not consider a negative specification of
  ;; a proposition, i.e., we do not allow a formula in the form
  ;; of (NOT PHI _ PROPOSNAME).
  ;;
 (cond ((and (equal (getp phi 'wff-type) 'episodic)
	     (equal (getp phi 'pred) '_)
	     (skolemizable (car args)) )
	(setq res
	     (list (skolemize-in-contxt contxt-name (car args))
		   '_  (cadr args))) )
       ((skolemizable phi)
	(setq var (getp phi 'variable))
	(setq c (new-symbol (if (getp var 'SORT)
				(build-symbol (list (getp var 'SORT) '- 'C) 'epilog) 'C)))
	(if (equal (getp phi 'quantifier) 'E)
	    (setp c 'INDEF t)
	    (setp c 'DEF t))
	(setp c 'SORT (getp var 'SORT))
	(setp c 'logical-form contxt-name)
	(if (and (numberp (getp contxt-name 'RANK))
		 (zerop (getp contxt-name 'RANK)) )
	    (change-property var 'skol-const c) )
	(if (getp phi 'restrict)
	    (setq res
	      (subst c var
		(list (skolemize-in-contxt contxt-name (getp phi 'restrict))
		     'AND
		    (skolemize-in-contxt contxt-name (getp phi 'main)))))
	    (setq res
	      (subst c var
		(skolemize-in-contxt contxt-name (getp phi 'main)))) )
	))
  res ) ; end SKOLEMIZE-IN-CONTXT

; check to see if something is quantified in one part and referenced in another
; (without being requantified there)
(defun check-for-split-quants (phi &aux
				   (args (prop-args phi)))
  (and (is-quantifier (if (listp (car args)) (caar args) nil))
       (member-if #'(lambda (a) (member-if #'(lambda (b) (epi-variable-p b))
					   (prop-args a))) (cdr args)))
  )
(defun check-for-split-quants-sym (phi &aux
				       (args (getp phi 'arguments)))
  (and (equal (getp (car args) 'wff-type) 'quantified)
       (member-if #'(lambda (a) (member-if #'(lambda (b) (epi-variable-p b))
					   (getp a 'arguments)))
		  (cdr args))
       )
  )

(defun splittable (phi &aux pred)
  (if (atom phi)
      (and (symbolp phi) 
	   (or (and (equal (getp phi 'wff-type) 'logical)
		    (equal (getp phi 'pred) 'and)
		    (if *qa-mode* (not (check-for-split-quants-sym phi)) t))
	       (if *split-episodic*
	       (and (equal (getp phi 'wff-type) 'episodic)
		    (member (getp phi 'pred) '(** * @))
		    (splittable (car (getp phi 'arguments)))))
	       ))
      (and (setq pred (prop-pred phi))
	   (or (and (equal pred 'AND) 
		    (if *qa-mode* (not (check-for-split-quants phi)) t))
	       (if *split-episodic*
		   (and (member pred '(** * @)) (splittable (car phi))))))
      )
  )

(defun split (phi &aux 
		  (ep-op (case (prop-pred phi) 
			       ((**  *) '*)
			       (@ '@) ))
		  (args (prop-args phi))
		  (ep-name (if ep-op (cadr args)))
		  res plist)
  ;;
  ;; Create a new formula with a new name for each conjunct of
  ;; the given conjunctive formula.
  ;;
  (dolist (a args)
	  (unless (or (listp a) (symbolp a))
		  (print-error "Skolemization: Invalid formula " a " in " phi)
		  (return-from split nil)) )

  (if ep-op
      (setq plist (prop-args (car args)))
      (setq plist args))
  (dolist (x plist) 
	  ; will never be negated
	  (setq res (append res (list (if ep-op (list x ep-op ep-name) x)))) 
	  )
  res
 ) ; end of SPLIT

(defun reverse-lambda (wff &optional force &aux
			   (args (getp wff 'arguments))
			   (pred (getp wff 'pred))
			   (negp (neg-p wff))
			   )
  (if (member pred '(** * @ _))
      (let ((new (reverse-lambda (car args) force))
	    neww)
	   (when new
		 (setq neww (new-symbol 'wff))
		 (setp neww 'entity-type 'wff)
		 (setp neww 'wff-type 'episodic)
		 (setp neww 'pred pred)
		 (setp neww 'negated negp)
		 (setp neww 'print (append (if negp (list 'not) )
					   (list (print-info new) pred 
					 (print-info (cadr args)))))
		 (setp neww 'prob (getp wff 'prob))
		 (setp neww 'c-prob (getp wff 'c-prob))
		 (setp neww 'rank (getp wff 'rank))
		 (setp neww 'input-index (getp wff 'input-index))
		 (setp neww 'arguments (append (list new) (cdr args)))
		 neww))
      ; maybe only do for "action" predicates?
      (unless (or (null pred) 
		  (member pred '(and or implies ~> <=> cause-of because))
		  (numberp pred)
		  (if force nil (type-pred-p pred))
		  (if force nil (< (list-length args) 2) ))
      (let* ((s (find-sort (car args)))
	     (v (make-varname (if (member s *sorts*) s nil)
		 *cur-var-matchable* *cur-varnum-matchable* 'L))
	    newpred neww)
	   (setp v 'VARIABLE t)
	   (if (equal (getp v 'sort) 'operator)
	       (setp v 'operator t))
	   (setq *new-vars* nil)
	   (setq newpred (norm-pred (list 'L v 
			  (substitute-prop wff (list (list (car args) v))))))
	   (setq *new-vars* t)
	   (setq neww (new-symbol 'wff))
	   (setp neww 'negated negp)
	   (setp neww 'entity-type 'wff)
	   (setp neww 'pred newpred)
	   (setp neww 'wff-type (getp wff 'wff-type))
	   (setp neww 'arguments (list (car args)))
	   (setp neww 'prob (getp wff 'prob))
	   (setp neww 'c-prob (getp wff 'c-prob))
	   (setp neww 'rank (getp wff 'rank))
	   (setp neww 'print (append (if negp (list 'not) )
	     (list (print-info (car args)) (print-info newpred))))
	   (setp neww 'input-index (getp wff 'input-index))
	   neww))
      )
  )

; returns either nil (no simplification), an atom (the simplification of the
; whole thing), or a list consisting of the restriction and main clause parts
(defun eval-conditional (rest rans main mans &optional equiv non-stand) 
  (when (or rans mans)
	(cond ((or (equal rans 'inconsistent)
		   (equal mans 'inconsistent))
	       'inconsistent)
	      ((equal rans 'YES)
	       (cond ((equal mans 'YES) 'YES)
		     ((equal mans 'NO) 'inconsistent)
		     (t (list nil (if mans mans main))) ))
	      ((equal rans 'NO)
	       (cond ((equal mans 'YES) 'inconsistent)
		     ((equal mans 'NO) 'inconsistent)
		     (t (if (and equiv (not *simplification-schemas*))
			    (list nil (negate-atom main))
			    'inconsistent)) ))
	      ((equal mans 'YES) 
	       (if (or (and equiv (not *simplification-schemas*)) non-stand) 
		   (list nil rest) 'YES))
	      ((equal mans 'NO) 
	       (if (and equiv *simplification-schemas*) 'inconsistent
		   (list nil (negate-atom rest))) )
	      ))
  )

  
(defun Oeval-conditional (rest rans main mans) 
  (when (or rans mans)
	(cond ((equal rans 'YES)
	       ; could infer main then
	       (cond ((equal mans 'YES) 'YES) 
		     ((equal mans 'NO) 'NO)
		     (t (list nil (if mans mans main))) ))
	      ((equal rans 'NO)
	       (cond ((equal mans 'YES)
		      (if (equal main 'YES) 'YES)) ;inconsistent
		     ((equal mans 'NO) 'NO)
		     (t 'INCONSISTENT)) )
	      ((equal mans 'YES)
	       (if *qa-mode* 
		   (list nil (if rans rans rest))
		   (if (equal main 'YES) 'YES) ) )
	      ((equal mans 'NO) 
	       (list nil (negate-atom (if rans rans rest))))
	      (rans (list (if rans rans rest) (if mans mans main))) )
	)
  )

(defun negate-atom (phi)
  (cond ((equal phi 'YES) 'NO)
	((equal phi 'NO) 'YES)
	(t (norm-wff (negate phi) '+)) )  ; normalize '+ nil nil
  )

(defun negate (phi)
  (cond ((equal phi 'YES) 'NO)
	((equal phi 'NO) 'YES)
	((atom phi)
	 (if (symbolp phi)
	     (negate (getp phi 'print))))
	(t (distr-neg (if (equal (car phi) 'not) (cdr phi) (cons 'not phi))))
	)
  )

(defun eval-logical (arguments pred)
  (cond ((member 'inconsistent arguments)
	 'inconsistent)
	((or (and (member 'NO arguments) (equal pred 'AND))
	     (and (not (find-if-not #'(lambda (a) (equal a 'NO)) arguments)) 
		  (equal pred 'OR)) )
	 'NO)
	((or (and (member 'YES arguments) (equal pred 'OR))
	     (and (not (find-if-not #'(lambda (a) (equal a 'YES)) arguments)) 
		  (equal pred 'AND)) )
	 'YES)
	((equal pred 'OR)
	 (remove 'NO arguments))
	((equal pred 'AND)
	 (remove 'YES arguments))
	)
  )

(defun eval-quantified (rest rans main mans)
  (when (or rans mans)
  (unless rest
	  (setq rest 'YES)
	  (setq rans 'YES))
  (cond ((or (equal rans 'inconsistent) (equal mans 'inconsistent))
	 'inconsistent)
	((equal rans 'YES)
	 (cond ((equal mans 'YES) 'YES)
	       ((equal mans 'NO) 'NO)
	       (t (list nil (if mans mans main)))))
	((equal rans 'NO) 'NO)
	((equal mans 'YES) (if (and rest (not (equal rest 'YES)))
				    (list nil rest) 'YES))
	((equal mans 'NO) 'NO)
	(t (list (if rans rans rest) (if mans mans main)))
	))
  )

(defun new-record (contents &optional sort)
  (if sort
      (normalize-term
       (append (list (car *record-indicators*) (list 'quote sort)) contents))
      (normalize-term
       (append (list (car *record-indicators*) 'term) contents)))
  )

