;       Last change    August 1993
(in-package equal)  

; the following routines are used to determine whether two literals are 
; resolvable or factorable


; classify-lits classifies the given literals, and returns a list
; of substitution orders and which literal to enter.
; Each item in the list is of the form (orderlist . number of lit to enter)
; Each item in orderlist is of the form 
;     (element number from 1st lit . element number from second)

(defun classify-lits ()
  ; each literal has only 2 args which can be unified
  ; only two possible unifications - same direction and reversed
 (list '((1 . 1)(2 . 2)) '((1 . 2)(2 . 1)))
  )


; test-residue does the necessary substitutions, and then resolves or
; factors the resulting literals.

(defun test-residue (incompatible net lit-1 lit-2 ordlist
			     neg-p1 reln1 arglist1
			     neg-p2 reln2 arglist2 effort &aux
			     (newargs1 arglist1)
			     (newargs2 arglist2)
			     (eval1 'unknown)
			     (eval2 'unknown)
			     newlit1 newlit2
	                     ulist subs1 subs2)
  (unless (null ordlist)
	  ; unify the two literals according to specification given
          (setq ulist (unify-lists arglist1 arglist2 ordlist nil))

	  ; make substitutions in the two literals
          (unless (null ulist)
                  (setq subs1 (car ulist))
	          (setq subs2 (cdr ulist))
	          (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	          (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
		  ))

  (unless (or (null newlit1)
	      (null newlit2))
          
	  ; evaluate the arguments 
	  (setq newargs1 (spec-eval-args newlit1))
	  (setq newargs2 (spec-eval-args newlit2))

	  ; evaluate the literals
	  (setq eval1 (eval-with-neg net neg-p1 reln1 newargs1 effort))
	  (setq eval2 (eval-with-neg net neg-p2 reln2 newargs2 effort))

	  ; make sure substitution list complete
          (when (equal eval1 'unknown)
                (setq subs1 (union subs1 
		                (arg-sub-list newargs1 
						   (getp newlit1 'arguments) 1)))
	        (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	        )
          (when (equal eval2 'unknown)
                (setq subs2 (union subs2 
		              (arg-sub-list newargs2 
						 (getp newlit2 'arguments) 2)))
	        (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
	        )

	  ; then use evaluations if possible

              (incompatible-with-eval lit-1 lit-2 eval1 eval2 subs1 subs2 
		                        (traced-p 'epilog::equality-test)
		                        "    Equality Specialist:"
					incompatible)
	  )
  )

; unify and test for residues when both literals are subset-of literals

; incompatible-lits tries to determine whether lit1 and lit2 are resolvable
; it tries all possible substitutions to find out

(defun incompatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort &aux
		           (net (determine-net netname))
			   (all-constants (not (or
					   (non-constant-arg arglist1)
					   (non-constant-arg arglist2))))
			   )

  ; see if arguments unifiable, and in what order

  (if (traced-p 'epilog::equality-test)
      (print-line "    Equality Specialist: Trying to determine incompatibility of " 
		  (print-info lit1) " and " (print-info lit2))
      )

  (if all-constants
      (let ((eval1 (eval-with-neg net neg-p1 pred1 arglist1 effort))
	    (eval2 (eval-with-neg net neg-p2 pred2 arglist2 effort)))
	   ; try evaluations first
	   ;unless
	    (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil
				      (traced-p 'epilog::equality-test)
				      "    Equality Specialist:"))
		   ; if no result there - see if a membership test - otherwise
		   ; cnanot do anything
		   ;cond 
		   ))
      ; if not constants
      (let* ((ord-entry (classify-lits)))
	    (if (null ord-entry)
	        nil
		(let (r)
		     (mapcar #'(lambda (o)
				       (setq r (append r
						       (test-residue t
					 	        net lit1 lit2 o 
						        neg-p1 pred1 arglist1
					                neg-p2 pred2 arglist2 
						        effort))))
				      ord-entry)
		     r)
		)
	    ))
  )

; compatible-lits tries to determine whether lit1 and lit2 are factorable.
; it tries all possible substitutions to do this.

(defun compatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort &aux
		           (net (determine-net netname))
	                   (all-constants (not (or 
						(non-constant-arg arglist1)
				                (non-constant-arg arglist2)
						)))
			   )

  ; see if arguments unifiable, and in what order

  (when (traced-p 'epilog::equality-test)
        (print-line "    Equality Specialist: Trying to determine compatibility of " 
		  (print-info lit1) " and " (print-info lit2))
	(if (not (equal (episet-subnet-identifier net) '/))    
	    (print-it "    in subnet " (episet-subnet-identifier net)))
      )

  (if all-constants
      (let ((eval1 (eval-with-neg net neg-p1 pred1 arglist1 effort))
	    (eval2 (eval-with-neg net neg-p2 pred2 arglist2 effort)))
	   (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil 
		                        (traced-p 'epilog::equality-test)
		                        "    Equality Specialist:"
					nil))
		   ;cond 
		   ))

      (let* ((ord-entry (classify-lits)))
	    (if (null ord-entry)
	        nil
		(let (r)
		     (mapcar #'(lambda (o)
				       (setq r (append r
						       (test-residue nil
					 	        net lit1 lit2 o 
						        neg-p1 pred1 arglist1
					                neg-p2 pred2 arglist2 
						        effort))))
				      ord-entry)
		     r)
		)
	    ))
  )
