;; February 20 2005: modified the getp macro to test if the input
;; argument is a symbol. Thanks to Alexander Yeh to report this
;; problem. FM.

;; January 2005: patch to the function check-online-status to work
;; also in windows. FM.

;;       Last change    August 1993
(in-package lib)

 ;;;
 ;;; The following macros are to set and get values
 ;;; to/from Property lists, Arrays, and Hash tables.

(DEFMACRO SETP (X PROP VAL) `(SETF (GET ,X ,PROP) ,VAL))

(DEFMACRO SETA (ARRAY IND VAL) `(SETF (AREF ,ARRAY ,IND) ,VAL))

(DEFMACRO SETH (KEY HASH-TABLE VAL) `(SETF (GETHASH ,KEY ,HASH-TABLE) ,VAL))

;; Previous version of the getp macro
;;(DEFMACRO GETP (X PROP) `(GET ,X ,PROP))
;; new version of the same macro.
(DEFMACRO GETP (X PROP)
  `(AND
    (OR
     (SYMBOLP ,X)
     (PRINT-LINE
      "**ATTENTION: GETP automatically returned NIL because the argument \'" ,X "\' is not a symbol.")
     )
    (GET ,X ,PROP)
    )
  )

(DEFMACRO GETA (ARRAY IND) `(AREF ,ARRAY ,IND))

(DEFMACRO GETH (KEY HASH-TABLE) `(GETHASH ,KEY ,HASH-TABLE))


;; Symbol creation and manipulation routines
;; PACK and UNPACK are th most useful
;; PACK creates a new symbol from a list of symbols
;; UNPACK splits a symbol name into a list of symbols

(DEFUN FIX-NUMBERS (SLIST &aux n)
  (MAPCAR #'(LAMBDA (X)
		    (COND ((and (symbolp x)
				(setq n (symbol-name x))
				(equal (array-dimensions n) '(1)))
			   (cond ((DIGIT-CHAR-P (COERCE X 'CHARACTER)))
				 (t x)))
			  (T X)))
	  SLIST))

(DEFUN PACK (SYMBOL-LIST)
  (INTERN (APPLY #'CONCATENATE
		 (CONS 'STRING (MAPCAR #'PRIN1-TO-STRING 
				       (FIX-NUMBERS SYMBOL-LIST))))))

(DEFUN PACK* (&REST SYMBOL-LIST)
  (PACK SYMBOL-LIST))

(DEFUN UNPACK (SYM &OPTIONAL (FLG NIL)) 
  (MAP 'LIST #'(LAMBDA (CH) (INTERN (STRING CH)))
       (COND (FLG (PRINC-TO-STRING SYM)) (T (PRIN1-TO-STRING SYM)))))

(DEFUN NCHARS (SYM &OPTIONAL (FLG NIL))
  (LENGTH (COND (FLG (PRINC-TO-STRING SYM))
		(T (PRIN1-TO-STRING SYM)))))

(DEFUN NTHCHAR (SYM N &OPTIONAL (FLG NIL))
  (SETQ SYM (COND (FLG (PRINC-TO-STRING SYM)) (T (PRIN1-TO-STRING SYM))))
  (COND ((ZEROP N) NIL)
	((> N 0) (AND (<= N (LENGTH SYM))
		      (INTERN (STRING (ELT SYM (1- N))))))
	(T (SETQ N (+ (LENGTH SYM) N))
	   (AND (>= N 0) (INTERN (STRING (ELT SYM N)))))))

(DEFUN PACKC (CHAR-CODES)
  (INTERN (MAP 'STRING #'CHARACTER CHAR-CODES)))

(DEFUN CHCON (SYM &OPTIONAL (FLG NIL))
  (MAP 'LIST #'CHAR-CODE
       (COND (FLG (PRINC-TO-STRING SYM)) (T (PRIN1-TO-STRING SYM)))))

(DEFUN CHCON1 (SYM)
  (CHAR-CODE (ELT (STRING SYM) 0)))

(DEFUN CHARACTER-ATOM (CH-CODE) (INTERN (STRING (CHARACTER CH-CODE))))

(DEFUN FILTER1 (PRED LIST)
  ;;
  ;; Returns a list consisting of the first element of LIST
  ;; which satisfies PRED followed by the remaining elements
  ;; of LIST. If none satisfy PRED, the result is NIL.
  ;;
  (COND ((NULL LIST) NIL)
	((FUNCALL PRED (CAR LIST)) LIST)
	(T (FILTER1 PRED (CDR LIST))) ))


(DEFUN FILTER (PRED LIST)
  ;;
  ;; Returns the elements of LIST satisfying PRED
  ;;
  (COND ((NULL LIST) NIL)
	((FUNCALL PRED (CAR LIST))
	 (CONS (CAR LIST) (FILTER PRED (CDR LIST))))
	(T (FILTER PRED (CDR LIST))) ))


(DEFMACRO LAST1 (X) `(CAR (LAST ,X)))

(DEFUN REVERS (X) (IF (ATOM X) X (REVERSE X)))


;; strip-pkg explodes the item and strips off the package name if there is one
(defun strip-pkg (item &aux
		       (ilist (unpack item))
		       (p (position (intern ":") ilist :from-end t)))
					;'|:| 
  (if (null p)
      ilist
    (subseq ilist (+ p 1)))
  )

					; squish concatenates several strings together
(defun squish (s) (apply #'concatenate (cons 'string s)))

(defun build-symbol (slist &optional package)
  (if package
      (intern (squish (mapcar 'princ-to-string slist)) package)
    (pack slist))
  )

(defun average (&rest items)
  (if items (float (/ (apply '+ items) (length items))) 0) )

					; non-destructive rplacd
(defun nondest-rplacd (old item new &aux nl (p (assoc item old)))
  (setq nl (remove p old :test 'equal))
  (setq nl (append nl (list (cons item new))))
  nl)

					;check-online-status returns t if this is an interactive process, nil otherwise
					;note: this returns nil in a lisp run via the emacs-lisp interface.
(defun check-online-status (&aux result)
  #+lucid (multiple-value-bind (x y z q) 
	      (run-unix-program "tty" :arguments "-s")
	    (if (equal z 0) (setq result t)))
  #+allegro (multiple-value-bind (x y z q) 
		;; see "man tty" on a unix machine
		(ignore-errors (excl:run-shell-command "tty -s"))
	      (declare (ignore y z q))
	      
	      ;; 0 means that tty (and so lisp) is executed from a
	      ;; real shell (not x) is true if the run-shell-command
	      ;; doesn't exist. this is a short cut to know that it is
	      ;; run from windows (find a better solution). FM
	      (if (or (not x) (equal x 0)) (setq result t)))
  result)

