;       Last change    August 1993
(in-package lib)


; Each element of table consists of a list of dotted pairs - the first part is
; the key and the second the atom name associated with it

(defstruct element-table
  (name nil)			; name of table
  (size 17)			; current size
  (size-index 0)		; index into size table
  (rehash-threshold .8)		; when to rehash
  (current-rehash 13)		; number of load elements to rehash after
  (load 0)			; number of entries
  (hash-func nil)		; hash function for entries
  (test-func nil)		; equality function for entries
  (vector (make-array 17 :initial-element nil)	; the table itself
	  :type simple-vector)
  )


(defconstant *hash-table-sizes* 
  '#(17 31 67 131 257 521 1031 2053 4099 8209 16411 32771 65539))

; use sxhash
; store name in location
;  e.g. (lrrh happy)  => wff3
;       (K wolf)      => term6

(defun put-element (element name table &optional temp &aux index)
  ; if array is too full, expand it first
  (if (> (element-table-load table)
	 (element-table-current-rehash table))
      (expand-element-table table))

  ; puts name into the array where element hashes to
  ; assumes that element is not already there
  (if temp
      (setf (element-table-load table) (+ (element-table-load table) 1))
      (change-field 'element-table-load table
		    (+ (element-table-load table) 1)) )
  (setq index (mod (funcall (element-table-hash-func table) element) 
		   (element-table-size table)))
  (if temp
      (seta (element-table-vector table) index 
	    (append (list (cons element name))
		    (aref (element-table-vector table) index)))
      (change-array (element-table-vector table) index
		    (append (list (cons element name))
			    (aref (element-table-vector table) index))) )
  )

(defun get-element (element table)
  ; looks to see if element has a corresponding name in the array
  ; if so, the name is returned, otherwise nil is
  (cdr (assoc element (aref (element-table-vector table)
	  (mod (funcall (element-table-hash-func table) element) 
	       (element-table-size table)))
	      :test (element-table-test-func table)))
  )

; does not use rewind stuff - this is a "mini rewind"
(defun delete-element (element table &aux index
			       item)
  (setq index (mod (funcall (element-table-hash-func table) element) 
		   (element-table-size table)))
  (setq item (aref (element-table-vector table) index))
  (setf (aref (element-table-vector table) index)
	(remove item (aref (element-table-vector table) index)
		:test 'equal :count 1))
  (decf (element-table-load table))
  item)

(defun initialize-element-table (name)
  (make-element-table :hash-func #'sxhash
		      :test-func #'equal
		      :name name
		      ))

(defun expand-element-table (table)
  (cond ((>= (element-table-size-index table)
	     (length *hash-table-sizes*))
	 (if (>= (element-table-load table)
		 (element-table-size table))
	     (let ()
		  (print-line "** WARNING: No more room in hash table "
				 (element-table-name table))
		  )
	     (let ()
		  (print-line "** WARNING: Hash table " 
				 (element-table-name table)
				 " has exceeded its"
				 " capacity for efficient hashing")
		  (setf (element-table-current-rehash table)
			(element-table-size table))
		  )))
	(t (let ((oldtable (element-table-vector table)))
		(setf (element-table-vector table)
		      (make-array
		       (list (setf (element-table-size table)
				   (aref *hash-table-sizes*
					 (incf (element-table-size-index
						table)))))
		       :initial-element nil))
		(setf (element-table-current-rehash table)
		       (* (element-table-rehash-threshold table)
			  (element-table-size table)))
		(change-field 'element-table-load table 0)
		(map nil
		     #'(lambda (entry)
			       (mapc #'(lambda (value)
				    (put-element (car value) (cdr value) table))
				     entry))
		     oldtable)))
	))


