;       Last change    August 1993
(in-package user)
(eval-when (compile load eval)   
(provide 'lib)
(defpackage lib 
  (:use user lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl
	#+allegro cltl1)
  (:export
	  ;; General Routines
	  |:|			; package separation symbol for strip-pkg
	  setp 			; set a symbol's property
	  getp 			; get a symbol's property
	  seta 			; set array element
	  geta 			; get array element
	  seth 			; set hash table entry
	  geth 			; get hash table entry
	  pack 			; build a new symbol from a list
	  unpack		; split a symbol into individual characters
	  nthchar 		; returns the nth character in a symbol name
	  character-atom	; character corresponding to char code
	  strip-pkg		; same as unpack with package name removed
	  build-symbol		; creates a symbol from a list of things
	  filter1 		; first element of list satisfying filter
	  filter		; all elements of list satisfying filter
	  last1 		; returns last element of list
	  revers 		; reverses list
	  nondest-rplacd	; non-destructive rplacd
	  squish 		; squish several strings into one
	  average		; calculate the average of a list of numbers
	  check-online-status	; determine whether process is interactive

	  ;; Printing Routines
	  *nl*  		; constant indicating "start new line"
	  *print* 		; flag indicating whether or not to print
	  *warn* 		; flag indicating whether to print warnings
	  *indent* 		; current indent level
	  *spaces*		; current number of spaces for indenting
	  indent		; print line with n spaces
	  depth-indent 		; indent twice n spaces
	  set-indent 		; set indent to n
	  reset-indent 		; set indent and spaces back to given values
	  print-line 		; print, starting on a new line
	  print-it 		; print, continuing on current line
	  print-error 		; print error with given list
	  fatal-error 		; print error message and throw to given spot
	  warn-error		; print warning

	  ;; Help
          *help-path*		; path to use to find help files
	  help			; retrieves help info from file

	  ;; display
	  *display-default*     ; default display item if none given
	  -full -f -brief -b -key -k    ; flags accepted by display
	  set-display-function 	; set up display function
	  print-display		; prints the display options
	  display		; display requested information
	  options		; display option which prints the options
	  ?                     ; display help

	  ;; rewind
	  change-array 		; change array entry, saving on rewind list
	  change-hash 		; change hash table entry, saving on rewind list
	  change-field 		; change field value, saving on rewind list
	  change-var 		; change variable value, saving on rewind list
	  change-property 	; change property value, saving on rewind list
	  increment 		; increment variable, saving on rewind list
	  checkpoint 		; set up named or anonymous checkpoints
	  retract 		; retract to named, or back n checkpoints
	  checkpointing-p	; t if checkpointing is on
	  display-ckpts 	; display current checkpoints
	  anon-checkpoint 	; start anonymous (numbered) checkpoint
	  start-checkpoint 	; set up checkpoint
	  get-last-checkpoint	; returns name of last checkpoint set
	  checkpoint-question	; sets a checkpoint and stores its name 
	  rewind-question	; rewinds past the stored question checkpoint
	  rewind-all		; rewinds past all checkpoints

	  ;; trace
	  *traced-items* 	; items currently being traced
	  *traceable-items* 	; items which may be traced
	  traced-p 		; tests whether an item is being traced
	  traceable 		; sets up an item which may be traced
	  trace-item 		; starts tracing an item 
	  untrace-item 		; stops tracing an item
	  trace-all 		; traces all known items 
	  untrace-all		; stops tracing all known items
	  print-traceables	; prints a description of each traceable item

	  ;; tweak
          boolean               ; possible types of tweakable 
          symbol-list           ;   parameters
          character-list
	  *tweakable-items*	; parameters which may be tweaked
	  print-tweakables 	; prints a description of each tweakable item
	  tweak 		; changes the value of a parameter
	  tweakable		; sets up a tweakable parameter

	  ;;hash
	  initialize-element-table 	; starts a hash table
	  get-element 			; gets an element from a hash table
	  put-element			; puts something in a hash table
	  delete-element		; deletes something from a hash table
	  element-table 		; element table fields 
	  element-table-load 
	  element-table-size 
	  element-table-vector 
	  element-table-size-index 
	  element-table-rehash-threshold 
	  element-table-current-rehash
	  ))
(unless (find-package 'lib) (make-package 'lib)))
(in-package lib)

(load "epilog:lib;general")
(load "epilog:lib;io")
(load "epilog:lib;display")
(load "epilog:lib;help")
(load "epilog:lib;rewind")
(load "epilog:lib;hash")
(load "epilog:lib;trace")
(load "epilog:lib;tweak")

