;       Last change    August 1993
(in-package other)

(tweakable 'epilog::*interpret-nil-as-no* nil 
  "Indicates how to interpret null evaluations from other routines"
  'boolean)

; Evaluation routines

; evaluate determines if the given literal is 	true, false or unknown

(defun evaluate (netname lit negp pred arglist effort &aux
			 (res 'unknown)
			 rtn p)
  (declare (ignore effort lit))
  (if (and (not (non-constant-arg arglist))
	   (equal netname '/)
	   (symbolp pred)
	   (equal (pred-type pred) 'constant)
	   )
      (if (and (setq p (getp pred 'package)) 
	       (setq rtn (find-symbol (symbol-name pred) p)))
	  (setq res (call-other-function negp pred arglist rtn))
	  )
      )
  res
  )

; call-other-function calls the function in the package given for the
; predicate.
; Note the flag *interpret-nil-as-no* 
; Depending on the routines actually being called, they may return nil
; just as non-true - i.e. unknown or false
; Because the evaluation is called on entry as well, we don't want a
; contradiction found because no information was stored yet!
; This is especially true for functions that put properties on objects
; Just becasue the property isn't there yet, we don't want a contradiction!
; If however, the routines being called return unknown for unknown, and 
; no or nil for false, then the *interpret-nil-as-no* can be set to t
; without any danger

(defun call-other-function (negp pred arglist rtn &aux res)
  (setq res (apply rtn arglist))
  (if (or (equal res 'unknown)
	  (and (not epilog::*interpret-nil-as-no*) (null res)))
      (return-from call-other-function 'unknown))

  (setq res (find-symbol (symbol-name res) 'epilog))

  (cond ((member res '(no yes unknown)) )
	(res (setq res 'yes))
	(t (setq res 'no)))

  (if negp (setq res (cond ((equal res 'YES) 'NO) 
			   ((equal res 'NO) 'YES) 
			   (t res))))
  (if (traced-p 'epilog::other-eval)
      (print-line "    Other Specialist: Evaluated " 
		     (if negp "not" "") " "
		     (car arglist) " " pred " " (cond ((cdr arglist)) (t " "))
		     " to " res))
  res)

