;       Last change    August 1993
(in-package part)


(tweakable 'epilog::*part-assert* t 
  "Indicates whether allowed to make assertions" 'boolean)

; enter enters the literal lit after first ensuring that it is in the
; part specialists domain and that it is not negated.

(defun enter (netname lit negp pred arglist &aux
		      (net (determine-net netname))
		      (a1 (car arglist))
		      (a2 (cadr arglist))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (when (and (not negp)
	     (not (non-constant-arg arglist))
	     (null a3)
	     (if (equal pred 'equal)
		 (or (part-item net a1) (part-item net a2))
		 t)
	     )
	(when (traced-p 'epilog::part-entry)
	      (print-line "    Part Specialist: Entering " (print-info a1) " " 
			(print-info pred) " " (if a2 (print-info a2) ""))
	      (if (not (equal (part-subnet-identifier net) '/))    
		  (print-it "    into subnet " (part-subnet-identifier net)))
	      )
	; if pred equal, only use for eval now, unless parts
	(cond ((equal pred 'part-of) (enter-part net a1 a2))
	      ((equal pred 'equal) (enter-part-equal net a1 a2))
	      (t (enter-part-type-info net a1 pred))
	      )
	)
  )

(defun part-item (net item)
  (or (findpart net item)
      ; check types of item to see if part type
      (member-if #'(lambda (a) (part-pred-p a)) (grab-types item)))
  )

(defun enter-part-equal (net a1 a2 &aux
			     (p1 (findpart net a1))
			     (p2 (findpart net a2))
			     (narray (part-subnet-parts net)))
  (when (traced-p 'epilog::part-entry)
	(print-line "    Part Specialist: Setting " (print-info a1) " " 
			"equal to " (if a2 (print-info a2) "")) )
  (cond ((null p1) (change-hash a1 narray p2))
	((null p2) (change-hash a2 narray p1))
	(t ; copy from p1 to p2, set p1 to point to p2
	   (change-field 'part-collection p2 (or (part-collection p1)
						 (part-collection p2)))
	   (change-field 'part-types p2 
			 (union (part-types p1) (part-types p2) :test 'equal))
	   (change-field 'part-parts p2
			 (union (part-parts p1) (part-parts p2) :test 'equal))
	   (change-field 'part-super-parts p2
			 (union (part-super-parts p1) (part-super-parts p2) 
				:test 'equal))
	   (change-field 'part-owning-entity p2
			 (cond ((part-owning-entity p2)) 
			       (t (part-owning-entity p1))))
	   (change-hash a1 narray p2)
	   ))
  )

(defun enter-part-type-info (net a1 pred &aux nt fn
				 (ppart (getpart net a1)))
  (if (part-set a1) (change-field 'part-collection ppart t))
  (cond ((part-collection ppart)
	 (setq fn (make-fn-arg 'epilog::member-type-of a1))
	 (setq nt (spec-eval-fn fn))
	 (when (and nt (not (equal nt fn)))
	 (when (traced-p 'epilog::part-entry)
	       (print-line "    Part specialist: Adding member part type " 
			   (print-info nt) " to " (print-info a1)))
	 (change-field 'part-types ppart (append (part-types ppart) (list nt)))
	 )
	 )
	(t (when (traced-p 'epilog::part-entry)
		 (print-line "    Part specialist: Adding part type " 
			     (print-info pred) " to " (print-info a1)))
	   (change-field 'part-types ppart (append (part-types ppart) 
						   (list pred)))
	   ))
  )

(defun enter-part (net part entity &aux oe
		       ppart epart
		       (indv (individual entity)))
  ; entity may be an individual entity or a part
  ; part may be a single part of a collection of them
  (when (traced-p 'epilog::part-entry)
	(print-line "    Part specialist: Adding part " (print-info part) " to " (print-info entity)))
  (setq ppart (getpart net part))
  (setq epart (getpart net entity))
  (if (part-set part) (change-field 'part-collection ppart t))
  (unless (or (part-types epart) indv)
	  (change-field 'part-types epart 
			(remove 'entity (grab-types entity))))

  ; ensure owning entity 
  (setq oe (cond ((part-owning-entity epart)) ((part-owning-entity ppart))))

  (unless oe (if indv (setq oe entity)))
  (when oe
	(change-field 'part-owning-entity epart oe)
	(change-field 'part-owning-entity ppart oe)
	(unless indv
		(mapcar #'(lambda (a) (update-owner net a oe 'super)) 
			(part-super-parts epart))
		(mapcar #'(lambda (a) (update-owner net a oe 'sub)) 
			  (part-parts epart))
		)
	(mapcar #'(lambda (a) (update-owner net a oe 'super)) 
		(part-super-parts ppart))
	(mapcar #'(lambda (a) (update-owner net a oe 'sub)) (part-parts ppart))
	(when epilog::*part-assert*
	      (cond ((part-collection ppart)
		     (assert-part-members part oe (part-types ppart)
					  (grab-types oe)))
		    (t (assert-part-types part (part-types ppart)
					  (remove 'entity (grab-types oe)))) ) )
	)

  (change-field 'part-parts epart (union (part-parts epart) (list part)))
  (unless indv
  (change-field 'part-super-parts ppart (union (part-super-parts ppart) 
					       (list entity))) )
  t)

(defun part-set (part)
  (equal (find-sort part) 'set))

(defun assert-part-members (pset oe ptypes otypes &aux
				 pp as usei)
  (dolist (ptype ptypes)
	  (enter-single-valued pset oe ptype otypes)
	(if (setq pp (pred-parts ptype) )
	    (setq usei (cadr pp))
	    (setq usei ptype))
	(dolist (o otypes)
		(setq as (list 'A 'x (list 'x 'member-of pset) 
			       (list 'x (pack (list o '- usei)))))
		(when (traced-p 'epilog::part-assert)
		      (print-line "    Part Specialist: Asserting " as))
		(setq *assertions* (append *assertions* 
					   (make-spec-inference (list as))))
		)
	)
  )
	  
(defun enter-single-valued (pset oe ptype otypes &aux as c)
  ; if cardinality of set = max of that type of part, then assert
  ; that any part of that type must be a member of the set
  (setq c (spec-eval-fn (make-fn-arg 'cardinality-of pset)))
  (when (and c (numberp c) (equal c (part-total ptype otypes)))
	(setq as (list 'A 'x (list (list (list 'x ptype) 'and (list 'x 'part-of oe)) '<=>
		       (list 'x 'member-of pset))) )
	(when (traced-p 'epilog::part-assert)
	      (print-line "    Part Specialist: Asserting " as))
	(setq *assertions* (append *assertions* 
				   (make-spec-inference (list as))))
	)
  )

(defun part-total (ptype otypes &aux total bounds
			 (totals (getp ptype 'total)))
  (dolist (p (apply 'append (mapcar 'find-part-hiers otypes)))
	  (setq bounds (cdr (assoc p totals)))
	  (when bounds (setq total (cadr bounds)))
	  )
  total)

; should this be asserted to epilog? or just kept on type list of part here?
(defun assert-part-types (item itypes otypes &aux pp
			       usei as)
  (dolist (i itypes)
	  ; if pred parts, use last part for itypes
	  (if (setq pp (pred-parts i) )
	      (setq usei (cadr pp))
	      (setq usei i))
	  (unless (member usei *remainder-indicators*)
	  (dolist (o otypes)
		  (setq as (list item (pack (list o '- usei))))
		  (when (traced-p 'epilog::part-assert)
			(print-line "    Part Specialist: Asserting " as))
		(setq *assertions* (append *assertions* 
					(make-spec-inference (list as))))
		  )))
  )
(defun update-owner (net item oe flag &aux
			 (part (findpart net item)))
  (when (and part (not (part-owning-entity part)))
	(change-field 'part-owning-entity part oe)
	(when (equal flag 'super)
	      (mapcar #'(lambda (a) (update-owner net a oe 'super)) 
		      (part-super-parts part)))
	(when (equal flag 'sub)
	      (mapcar #'(lambda (a) (update-owner net a oe 'sub)) 
		      (part-parts part)))
	)
  )

(defun individual (item &aux tps)
  (setq tps (grab-types item))
  (and (not (equal tps '(entity)))
       (member-if #'(lambda (a) (and (type-pred-p a) (not (part-pred-p a)))) tps))
  )

(defun part-pred-p (pred &aux pp)
  (cond ((setq pp (pred-parts pred))
	 (part-pred-p (cadr pp)))
	(t (and (type-pred-p pred)
		(member-if #'(lambda (b) 
				     (or (getp b 'part-type-hier)
					 (and (getp b 'part-hier)
					      (not (equal (getp b 'root)
							  pred)))))
			   (getp pred 'hier))) )
	))



