;       Last change    August 1993
(in-package response)


(defmacro fragment-fragtype (item) `(getp ,item 'frag-type))
(defmacro fragment-name (item) `(getp ,item 'name))
(defmacro fragment-objtype (item) `(getp ,item 'objtype))
(defmacro fragment-head (item) `(get ,item 'head)) ;; was-getp
(defmacro fragment-negated (item) `(getp ,item 'negated))
(defmacro fragment-required (item) `(getp ,item 'required))
(defmacro fragment-advs (item) `(getp ,item 'advs))
(defmacro fragment-advp (item) `(getp ,item 'advp))
(defmacro fragment-pre (item) `(getp ,item 'pre))
(defmacro fragment-conj (item) `(getp ,item 'conj))
(defmacro fragment-parts (item) `(getp ,item 'parts))
(defmacro fragment-subject (item) `(getp ,item 'subject))
(defmacro fragment-type (item) `(getp ,item 'type))
(defmacro fragment-number (item) `(getp ,item 'number))
(defmacro fragment-quantifier (item) `(getp ,item 'quantifier))
(defmacro fragment-determiner (item) `(getp ,item 'determiner))
(defmacro fragment-pronoun (item) `(getp ,item 'pronoun))
(defmacro fragment-ap (item) `(getp ,item 'ap))
(defmacro fragment-noun (item) `(getp ,item 'noun))
(defmacro fragment-proper-noun (item) `(getp ,item 'proper-noun))
(defmacro fragment-pp (item) `(getp ,item 'pp))
(defmacro fragment-Relc (item) `(getp ,item 'Relc))
(defmacro fragment-S (item) `(getp ,item 'S))
(defmacro fragment-adv (item) `(get ,item 'adv)) ;; was-getp
(defmacro fragment-adj (item) `(getp ,item 'adj))
(defmacro fragment-prep (item) `(getp ,item 'prep))
(defmacro fragment-np (item) `(getp ,item 'np))
(defmacro fragment-vp (item) `(getp ,item 'vp))
(defmacro fragment-verb (item) `(getp ,item 'verb))
(defmacro fragment-tense (item) `(getp ,item 'tense))
(defmacro fragment-aspect (item) `(getp ,item 'aspect))
(defmacro fragment-mood (item) `(getp ,item 'mood))
(defmacro fragment-objects (item) `(getp ,item 'objects))
(defmacro fragment-particles (item) `(getp ,item 'particles))
(defmacro fragment-print (item) `(getp ,item 'print))

(defun print-conj (frag indent)
  (unless (null frag)
	  (print-line (fragment-fragtype frag) " " (fragment-name frag) " "
		      (cond ((fragment-PRE frag)) (t " "))
		      " " (fragment-CONJ frag))
	  (dolist (f (fragment-PARTS frag)) (print-frag f (+ indent 2)))
	  ))

; print-frag prints fragment information with treelike indentation
(defun print-frag (frag &optional (indent 0) &aux
			(os *spaces*)
			(oi *indent*))
  (set-indent indent)
  (unless (null frag)
	  (print-line (symbol-name frag) ": " (fragment-fragtype frag) " " 
		      (fragment-name frag) " "
		      (cond ((fragment-negated frag) "negated ") (t "")) " "
		      (cond ((fragment-print frag)) (t "")) " "
		      (cond ((fragment-number frag)) (t "")) " " 
		      (cond ((fragment-adv frag)) (t "")) " ")
  (when (equal (fragment-head frag) frag)
      (print-line "frag head same as frag"))
	  (case (fragment-fragtype frag)
		(S ; either an S, Sconj, or NP and VP
		   (cond ((fragment-parts frag) (print-conj frag indent))
			 (t (when (fragment-aspect frag)
				  (print-it (fragment-aspect frag)))
			    (when (fragment-tense frag)
				  (print-it " ")
				  (print-it (fragment-tense frag)))
			    (when (fragment-mood frag)
				  (print-it " ")
				  (print-it (fragment-mood frag)))
			    (print-frag (fragment-NP frag) (+ indent 2))
			    (print-frag (fragment-VP frag) (+ indent 2))))
		   (print-frag (fragment-ADVS frag) (+ indent 2))
		   (print-frag (fragment-ADVP frag) (+ indent 2))
		   )
		(NP 
		 (print-it (if (fragment-SUBJECT frag) "subject" "object") " ")
		 (cond ((fragment-pronoun frag)
			(print-it (fragment-pronoun frag)))
		       (t
			(print-it (cond ((fragment-proper-noun frag)) (t ""))
				 " " (cond ((fragment-determiner frag)) (t ""))
				 " " (cond ((fragment-quantifier frag)) (t ""))
				 " " (cond ((fragment-noun frag)) (t "")))
			(print-frag (fragment-AP frag) (+ indent 2))) )
		 (when (fragment-parts frag) (print-conj frag indent ))
		 (when (fragment-PP frag)
		      (print-line "  Prepositional clauses")
		      (print-frag (fragment-PP frag) (+ indent 2)))
		 (when (fragment-RelC frag)
		      (print-line "  Relative clauses")
		      (print-frag (fragment-RelC frag) (+ indent 2)))
		 (when (fragment-S frag) 
			(print-frag (fragment-S frag) (+ indent 2)))
		 )
		(AP
		 (cond ((fragment-parts frag) (print-conj frag indent))
		       (t (print-it (cond ((fragment-adj frag)) (t "")))) ))
		((PP ADVP)
		 (print-it (cond ((fragment-pre frag) ) (t "")) " "
			   (fragment-prep frag))
		 (cond ((fragment-NP frag) 
			(print-frag (fragment-NP frag) (+ indent 2)))
		       ((fragment-S frag) 
			(print-frag (Fragment-S frag) (+ indent 2))))
		 (when (fragment-parts frag) (print-conj frag indent ))
		 )
		((VP RelC)
		 (cond ((fragment-parts frag) 
			(print-it (cond ((fragment-particles frag)) (t "")) " "
				  (cond ((fragment-tense frag)) (t "")) " "
				  (cond ((fragment-aspect frag)) (t "")) " "
				  (cond ((fragment-mood frag)) (t "")))
			(print-conj frag indent) )
		       (t (print-it (fragment-verb frag) " "
				   (cond ((fragment-particles frag)) (t "")) " "
				   (cond ((fragment-tense frag)) (t "")) " "
				   (cond ((fragment-aspect frag) ) (t "")) " "
				   (cond ((fragment-mood frag) ) (t ""))) 
			  ))
		 (dolist (o (fragment-OBJECTS frag))
			 (print-frag o (+ indent 2)))
		 (print-frag (fragment-ADVS frag) (+ indent 2))
		 (print-frag (fragment-ADVP frag) (+ indent 2))
		 )
		)
	  )
  (reset-indent os oi)
  nil)

