;       Last change    August 1993
(in-package response)

(defun pinfo (item)
  (cond ((not (symbolp item)) item)
	((getp item 'print))
	(t item))
  )

; add-determiner adds a determiner to a fragment if it doesn't already have one.
(defun add-determiner (det item)
  (unless (fragment-DETERMINER item)
	  (if (fragment-parts item)
	      (progn
	       (if (and (equal det "any") 
			(equal (fragment-pre item) 'either))
		   (setp item 'pre nil))
	       (add-determiner det 
			(car (fragment-parts item)))
	       (dolist (f (cdr (fragment-parts item)))
		       (if (not (equal (fragment-number f) 'plural))
			   (add-determiner det f)
			   (add-determiner "" f))))
	      (when (fragment-noun item)
		    (setp item 'DETERMINER det)))
	  )
  )

; add-number adds a number indicator to a fragment if it doesn't
; already have one.

(defun add-number (num item &optional force)
  (when item
  (unless (or (if (equal (fragment-fragtype item) 'NP)
		  (equal (fragment-objtype item) 'constant))
	      (fragment-NUMBER item))
          (setp item 'NUMBER num)
	  (mapcar #'(lambda (a) (add-number num a force))
		  (fragment-parts item))
	  (add-number num (fragment-PP item) force)
	  (add-number num (fragment-ADVP item) force)
	  (unless (equal (fragment-fragtype item) 'NP)
		  (add-number num (fragment-S item) force))
	  (add-number num (fragment-RELC item) force)
	  (when (or force
		    (and (equal (fragment-fragtype item) 'PP) 
			 (equal (fragment-prep item) 'of)))
		(add-number num (fragment-NP item) force))
	  (when force
		(mapcar #'(lambda (a) (add-number num a force))	  
			(fragment-objects item))
		(add-number num (fragment-VP item) force) )
	  ))
  )

; add-determiner adds a noun to a fragment if it doesn't already have one.

(defun add-noun (n item)
  (when n
  (unless (and (fragment-NOUN item) 
	       (not (equal (fragment-noun item) 'thing)))
          (setp item 'NOUN n))
  ))

(defun add-adverb (adverb frag)
  (when adverb (setp frag 'adv (pushnew adverb (fragment-adv frag))))
  frag
  )

(defun add-qinfo (frag qinfo &aux np)
  (case (fragment-fragtype frag)
        (NP (setq np frag))
        (PP (setq np (make-frag 'NP (cons 'name (fragment-name frag))
				(cons 'PP frag))))
        (AP (setq np (make-frag 'AP (cons 'name (fragment-name frag))
				(cons 'AP frag))))
        (otherwise (setq np frag)))
  (unless (fragment-quantifier np) (setp np 'quantifier qinfo))
  np)

; should we be also using the varialbe to get +/- context as well?
(defun decide-adverb (quantifier)
  (case quantifier
	(A nil) ;'always
	(most 'usually) 
	(many 'likely) ;'frequently
	(some 'sometimes)
	(few 'occasionally)
	(none 'never)
	(otherwise
	 (if (symbolp quantifier)
	     (if (equal (getp quantifier 'quant-type) 'modified)
		 (quant-adverb quantifier)
		 quantifier)))
	))

(defun quant-adverb (quant &aux q a)
  (setq q (decide-adverb (getp quant 'operator)))
  (setq a (mapcar 'decide-adverb (getp quant 'arguments)))
  (if (listp q) (append q a) (append (list q) a))
  )

(defun decide-adv-qinfo (pred &aux adverb qinfo)
  (cond ((and pred 
	      (or (not (numberp pred))
		  (= pred 1.0)))
	 (setq adverb nil) ;'always
	 (setq qinfo 'A))
	((>= pred 0.75)
	 (setq adverb 'usually) 
	 (setq qinfo 'most))
	((>= pred 0.5)
	 (setq adverb 'likely) ;'frequently
	 (setq qinfo 'many))
	((>= pred 0.25)
	 (setq adverb 'sometimes)
	 (setq qinfo 'some)) 
	((> pred 0.0)
	 (setq adverb 'occasionally)
	 (setq qinfo 'few))
	(t
	 (setq adverb 'never)
	 (setq qinfo 'none))
	)
  (cons adverb qinfo)
  )

(defun adverb-for (pred)
  (if (numberp pred)
      (cond ((= pred 1) nil)
	    ((>= pred 0.75) 'usually) 
	    ((>= pred 0.5) 'likely) ;'frequently
	    ((>= pred 0.25) 'sometimes)
	    ((>= pred 0.0) 'rarely)
	    ((= pred 0) 'never)))
  )     

; set-required checks all subfrags in frag to see if they are mentioned in
; postfrag - if so,  set required to t (they must be filled in)

;; I have commented this out because it's not used, and it contains an
;; error that the CMU compiler chokes on: fragment-name is a macro, not a
;; function, so is not a legal argument for mapcar. -ANK

;(defun set-required (frag postfrag &aux
;                          (allfrags (if (fragment-VP frag)
;                                   (fragment-objects (fragment-VP frag))))
;                          (checklist (mapcar #'fragment-name postfrag)))
;  (dolist (a allfrags)
;          (when (member (fragment-name a) checklist)
;                (setp a 'required t)
;                (if (and (equal (fragment-objtype a) 'variable)
;                         (equal (fragment-fragtype a) 'NP))
;                    (add-determiner "some" a))
;                ))
;  )


