;       Last change    August 1993
(in-package response)

; Organize breaks the clauses up into sentences (for multiple sentence answers).
; Currently, all conditional statements are sentences by themselves (since they
; can't combine with facts anyway), and any facts are assumed to be in one 
; sentence, and will be broken up by the verbalization stage, if necessary.

; Format of returned "logical form"
; outermost, a list of sentences
; for each sentence, a list of wffs
; Example:
;(lrrh girl) (lrrh pretty)      (A x ((x girl) and (x pretty)) (x happy)) 
;( (  wff1        wff2    )   ( wff3 )      )
;     sentence1                   sentence2

; organize takes the given list of clauses and returns a list of sentence
; collections.  Note that the clauses have already been ordered in order of
; likelihood of being known by the filtering stage. Preserve that order as
; much as possible when separating facts from rules.
(defun organize (clauses &aux
			usedclauses yukkies avoid
			nfacts facts rules newf)
  ; returns list of logical form of clauses - one element in list per sentence
  ; assume all in one sentence for now
  ; transform each clause into appropriate list form, and return the whole mess
  ; as one sentence
  (dolist (wffc clauses)
	  ; avoid duplication if a clause used more than once
	  (unless (member wffc usedclauses)
		  (if (getp wffc 'conditional); (getp wffc 'PRINT)
		      (setq rules (append rules (list wffc)))
		      (setq facts (append facts (list wffc))))
		  (pushnew wffc usedclauses)
		  )
	  )
  ; go through facts and remove episodic type literals and anything involving
  ; them
  (dolist (f facts)
	  (when (and (type-lit-p f) 
		     (equal (getp f 'wff-type) 'episodic))
		(setq yukkies (append yukkies (list f)))
		(setq newf (append newf (list (car (getp f 'arguments)))))
		(setq avoid (append avoid (last (getp f 'arguments))))
		))
  (setq nfacts (remove-if #'(lambda (a) 
	   (or (member a yukkies) (intersection (getp a 'arguments) avoid)))
			 facts))
  (setq nfacts (append nfacts newf))
  (if nfacts (setq facts nfacts))
  (append (cond (facts (list facts)) (t nil)) (mapcar 'list rules))
  )
