;       Last change    August 1993
(in-package set-equal)
;        Routines to get and create set nets and sets

; Getnet returns the set net corresponding to the context given. If no
; such net exists, one is created.

(defun getnet (context)
  (let ((context-array (retrieve-specialist-info
			*environment*
			'set-equal
			(lambda ()
			  (make-hash-table :test 'eql
					   :size '100
					   :rehash-size '10
					   :rehash-threshold '.8)))))
    (when (null (geth context context-array))
      (seth context context-array (newnet context)))
    (geth context context-array)))

; determine net gets the set net associated with the subnet for the
; netname given (/concpet1/concept2/...) and sets the current net in
; case any functions are used

(defun determine-net (netname &aux
			      (net nil))
  (if (null netname)
      (setq net (getnet '/))
      (setq net (getnet netname))
      )
  (setq current-net net)
  net
  )

; Getset returns the set for item in the set net. If there is no
; entry, once is created and returned.

(defun getset (net item &aux
		     (narray (episet-subnet-sets net)))
  (when (null (geth item narray))
	(change-hash item narray (make-episet :identifiers (list item)))
	)
  (geth item narray)
  )

; Findset returns the set in the net corresponding to item if there
; is one. If not, nil is returned

(defun findset (net item)
  (let ((result (geth item (episet-subnet-sets net))))
    (if result
	result
      (dolist (e (environment-use-environments *environment*))
	(let* ((*environment* e)
	       (result (findset (determine-net (episet-subnet-identifier net))
				item)))
	  (when result
	    (return result)))))))


(defun writeable-copy (net set &optional (env *environment*))

  "Takes a set-subnet 'net', an episet structure 'set', and optionally
  an environment 'env', which defaults to the value of *environment*.
  If set belongs to env, returns set; otherwise, returns a copy of set
  that belongs to env under net."
  
  (if (eq (episet-environment set) env)
      set
    (let* ((*environment* env)
	   (table (episet-subnet-sets
		   (getnet (episet-subnet-identifier net))))
	   (ids (episet-identifiers set))
	   (newset (copy-episet set)))
      (setf (episet-environment newset) *environment*)
      (mapc #'(lambda (setname)
		(change-hash setname table newset))
	    ids)
      newset)))

; Newnet creates a new set net, including the set table

(defun newnet (ident &aux

		     ; hash table for set

		     (nt (make-hash-table :test 'eql
					  :size '100
					  :rehash-size '10
                                          :rehash-threshold '.8))) ; changed!
                                        ;; threshold value changed from 9
                                        ;; [2/24/95 CH]



  (make-episet-subnet :identifier ident
		   :sets nt)
  )

; get-subnet returns the appropriate set subnet
(defun get-subnet (net)
  (if (null net)
      (let ()
           (if (null current-net)
               (setq current-net (determine-net *current-netname*)))
           current-net
           )
      (determine-net net))
  )

; get-contents returns the list of mebers associated with set arg

(defun get-contents (net arg &aux
			     (name arg)
			     (setitem (findset net name)))
  (if (null setitem)
      nil
    (episet-contents setitem)
    )  
  )

; get-contents-0 is the same as get-contents, excpet that any member which is
; itself a set is replaced by its contents, recursively.

(defun get-contents-0 (net arg &aux
                         (name arg)
                         (setitem (findset net name)))
  (if (null setitem)
      (list arg)
      (get-more-0 net (episet-contents setitem))
      )  
  )

; for every item in list, if it is a set, get ists contents, else use it

(defun get-more-0 (net alist &aux
                           (arg (car alist)))
  (if (null arg)
      nil
      (union (get-contents-0 net arg) (get-more-0 net (cdr alist)))
      )  
  )

; get-set-immed returns the contents of a set specified in a quoted expression

(defun get-set-immed (arg)
  (record-contents arg) 
  )

; get-set-immed-0 is the same as get-set-immed, but for any member of that
; set which is itself a set, its members are returned instead of its name

(defun get-set-immed-0 (net arg)
  (get-more-0 net (record-contents arg)) 
  )

(defun +equal (setmem)
  (equal setmem 'epilog::+)
  )

(defun +member (setmems)
  (member 'epilog::+ setmems :test #'equal)
  )

(defun immed-set (set)
  (and (symbolp set) (equal (getp set 'term-type) 'record)))

(defun equiv-set-name (item &optional new)
  (if new
      (pack (list (print-info item) '- 'equal '_set) )
      (pack (list (print-info item) '- 'equal) ))
  )

(defun non-equal-set-name (item &optional new)
  (if new
      (pack (list (print-info item) '- 'epilog::non-equal '_set) )
      (pack (list (print-info item) '- 'epilog::non-equal) ))
  )

(defun check-equivalences (net a1 a2 &aux
                               (s1name (equiv-set-name a1))
                               (s1 (findset net s1name)))
  (when s1
        (member a2 (episet-contents s1)))
  )

(defun check-non-equal (net a1 a2 &aux
                               (s1name (non-equal-set-name a1))
                               (s1 (findset net s1name)))
  (cond (s1
         (member a2 (episet-contents s1)
                :test #'(lambda (a b) (or (equal a b)
                                          (check-equivalences net a b)))))
        ((setq s1 (findset net (non-equal-set-name a2)))
         (member a1 (episet-contents s1)
                 :test #'(lambda (a b) (or (equal a b)
                                           (check-equivalences net a b)))))
        )
  )

