;       Last change    August 1993
(in-package set)


(tweakable 'epilog::*unique-names-assumption* t
  "Indicates whether or not we are assuming that user assigned names indicate unique individuals"
  'boolean)

; evaluation functions

; evaluate tries to determine whether the  given literal is true, false
; or unknown, after first ensuring that it is in the set specialist's domain.

(defun evaluate (netname lit negp pred arglist effort &aux
		      (net (determine-net netname))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (if (and (null a3)
	   (if (equal pred 'equal) (or (equal (find-sort (car arglist)) 'set)
				       (equal (find-sort (cadr arglist)) 'set))
	       t)
	   (not (non-constant-arg arglist)))
      (eval-with-neg net negp pred arglist effort)
      'unknown)
  )

; eval-with-neg evaluates the given literal, taking into account whether it
; is negated. If so, the positive literal is evaluated, and the result reversed. 
(defun eval-with-neg (net negp reln args effort &aux
			  (posans (evaluate-set net reln args effort)))
    
  (cond ((equal posans 'yes)
         (if negp
	     'no 
	     'yes))
	((equal posans 'no)
	 (if negp
	     'yes
	     'no))
	(t posans))
  )
 
; test-equal tests to see if two arguments are equal

(defun test-equal (net a1 a2 effort &aux
		       (sort1 (find-sort a1))
		       (sort2 (find-sort a2)))
  (cond ((equal a1 a2) 'yes)

	((and (equal sort1 'epilog::set)  ; if both are sets, compare contents
	      (equal sort2 'epilog::set))
	 (test-set-equal net a1 a2 effort))

	; otherwise we don't know
	(t 'unknown)
	)
  )


(defun test-set-equal (net a1 a2 effort &aux
			   s1 s2 mems1 mems2 type1 type2 all1 all2 fin1 fin2 cp)
  (declare (ignore effort))
  (if (traced-p 'epilog::set-eval)
      (print-line "    Set Specialist: Testing to see if set " (print-info a1) 
		  " is equal to set " (print-info a2) " " ))
  (cond ((immed-set a1) (setq mems1 (get-set-immed a1))
	 (setq fin1 (not (+member mems1))))
	 (t (if (setq s1 (findset net a1))
		(setq mems1 (episet-contents s1))
		(return-from test-set-equal 'unknown))
	    (setq type1 (episet-member-type s1))
	    (setq all1 (episet-all s1))
	    (setq fin1 (episet-finished s1))) )
  (cond ((immed-set a2) (setq mems2 (get-set-immed a2))
	 (setq fin2 (not (+member mems2))))
	 (t (if (setq s2 (findset net a2))
		(setq mems2 (episet-contents s2))
		(return-from test-set-equal 'unknown))
	    (setq type2 (episet-member-type s2))
	    (setq all2 (episet-all s2))
	    (setq fin2 (episet-finished s2))) )
	 
  ; if sets point to same set structure, they are equal
  (cond ((and s1 s2 (equal s1 s2))
	 t)

	; if members are identical and both are fully known sets, then true
	((and (if (and type1 type2 
		       (setq cp (spec-compare-preds type1 type2)))
		  (and (equal cp 'epilog::equivalent) (equal all1 all2))
		  t)
	      (null (set-exclusive-or mems1 mems2 :test #'(lambda (a b) 
					   (or (equal a b)
					       (check-equivalences net a b)))))
	      (and fin1 fin2))
	 t)

	; if both are fully known and the members are not the same
	; then they are not equal - otherwise unknown
	((and fin1 fin2
	      ; at least one member not equal to another member, using assump
	      (member-if #'(lambda (a) 
		 (member-if #'(lambda (b) (and (not (equal a b)) 
				 (check-non-equal net a b))) mems2)) mems1))
	 nil)

	; if types are incompatible they can't be equal
	((and s1 s2 type1 type2 (equal cp 'epilog::disjoint)) nil)

	(t 'unknown))
  )

(defun check-members (a mems fin &optional effort &aux
			  final-res)
  (dolist (m (remove '+ mems))
	  (case (eval-with-spec *current-netname* 
				(normalize (list a 'equal m))
				effort)
		(YES (return-from check-members t))
		(NO)
		(unknown (setq final-res 'unknown)))
	  )
  (if (and (not fin) (null final-res)) (setq final-res 'unknown))
  final-res)

; test-member tests to see if a1 is a member of the set a2. If it is
; identical to one of the members, it is, otherwise if the set is
; finished, it is not, otherwise unknown

(defun test-member (net a1 a2 effort zero &aux
			set1 st fin)
  (if (traced-p 'epilog::set-eval)
      (print-line  "    Set Specialist: Testing to see if " (print-info a1) 
		 " is a member of set " (print-info a2))) 
  (cond ((immed-set a2)
	 (setq st (if zero (get-set-immed-0 net a2) (get-set-immed a2)))
	 (setq fin (not (+member st))))
	(t (if (setq set1 (findset net a2))
	       (setq st (if zero (get-contents-0 net a2) 
			     (get-contents net a2)))
	       (return-from test-member 'unknown))
	   (setq fin (episet-finished set1))
	   ))

  (is-a-member a1 set1 st fin effort )
  )

(defun is-a-member (a1 set1 st fin &optional effort &aux 
		       type1 type2 res card-max)
  ; if a1 is a member of set a2, then true
  (cond ((member a1 st :test 'equal) t) ; quick and easy check
	
	; if a1's type is incompatible with the set's member type, then no
	((and set1 (setq type1 (episet-member-type set1))
	      (setq type2 (epilog::grab-types a1))
	      (member-if #'(lambda (a) 
		     (equal (spec-compare-preds type1 a) 'epilog::disjoint)) 
			 type2))
	 nil)

	; try more detailed member check using equality specialist to help
	((and st 
	      (not (equal (setq res (check-members a1 st fin effort )) 
			  'unknown)))
	 res)

	; if cardinality of set is less than 1, can't have any members
	((and (setq card-max (if set1 (episet-cardinality-max set1)
				 (if fin (list-length st))))
	      (numberp card-max)
	      (< card-max 1))
	 nil)

	;otherwise can't tell
	(t 'unknown)
      )
  )
  
; test-subset checks to see if set a1 is a subset of set a2

(defun test-subset (net a1 a2 effort &aux
			set1 set2 st1 st2 fin1 fin2 type1 type2 cp
			res card-min card-max)
  (if (traced-p 'epilog::set-eval)
      (print-line  "    Set Specialist: Testing to see if set " (print-info a1) 
		   " is a subset of set " (print-info a2) ))
  (cond ((immed-set a1)
	 (setq st1 (get-set-immed-0 net a1))
	 (setq fin1 (not (+member st1))))
	(t (if (setq set1 (findset net a1))
	       (setq st1 (get-contents-0 net a1))
	       (return-from test-subset 'unknown))
	   (setq fin1 (episet-finished set1))
	   (setq type1 (episet-member-type set1))
	   ))
  (cond ((immed-set a2)
	 (setq st2 (get-set-immed-0 net a2))
	 (setq fin2 (not (+member st2))))
	(t (if (setq set2 (findset net a2))
	       (setq st2 (get-contents-0 net a2))
	       (return-from test-subset 'unknown))
	   (setq fin2 (episet-finished set2))
	   (setq type2 (episet-member-type set2))
	   ))

  ; check to see if there is a connection of the appropriate type

  ; if a1 is a set name, and is a member of a2, then it is a nested
  ; set, so answer true
  (cond ((equal a1 a2) t)
	
	((and st2 (equal (find-sort a1) 'epilog::set)
	      (member a1 st2 :test 'equal))
	 t)

	((and set1 set2 
	      (subset-conn net set2 (episet-connections set1)))
	 t)

	; see if a1's contents are a subset of a2's contents
	((and st1 st2 fin1 (subsetp st1 st2 :test 'equal))
	 t)

	((and set1 set2 type1 type2
	      (setq cp (spec-compare-preds type1 type2))
	      (member cp '(epilog::equivalent epilog::subsumed))
	      (episet-all set2))
	 t)

	((and set1 set2 type1 type2 cp
	      (or (equal cp 'epilog::disjoint)
		  (and (episet-all set1) (not (episet-all set2)))) )
	 nil)

	; use equality specialist to help
	; if all members of first set are equivalent to those of second, t
	; if a member of first is NOT equal to any member of second, nil
	((and st1 st2 fin1 
	    (not (equal (setq res (bigger-subsetp st1 st2 fin2 effort )) 
			'unknown)))
	 res)

	; if cardinality of set1 > cardinality of set2, then no
	((and (if set1 (setq card-min (episet-cardinality-min set1))
		  (if fin1 (setq card-min (list-length st1))))
	      (if set2 (setq card-max (episet-cardinality-max set2))
		  (if fin2 (setq card-max (list-length set2))))
	      (numberp card-min) (numberp card-max)
	      (> card-min card-max)
	      )
	 nil)

	; otherwise can't tell
	(t 'unknown)
	)
  )

(defun bigger-subsetp (mems1 mems2 fin2 effort &aux
			   final-res m1res)
  ; note, mems1 are all the members in the first set i.e. fin1 is true
  (dolist (m1 mems1)
	  (setq m1res nil)
	  (dolist (m2 mems2)
		  (case (setq m1res 
			      (eval-with-spec *current-netname* 
					      (normalize (list m1 'equal m2))
					      effort))
			(YES (return)) ; stop looking for m1
			)
		  )
	  (cond ((equal m1res 'unknown) (return-from bigger-subsetp 'unknown))
		((equal m1res 'no) 
		 (if fin2 (return-from bigger-subsetp nil)
		     (return-from bigger-subsetp 'unknown)))
		(t ;yes
		   (setq final-res t)))
	  )
  (if final-res final-res 'unknown))


(defun subset-conn (net set-struct conn &aux
			      ss how sets)
  (member-if #'(lambda (a) 
		       (setq how (episet-connect-how a))
		       (setq sets (episet-connect-sets a))
		       (cond ((member how '(subset-of))
			      (and (setq ss (findset net (car sets)))
				       (or (equal ss set-struct)
					   (subset-conn net set-struct (episet-connections ss)))))
			     ((member how '(set-diff))
			      (and (setq ss (findset net (car sets)))
				       (equal ss set-struct)))
			     ((member (episet-connect-how a) '(intersect))
			      (member-if #'(lambda (b) 
					   (equal (findset net b) set-struct))
					 sets))
			     ))
	     conn) )

(defun test-cardinality (net setname card effort &aux
			     s1 min max testmin testmax)
  (declare (ignore effort))
  (if (not (immed-set setname))
      (when (setq s1 (findset net setname))
	    (setq min (episet-cardinality-min s1))
	    (setq max (episet-cardinality-max s1))
	    )
      (setq min (setq max (list-length (get-set-immed setname)))))
  (setq testmin (if (numberp card) card
		    (spec-eval-fn (make-fn-arg 'epilog::min-of card))))
  (setq testmax (if (numberp card) card
		    (spec-eval-fn (make-fn-arg 'epilog::max-of card))))
  (cond ((and (equal min testmin) (equal max testmax))
	 'yes)
	((and (numberp min) (numberp testmin) (<= testmin min)
	      (numberp max) (numberp testmax) (>= testmax max))
	 'yes)
	((or (and (numberp min) (numberp testmin) (> testmin min))
	     (and (numberp max) (numberp testmax) (< testmax max)))
	 'no)
	(t 'unknown))
  )

(defun evaluate-with-op (net op pred arglist effort &aux
			     (s1 (car arglist))
			     as curset)
  (unless (or (null s1) (cdr arglist) (null op) (null pred))
	  (cond ((immed-set s1) 'unknown) ; maybe more?
		((null (setq curset (findset net s1))) 'unknown)
		((member op '(coll plur plural))
		 ; compare member-type of set with pred
		 ; need to have incompatibility for no - should it be no if
		 ; disjoint?
		 (case (spec-compare-preds (episet-member-type curset) pred) 
		       ((equivalent subsumed) 'YES)
		       (disjoint 'NO)
		       (otherwise 'unknown))
		 )
		((and (equal (op-type op) 'modified)
		      (equal (getp op 'operator) 'coll-of))
		 ; compare both member type and cardinality
		 (case (spec-compare-preds (episet-member-type curset) pred)
			    
		       ((equivalent subsumed)
			(if (and (setq as (getp op 'arguments))
				 (null (cdr as))
				 (or (numberp (car as)) (number-arg (car as))))
			    ; check cardinality
			    (test-cardinality net s1 (car as) effort)
			    'unknown) )
		       (subsumes
			(if (and (setq as (getp op 'arguments))
				 (null (cdr as))
				 (or (numberp (car as)) (number-arg (car as))))
			    ; check cardinality
			    (if (equal (test-cardinality net s1 (car as) effort)
				       'NO)
				'NO
				'unknown)
			    'unknown) )
		       (disjoint 'NO)
		       (otherwise 'unknown) ) )
		)
	  )
  )


; evaluate-set tries to evaluate whether the given relation holds with the
; arguments.
; Can test limited equality, and membership.

(defun evaluate-set (net pred arglist effort &aux
			 (reln (find-symbol (symbol-name pred) 'set))
			 (a1 (car arglist))
			 (a2 (cadr arglist))
			 (res 'unknown)
			 op pr)
  (setq res
  (unless (null arglist)
	  (when (traced-p 'epilog::set-eval)
	        (print-line "    Set Specialist: Evaluating " (print-info a1) 
			  " " (print-info pred) " " (if a2 (print-info a2) ""))
		(if (not (equal (episet-subnet-identifier net) '/))    
		    (print-it "    in subnet " (episet-subnet-identifier net)))
		)
	  (if (and (symbolp pred) (equal (getp pred 'pred-type) 'modified)
		   (setq op (getp pred 'operator))
		   (setq pr (car (getp pred 'arguments)))
		   )
	      (evaluate-with-op net op pr arglist effort)
	  (case reln
		(equal (setq res (test-equal net a1 a2 effort)))
		(has-cardinality (setq res (test-cardinality net a1 a2 effort)))
		(subset-of (setq res (test-subset net a1 a2 effort)))
		(member-of (setq res (test-member net a1 a2 effort nil)))
		(member-of-0 (setq res (test-member net a1 a2 effort t)))
		) )
	))
  ; return a yes, no or unknown answer
  (if (null res) (setq res 'no))
  (if (and res (not (member res '(unknown no)))) (setq res 'yes))
  (if (and (traced-p 'epilog::set-eval) (not (equal res 'unknown)))
      (print-line "    Set Specialist: Evaluated " 
	  		       (print-info a1) " " (print-info reln) " " 
			       (if a2 (print-info a2) "") " to " res))
  res)
