;       Last change    August 1993
(in-package time)
;                         FUNCTIONS
; These functions are functions the user may have in propositions
; When evaluated, they use the global specialist effort level for effort

; declare them as functions to the parser

; start-of returns the start point of an event (if concept is an event)

(defun start-of (&rest args &aux
		       (concept (car args))
		       res)
  (setq res 
  (if (or (null concept)
	  (not (symbolp concept))
	  (not (constant-arg concept))
	  (not (member (find-sort concept) '(epilog::episode epilog::time 
					    epilog::event epilog::interval)))
	  (absolute-time concept))
      nil
      (get-start concept) 
      ))
  (when res (change-property res 'epilog::SORT 'epilog::TIME))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::time-function-eval))
      (print-line "    Time Specialist: START-OF " args " evaluated to " 
		  (print-info res)))
  res)

; end-of returns the end point of an event (if concept is an event)

(defun end-of (&rest args &aux
	             (concept (car args))
		     res)
  (setq res
  (if (or (null concept)
	  (not (symbolp concept))
	  (not (constant-arg concept))
	  (not (member (find-sort concept) '(epilog::episode epilog::time 
					    epilog::event epilog::interval)))
	  (absolute-time concept))
      nil
      (get-end concept) 
      ))
  (when res (change-property res 'epilog::SORT 'epilog::TIME))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::time-function-eval))
      (print-line "    Time Specialist: END-OF " args " evaluated to " 
		  (print-info res)))
  res)

; date returns an absolute time concept for the given time specification
; date resturned is a quoted expression of the form
;     (quote (time year month day hour minute second))

(defun date (&rest args &aux
		   (year (first args))
		   (month (second args))
		   (day (third args))
		   (hour (fourth args))
		   (minute (fifth args))
		   (second (sixth args))
		   res)
  (setq res
  (unless (member nil (list year month day hour minute second))
	  (new-record (list year month day hour minute second)
		      'time)
	  ))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::time-function-eval))
      (print-line "    Time Specialist: DATE evaluated to " (print-info res)))
  res)

; relation determines the temporal realtion which holds between a1 and a2

(defun relation (&rest args &aux
		       (a1 (car args))
		       (a2 (cadr args))
		       res)
  (unless (or (null a1) (null a2) )
	  (if (or (not (member (find-sort a1) '(epilog::time epilog::episode)))
	          (not (member (find-sort a2) '(epilog::time epilog::episode))))
	      nil
	      (let ()
		   (setq res
                       (if (or (absolute-time a1)
                               (absolute-time a2))
                           (find-absolute-reln a1 a2 
					       *specialist-eval-effort*)
                           (find-relation a1 a2 
					  *specialist-eval-effort*)))
		   (setq res (find-symbol (symbol-name res) 'epilog))
		   (if (or (traced-p 'epilog::function-eval)
			   (traced-p 'epilog::time-function-eval))
                       (print-line "    Time Specialist: RELATION " 
				      args " evaluated to " (print-info res)))
		   )))
  res)

; elapsed determines the most strict bounds on the duration between a1 and a2
; note - a1 must be before a2 for this
; it returns a quoted expression of the form (quote (min max)) - now
; returns only if there is just one number/bound name

(defun elapsed (&rest args &aux
		       (a1 (car args))
		       (a2 (cadr args))
		       res)
  (unless (or (null a1) (null a2))
	  (setq res
		(calc-duration (get-end a1) 
			      (get-start a2) 
			      *specialist-eval-effort*))
	  (setq res (if (listp res) nil res))
	  )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::time-function-eval))
      (print-line "    Time Specialist: ELAPSED " args " evaluated to " 
		  (print-info res)))
  res)

(defun duration-of (&rest args &aux
		       (a1 (car args))
		       res)
  (when (and a1 (event a1))
    (setq res (append '($ num)
		      (calc-duration (get-start a1)
				     (get-end a1) 
				     *specialist-eval-effort*)))
    ;; (setq res (if (listp res) nil res))
    )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::time-function-eval))
      (print-line "    Time Specialist: DURATION-OF " args " evaluated to " 
		  (print-info res)))
  res)
